Package: climatekit
Title: Unified Climate Indices for Temperature, Precipitation, and
        Drought
Version: 0.2.0
Authors@R: 
    person("Charles", "Coverdale", , "charlesfcoverdale@gmail.com", role = c("aut", "cre"))
Description: Compute the standard suite of climate indices from daily
    weather observations. Provides the canonical 'ETCCDI' 27 (Expert
    Team on Climate Change Detection and Indices), the 'ET-SCI'
    heatwave and cold-wave families plus the Excess Heat Factor of
    Nairn and Fawcett (2013), and agroclimatic, drought, and
    human-comfort families. Drought indices ('SPI', 'SPEI') accept a
    choice of distribution (gamma or Pearson III for SPI; log-logistic
    or generalised extreme value for SPEI). Reference
    evapotranspiration is available via Hargreaves and the FAO-56
    Penman-Monteith method (Allen et al. 1998). Percentile-based
    indices support the Zhang (2005) in-base bootstrap. Daily inputs
    are numeric vectors plus a 'Date' vector; outputs are tidy data
    frames. Optional gridded support via 'terra' applies any index
    over a 'SpatRaster' and reads 'netCDF' input. No external API
    calls; pairs with data packages such as 'readnoaa'. References:
    Alexander et al. (2006) <doi:10.1029/2005JD006290>; Zhang et al.
    (2011) <doi:10.1002/wcc.147>; Zhang et al. (2005)
    <doi:10.1175/JCLI3366.1>.
Depends: R (>= 4.1.0)
License: MIT + file LICENSE
Language: en-GB
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: cli (>= 3.6.0), stats, tools
Suggests: testthat (>= 3.0.0), terra, ncdf4, knitr, rmarkdown
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://charlescoverdale.github.io/climatekit/,
        https://github.com/charlescoverdale/climatekit
BugReports: https://github.com/charlescoverdale/climatekit/issues
NeedsCompilation: no
Packaged: 2026-05-09 07:46:09 UTC; charlescoverdale
Author: Charles Coverdale [aut, cre]
Maintainer: Charles Coverdale <charlesfcoverdale@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-09 16:30:23 UTC
Built: R 4.6.0; ; 2026-05-09 17:39:06 UTC; unix

Package: climatekit
Title: Unified Climate Indices for Temperature, Precipitation, and
        Drought
Version: 0.2.0
Authors@R: 
    person("Charles", "Coverdale", , "charlesfcoverdale@gmail.com", role = c("aut", "cre"))
Description: Compute the standard suite of climate indices from daily
    weather observations. Provides the canonical 'ETCCDI' 27 (Expert
    Team on Climate Change Detection and Indices), the 'ET-SCI'
    heatwave and cold-wave families plus the Excess Heat Factor of
    Nairn and Fawcett (2013), and agroclimatic, drought, and
    human-comfort families. Drought indices ('SPI', 'SPEI') accept a
    choice of distribution (gamma or Pearson III for SPI; log-logistic
    or generalised extreme value for SPEI). Reference
    evapotranspiration is available via Hargreaves and the FAO-56
    Penman-Monteith method (Allen et al. 1998). Percentile-based
    indices support the Zhang (2005) in-base bootstrap. Daily inputs
    are numeric vectors plus a 'Date' vector; outputs are tidy data
    frames. Optional gridded support via 'terra' applies any index
    over a 'SpatRaster' and reads 'netCDF' input. No external API
    calls; pairs with data packages such as 'readnoaa'. References:
    Alexander et al. (2006) <doi:10.1029/2005JD006290>; Zhang et al.
    (2011) <doi:10.1002/wcc.147>; Zhang et al. (2005)
    <doi:10.1175/JCLI3366.1>.
Depends: R (>= 4.1.0)
License: MIT + file LICENSE
Language: en-GB
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: cli (>= 3.6.0), stats, tools
Suggests: testthat (>= 3.0.0), terra, ncdf4, knitr, rmarkdown
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://charlescoverdale.github.io/climatekit/,
        https://github.com/charlescoverdale/climatekit
BugReports: https://github.com/charlescoverdale/climatekit/issues
NeedsCompilation: no
Packaged: 2026-05-09 07:46:09 UTC; charlescoverdale
Author: Charles Coverdale [aut, cre]
Maintainer: Charles Coverdale <charlesfcoverdale@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-09 16:30:23 UTC
Built: R 4.6.0; ; 2026-05-09 18:51:05 UTC; unix

# Generated by roxygen2: do not edit by hand

export(ck_apply_grid)
export(ck_available)
export(ck_branas)
export(ck_browse)
export(ck_catalogue)
export(ck_compute)
export(ck_convert_temp)
export(ck_cooling_degree_days)
export(ck_csdi)
export(ck_cwa)
export(ck_cwd)
export(ck_cwf)
export(ck_cwm)
export(ck_cwn)
export(ck_diurnal_range)
export(ck_dry_days)
export(ck_ehf)
export(ck_etccdi_27)
export(ck_fire_danger)
export(ck_first_frost)
export(ck_from_netcdf)
export(ck_frost_days)
export(ck_growing_degree_days)
export(ck_growing_season)
export(ck_heat_index)
export(ck_heating_degree_days)
export(ck_heavy_precip)
export(ck_huglin)
export(ck_humidex)
export(ck_hwa)
export(ck_hwd)
export(ck_hwf)
export(ck_hwm)
export(ck_hwn)
export(ck_ice_days)
export(ck_last_frost)
export(ck_max_1day_precip)
export(ck_max_5day_precip)
export(ck_metadata)
export(ck_pet)
export(ck_pet_pm)
export(ck_precip_intensity)
export(ck_r95p)
export(ck_r99p)
export(ck_spei)
export(ck_spi)
export(ck_summer_days)
export(ck_tn10p)
export(ck_tn90p)
export(ck_tnn)
export(ck_tnx)
export(ck_total_precip)
export(ck_tropical_nights)
export(ck_tx10p)
export(ck_tx90p)
export(ck_txn)
export(ck_txx)
export(ck_very_heavy_precip)
export(ck_warm_spell)
export(ck_wet_days)
export(ck_wind_chill)
export(ck_winkler)
export(ck_wsdi)
export(clear_cache)
importFrom(stats,pnorm)
importFrom(stats,qnorm)
importFrom(tools,R_user_dir)

# climatekit

[![CRAN status](https://www.r-pkg.org/badges/version/climatekit)](https://CRAN.R-project.org/package=climatekit) [![CRAN downloads](https://cranlogs.r-pkg.org/badges/climatekit)](https://CRAN.R-project.org/package=climatekit) [![Total Downloads](https://cranlogs.r-pkg.org/badges/grand-total/climatekit)](https://CRAN.R-project.org/package=climatekit) [![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable) [![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)

An R package for computing climate indices from daily weather observations. Takes vectors of temperature, precipitation, humidity, and wind data and returns tidy data frames: no file wrangling, no class coercion, no API calls.

## What are climate indices?

Climate indices are standardised summary statistics that reduce daily weather observations into meaningful measures of climate conditions. A single year of weather data for one station is 365 rows of temperature, precipitation, wind, and humidity readings. Climate indices compress that into interpretable numbers: how many frost days occurred, how long the growing season lasted, whether the region is in drought.

These indices matter because they are how climate science connects with the real economy. Energy companies use heating and cooling degree days to forecast demand. Agricultural ministries track growing degree days and frost dates. Water authorities monitor SPI and SPEI drought indices. Urban planners measure heat index exceedances. Insurance actuaries count extreme precipitation events. Viticulturists use Huglin and Winkler indices to assess grape-growing potential. Fire services monitor fire weather indices.

The definitions come from international standards bodies - the WMO Expert Team on Climate Change Detection and Indices (ETCCDI) defines 27 core indices, the Expert Team on Sector-specific Climate Indices (ET-SCI) extends these into health, agriculture, and energy domains, and individual research communities have added domain-specific measures like SPEI for drought and Huglin for viticulture.

## Getting started: where to get the data

`climatekit` computes indices from weather data that you provide. It doesn't download anything itself: you bring the data, it does the maths.

**If you already have data** (a CSV, a database export, an Excel file), all you need is a numeric vector of observations and a date vector:

```r
library(climatekit)

# Read your own data
weather <- read.csv("my_weather_station.csv")

# Compute frost days
ck_frost_days(weather$tmin, weather$date)
```

**If you don't have data yet**, the easiest way to get started is with [`readnoaa`](https://github.com/charlescoverdale/readnoaa), which downloads free daily weather observations from NOAA's global archive of 100,000+ stations. No API key needed:

```r
install.packages("readnoaa")  # or devtools::install_github("charlescoverdale/readnoaa")
library(readnoaa)
library(climatekit)

# Step 1: Find a station near you
noaa_nearby(lat = 51.5, lon = -0.1, radius_km = 25)
#>        station                    name latitude longitude distance_km
#>   UKE00105915     LONDON WEATHER CENTRE   51.517    -0.117        1.4

# Step 2: Download daily data
weather <- noaa_daily("UKE00105915", "2020-01-01", "2024-12-31",
                      datatypes = c("TMAX", "TMIN", "PRCP"))

# Step 3: Compute indices
ck_frost_days(weather$tmin, weather$date, period = "annual")
ck_spi(weather$prcp, weather$date, scale = 3)
ck_heating_degree_days((weather$tmax + weather$tmin) / 2, weather$date)
```

As long as you have a numeric vector and a date vector, `climatekit` will work with it.

### Common data sources

| Region | Source | Coverage | Access |
|---|---|---|---|
| Global | [NOAA GHCNd](https://www.ncei.noaa.gov/products/land-based-station/global-historical-climatology-network-daily) | 100,000+ stations worldwide | Free, no key. Use [`readnoaa`](https://github.com/charlescoverdale/readnoaa) |
| Global | [ERA5 reanalysis](https://cds.climate.copernicus.eu/) | Gridded, 0.25° resolution, 1940–present | Free, requires CDS account |
| UK | [Met Office MIDAS](https://catalogue.ceda.ac.uk/uuid/dbd451271eb04662beade68da43546e1) | ~1,000 UK stations, daily | Free via CEDA, requires registration |
| Europe | [ECA&D](https://www.ecad.eu/) | 20,000+ stations across Europe | Free download |
| US | [ACIS (RCC)](https://www.rcc-acis.org/) | All US cooperative & ASOS stations | Free, no key |
| Australia | [Bureau of Meteorology](https://www.bom.gov.au/climate/data/) | All BoM stations, daily | Free download |

---

## Why does this package exist?

R has the methods, but they are scattered across half a dozen packages with incompatible interfaces:

| Package | Coverage | Limitation |
|---|---|---|
| `ClimInd` | 138 indices (SPI, SPEI, heat/cold waves) | Returns raw vectors with no metadata, no dates, no units |
| `climdex.pcic` | 27 ETCCDI core indices | Requires a custom `climdexInput` S4 object; locked to ETCCDI standard |
| `SPEI` | SPI + SPEI drought indices | Single-purpose; only does drought |
| `heatwaveR` | Marine + atmospheric heatwaves | Single-purpose; only does heatwaves |
| `weathermetrics` | Unit conversions + heat index | No climate indices |

If you want frost days, degree days, SPI, and the Huglin index in the same analysis, you currently need four packages with four different input formats and four different output structures. One wants an S4 object, another wants a matrix, a third wants separate vectors, and none of them return a data frame with dates attached.

`climatekit` replaces all of that with a single interface: vectors in, data frames out. Every function takes the same kind of input (numeric vector + date vector), every function returns the same kind of output (a data frame with `period`, `value`, `index`, and `unit` columns), and the 50+ indices span temperature, precipitation, drought, agroclimatic, and comfort categories.

```r
# Without climatekit: four packages, four input formats, four output structures
library(climdex.pcic)
ci <- climdexInput.raw(tmax = ..., tmin = ..., prec = ..., ...)  # S4 object
fd <- climdex.fd(ci)  # returns named numeric vector, no dates

library(SPEI)
spi_result <- spi(ts(monthly_precip, frequency = 12), 3)  # returns S4, needs ts()

library(ClimInd)
gdd <- gdd(tavg_vector, 10)  # returns raw numeric, no metadata

# With climatekit: one package, one interface
library(climatekit)
ck_frost_days(tmin, dates)                     # → data.frame
ck_spi(precip, dates, scale = 3)               # → data.frame
ck_growing_degree_days(tavg, dates, base = 10)  # → data.frame
ck_huglin(tmin, tmax, dates, lat = 45)          # → data.frame
```

---

## Installation

```r
install.packages("climatekit")

# Or install the development version from GitHub
# install.packages("devtools")
devtools::install_github("charlescoverdale/climatekit")
```

---

## Functions

### ETCCDI canonical 27

The full ETCCDI core set (Alexander et al. 2006; Zhang et al. 2011) is implemented. `ck_etccdi_27()` returns an audit table mapping every code to its `climatekit` function.

| Code | Function | Description |
|---|---|---|
| FD | `ck_frost_days()` | Days where Tmin < 0 °C |
| ID | `ck_ice_days()` | Days where Tmax < 0 °C |
| SU | `ck_summer_days()` | Days where Tmax > 25 °C |
| TR | `ck_tropical_nights()` | Days where Tmin > 20 °C |
| TXx | `ck_txx()` | Annual / monthly max of Tmax |
| TNx | `ck_tnx()` | Annual / monthly max of Tmin (warmest night) |
| TXn | `ck_txn()` | Annual / monthly min of Tmax (coldest day) |
| TNn | `ck_tnn()` | Annual / monthly min of Tmin (coldest night) |
| DTR | `ck_diurnal_range()` | Mean daily temperature range |
| GSL | `ck_growing_season()` | Growing season length |
| TX10p | `ck_tx10p()` | % cool days (calendar-day base, optional Zhang 2005 bootstrap) |
| TN10p | `ck_tn10p()` | % cool nights (calendar-day base, optional bootstrap) |
| TX90p | `ck_tx90p()` | % warm days (calendar-day base, optional bootstrap) |
| TN90p | `ck_tn90p()` | % warm nights (calendar-day base, optional bootstrap) |
| WSDI | `ck_wsdi()` | Warm spell duration index |
| CSDI | `ck_csdi()` | Cold spell duration index |
| RX1day | `ck_max_1day_precip()` | Max 1-day precipitation |
| RX5day | `ck_max_5day_precip()` | Max 5-day precipitation |
| SDII | `ck_precip_intensity()` | Simple daily intensity index |
| R10mm | `ck_heavy_precip()` (default 10) | Days with precip >= 10 mm |
| R20mm | `ck_very_heavy_precip()` (default 20) | Days with precip >= 20 mm |
| Rnnmm | `ck_heavy_precip(threshold = nn)` | Days with precip >= nn mm |
| CDD | `ck_dry_days()` | Max consecutive dry days |
| CWD | `ck_wet_days()` | Max consecutive wet days |
| R95p | `ck_r95p()` | Total precip on very-wet days |
| R99p | `ck_r99p()` | Total precip on extremely-wet days |
| PRCPTOT | `ck_total_precip()` | Annual wet-day precip total |

### ET-SCI heatwave family

Period of >= 3 consecutive days with TX above the calendar-day 90th percentile, plus cold-wave duals (TN below 10th percentile).

| Code | Function | Description |
|---|---|---|
| HWN | `ck_hwn()` | Number of distinct heatwave events |
| HWF | `ck_hwf()` | Total days inside heatwave events |
| HWD | `ck_hwd()` | Longest heatwave duration |
| HWM | `ck_hwm(mode = "excess" / "absolute")` | Mean magnitude across event days |
| HWA | `ck_hwa(mode = "excess" / "absolute")` | Peak magnitude across event days |
| CWN | `ck_cwn()` | Cold-wave number |
| CWF | `ck_cwf()` | Cold-wave frequency |
| CWD | `ck_cwd()` | Cold-wave duration (note: ETCCDI also uses "CWD" for consecutive wet days, which is `ck_wet_days`) |
| CWM | `ck_cwm()` | Cold-wave magnitude |
| CWA | `ck_cwa()` | Cold-wave amplitude |
| EHF | `ck_ehf()` | Excess Heat Factor (Nairn & Fawcett 2013) |

### Drought, evapotranspiration

| Function | Description |
|---|---|
| `ck_spi(distribution = "gamma" / "pearsonIII")` | Standardized Precipitation Index |
| `ck_spei(distribution = "log-logistic" / "gev")` | Standardized Precipitation-Evapotranspiration Index |
| `ck_pet()` | Reference evapotranspiration (Hargreaves) |
| `ck_pet_pm()` | Reference evapotranspiration (FAO-56 Penman-Monteith) |

### Agroclimatic, comfort, energy

| Function | Description |
|---|---|
| `ck_huglin(lat)` | Huglin heliothermal index (viticulture) |
| `ck_winkler()` | Winkler index (wine region classification) |
| `ck_branas(lat)` | Branas hydrothermal index (disease pressure) |
| `ck_first_frost(lat)` | First autumn frost date (NH or SH) |
| `ck_last_frost(lat)` | Last spring frost date (NH or SH) |
| `ck_growing_degree_days()` | Accumulated GDD above base |
| `ck_heating_degree_days()` | Heating degree days |
| `ck_cooling_degree_days()` | Cooling degree days |
| `ck_warm_spell()` | Warm-spell days (series-quantile, simpler variant of WSDI) |
| `ck_wind_chill()` | Wind chill (Environment Canada / NWS) |
| `ck_heat_index()` | Heat index (Rothfusz / NWS) |
| `ck_humidex()` | Canadian humidex |
| `ck_fire_danger()` | Simplified fire-danger proxy (use `cffdrs` for full FWI) |

### Discovery, dispatch, gridded

| Function | Description |
|---|---|
| `ck_etccdi_27()` | Canonical ETCCDI 27 audit table |
| `ck_catalogue()` | Full implementation catalogue |
| `ck_browse(sector, standard, search)` | Filter the catalogue |
| `ck_compute(data, index, ...)` | Dispatch any index by name |
| `ck_available()`, `ck_metadata()` | Lightweight registry queries |
| `ck_convert_temp()` | Celsius / Fahrenheit / Kelvin |
| `ck_apply_grid(x, fun, dates, ...)` | Apply any function over a `terra::SpatRaster` |
| `ck_from_netcdf(path, var)` | Thin reader for netCDF input |
| `clear_cache()` | Clear the user-data cache |

---

## Examples

### How many frost days does a location get?

```r
library(climatekit)

# Daily minimum temperatures for a year
dates <- as.Date("2024-01-01") + 0:364
set.seed(42)
tmin <- sin(seq(0, 2 * pi, length.out = 365)) * 15 + 2

# Annual frost days
ck_frost_days(tmin, dates)
#>       period value      index unit
#>   2024-01-01   132 frost_days days

# Monthly breakdown
ck_frost_days(tmin, dates, period = "monthly")
#>       period value      index unit
#>   2024-01-01    25 frost_days days
#>   2024-02-01    17 frost_days days
#>   2024-03-01     4 frost_days days
#>   ...
```

---

### How much heating energy does a building need?

```r
# Heating degree days tell energy companies how much heating demand to expect.
# Each degree below the base temperature (default 18C) for each day adds to the total.

tavg <- sin(seq(0, 2 * pi, length.out = 365)) * 12 + 10
ck_heating_degree_days(tavg, dates, period = "monthly")
#>       period  value                index        unit
#>   2024-01-01 481.10 heating_degree_days degree-days
#>   2024-02-01 378.49 heating_degree_days degree-days
#>   2024-03-01 244.53 heating_degree_days degree-days
#>   ...

# Cooling degree days for air conditioning demand
ck_cooling_degree_days(tavg, dates, base = 22)
```

---

### Is a region in drought?

```r
# The Standardized Precipitation Index (SPI) fits a gamma distribution to
# monthly precipitation totals over a rolling window, then transforms to
# standard normal deviates. Values below -1 indicate moderate drought,
# below -1.5 severe drought, below -2 extreme drought.

dates_long <- seq(as.Date("2015-01-01"), as.Date("2024-12-31"), by = "day")
set.seed(42)
precip <- rgamma(length(dates_long), shape = 2, rate = 0.5)

spi <- ck_spi(precip, dates_long, scale = 3)
head(spi)
#>       period      value index        unit
#>   2015-03-01 -0.2891577   spi dimensionless
#>   2015-04-01  0.4458927   spi dimensionless
#>   ...

# SPEI adds evapotranspiration to capture temperature-driven drought
pet <- ck_pet(tmin, tmax, lat = 51.5, dates = dates)
```

---

### What wine regions does a climate support?

```r
# The Huglin heliothermal index classifies grape-growing potential:
# < 1500: too cool for viticulture
# 1500-1800: cool climate (Champagne, Mosel)
# 1800-2100: temperate (Burgundy, Oregon)
# 2100-2400: warm (Bordeaux, Napa)
# > 2400: hot (Barossa, Southern Spain)

dates_gs <- seq(as.Date("2024-04-01"), as.Date("2024-09-30"), by = "day")
set.seed(42)
tmin_gs <- rnorm(length(dates_gs), mean = 12, sd = 3)
tmax_gs <- tmin_gs + runif(length(dates_gs), 8, 15)

ck_huglin(tmin_gs, tmax_gs, dates_gs, lat = 45)
#>       period    value  index        unit
#>   2024-01-01 2129.284 huglin degree-days

# Winkler index (wine region classification)
tavg_gs <- (tmin_gs + tmax_gs) / 2
ck_winkler(tavg_gs, dates_gs)
```

---

### When did frost season start and end?

```r
# First and last frost dates matter for agriculture, construction, and transport.

dates_year <- as.Date("2024-01-01") + 0:364
set.seed(42)
tmin_year <- -10 + seq_along(dates_year) * 0.08 + rnorm(365, sd = 4)

ck_last_frost(tmin_year, dates_year)
#>       period value       date      index       unit
#>   2024-01-01   120 2024-04-29 last_frost day of year

ck_first_frost(tmin_year, dates_year)
```

---

### How dangerous is a heatwave?

```r
# The heat index combines temperature and humidity to estimate
# how hot it actually feels. Values above 40C are dangerous.

ck_heat_index(tavg = c(30, 33, 36, 39), humidity = c(60, 65, 70, 75))
#>      value      index unit
#>   32.94844 heat_index   °C
#>   38.67052 heat_index   °C
#>   47.57163 heat_index   °C
#>   60.56858 heat_index   °C

# Wind chill for cold conditions
ck_wind_chill(tavg = c(-5, -10, -15), wind_speed = c(20, 30, 40))

# Fire weather risk
ck_fire_danger(tavg = 35, humidity = 15, wind_speed = 30, precip = 0)
```

---

### Removing the in-base bias with the Zhang (2005) bootstrap

```r
# The percentile-day indices (TX10p, TN10p, TX90p, TN90p) compute
# thresholds from a reference period (default 1961-1990). For analysis
# years inside the reference period, the year being assessed contributes
# to its own threshold and biases the result toward 10% / 90%. Set
# bootstrap = TRUE to apply Zhang et al. (2005) leave-one-out resampling
# (the canonical climdex.pcic / climpact behaviour). Costs roughly
# N^2 percentile fits for an N-year reference; opt in for attribution
# work spanning the base.

ck_tx10p(tmax, dates, ref_start = 1961L, ref_end = 1990L, bootstrap = TRUE)
```

### Operational heatwave intensity (Excess Heat Factor)

```r
# EHF combines a 3-day mean temperature anomaly above the 95th
# percentile with an acclimatisation term. Positive EHF days are
# heatwave conditions; larger values indicate more severe events.

ck_ehf(tmax, tmin, dates, ref_start = 1961L, ref_end = 1990L,
       stat = "max")           # peak EHF in year
ck_ehf(tmax, tmin, dates, stat = "n_positive")    # count heatwave-condition days
ck_ehf(tmax, tmin, dates, stat = "sum_positive")  # severity-weighted total
```

### FAO-56 Penman-Monteith reference evapotranspiration

```r
# ck_pet() is the Hargreaves estimator (Tmin / Tmax / lat only).
# ck_pet_pm() is the international FAO-56 Penman-Monteith standard,
# with optional humidity, wind, solar-radiation, and elevation inputs.
# Sensible FAO-56 fallbacks are used where data are missing.

ck_pet_pm(tmin, tmax, lat = 45, dates = dates,
          elev = 200, wind = 2.5,
          rh_min = rh_min, rh_max = rh_max)
```

### Computing indices programmatically

```r
# If you are computing many indices over the same dataset, use ck_compute()
# with the index name as a string. This is useful in loops, Shiny apps,
# or any workflow where the index is selected at runtime.

weather <- data.frame(
  dates = as.Date("2024-01-01") + 0:364,
  tmin = sin(seq(0, 2 * pi, length.out = 365)) * 15 + 2,
  tmax = sin(seq(0, 2 * pi, length.out = 365)) * 15 + 12,
  precip = rgamma(365, shape = 0.5, rate = 0.2)
)

# Compute any index by name
ck_compute(weather, "frost_days")
ck_compute(weather, "total_precip", period = "monthly")

# See all available indices
ck_available()
#>                 index      category          unit
#>            frost_days   temperature          days
#>              ice_days   temperature          days
#>          summer_days    temperature          days
#>    tropical_nights     temperature          days
#>    ...
```

---

## Related packages

| Package | Description |
|---|---|
| [`readnoaa`](https://github.com/charlescoverdale/readnoaa) | NOAA weather and climate data (pairs with climatekit for data acquisition) |
| [`carbondata`](https://github.com/charlescoverdale/carbondata) | Carbon market data (EU/UK ETS, voluntary registries) |
| [`cer`](https://github.com/charlescoverdale/cer) | Clean Energy Regulator data (Australia) |
| [`aemo`](https://github.com/charlescoverdale/aemo) | Australian Energy Market Operator data |

---

## Migrating from `climdex.pcic`

`climdex.pcic` (Pacific Climate Impacts Consortium) was for many years the standard R implementation of the canonical ETCCDI 27. It was archived from CRAN in 2023. `climatekit` covers the same set with a simpler interface:

```r
# climdex.pcic
ci <- climdexInput.raw(tmax, tmin, prec, ..., base.range = c(1961, 1990))
fd <- climdex.fd(ci)        # named numeric vector

# climatekit
fd <- ck_frost_days(tmin, dates)   # tidy data frame
```

See `vignette("climdex-migration", package = "climatekit")` for the full function-name crosswalk and interface-shift notes.

## Citation

```r
citation("climatekit")
```

If you use the package in academic work, please also cite Alexander et al. (2006) and Zhang et al. (2011) (the canonical ETCCDI references), and Zhang et al. (2005) if you use the in-base bootstrap. `inst/CITATION` and the root-level `CITATION.cff` provide the bibentries.

## Issues

Please report bugs or requests at <https://github.com/charlescoverdale/climatekit/issues>.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# # climdex.pcic
# ci <- climdexInput.raw(
#   tmax, tmin, precip, tmax.dates, tmin.dates, prec.dates,
#   base.range = c(1961, 1990)
# )
# fd <- climdex.fd(ci)
# 
# # climatekit
# fd <- ck_frost_days(tmin, dates)

## ----eval = FALSE-------------------------------------------------------------
# ck_frost_days(tmin, dates)
# #> # period       value index        unit
# #> # 2020-01-01      72 frost_days   days
# #> # 2021-01-01      68 frost_days   days
# #> # ...

## ----eval = FALSE-------------------------------------------------------------
# ck_tx90p(tmax, dates, ref_start = 1981L, ref_end = 2010L)

## ----eval = FALSE-------------------------------------------------------------
# # Old climdex.pcic outputs
# old_fd <- climdex.fd(ci)
# old_txx <- climdex.txx(ci)
# 
# # New climatekit outputs
# new_fd  <- ck_frost_days(tmin, dates)
# new_txx <- ck_txx(tmax, dates)
# 
# # Compare; expect agreement up to NA-handling and base-period self-inclusion
# all.equal(old_fd, new_fd$value)
# all.equal(old_txx, new_txx$value)

#' Frost Days
#'
#' Count the number of days where minimum temperature is below 0 degrees C.
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmin`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' tmin <- c(-2, 3, -1, 5, -3, 0, 2, -4, 1, -1)
#' ck_frost_days(tmin, dates)
ck_frost_days <- function(tmin, dates, period = "annual") {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  period <- validate_period(period)

  result <- count_by_period(tmin < 0, dates, period)
  build_result(result$periods, result$values, "frost_days", "days", period)
}

#' Ice Days
#'
#' Count the number of days where maximum temperature is below 0 degrees C.
#'
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmax`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' tmax <- c(-2, 3, -1, 5, -3, 0, 2, -4, 1, -1)
#' ck_ice_days(tmax, dates)
ck_ice_days <- function(tmax, dates, period = "annual") {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  period <- validate_period(period)

  result <- count_by_period(tmax < 0, dates, period)
  build_result(result$periods, result$values, "ice_days", "days", period)
}

#' Summer Days
#'
#' Count the number of days where maximum temperature exceeds 25 degrees C.
#'
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmax`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-07-01") + 0:9
#' tmax <- c(22, 26, 28, 24, 30, 25, 27, 23, 31, 29)
#' ck_summer_days(tmax, dates)
ck_summer_days <- function(tmax, dates, period = "annual") {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  period <- validate_period(period)

  result <- count_by_period(tmax > 25, dates, period)
  build_result(result$periods, result$values, "summer_days", "days", period)
}

#' Tropical Nights
#'
#' Count the number of days where minimum temperature exceeds 20 degrees C.
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmin`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-07-01") + 0:9
#' tmin <- c(18, 21, 22, 19, 25, 20, 23, 17, 24, 21)
#' ck_tropical_nights(tmin, dates)
ck_tropical_nights <- function(tmin, dates, period = "annual") {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  period <- validate_period(period)

  result <- count_by_period(tmin > 20, dates, period)
  build_result(result$periods, result$values, "tropical_nights", "days", period)
}

#' Annual or Monthly Maximum of Daily Maximum Temperature (TXx)
#'
#' ETCCDI canonical index TXx. The maximum value of daily maximum
#' temperature (Tmax) within each reporting period.
#'
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmax`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' tmax <- c(5, 10, 18, 12, 4, 8, 22, 3, 7, 6)
#' ck_txx(tmax, dates)
ck_txx <- function(tmax, dates, period = "annual") {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  period <- validate_period(period)

  result <- aggregate_by_period(tmax, dates, period, max)
  build_result(result$periods, result$values, "txx", "\u00b0C", period)
}

#' Annual or Monthly Maximum of Daily Minimum Temperature (TNx)
#'
#' ETCCDI canonical index TNx. The maximum value of daily minimum
#' temperature (Tmin) within each reporting period (warmest night).
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmin`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-07-01") + 0:9
#' tmin <- c(15, 18, 22, 19, 14, 21, 23, 17, 20, 19)
#' ck_tnx(tmin, dates)
ck_tnx <- function(tmin, dates, period = "annual") {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  period <- validate_period(period)

  result <- aggregate_by_period(tmin, dates, period, max)
  build_result(result$periods, result$values, "tnx", "\u00b0C", period)
}

#' Annual or Monthly Minimum of Daily Maximum Temperature (TXn)
#'
#' ETCCDI canonical index TXn. The minimum value of daily maximum
#' temperature (Tmax) within each reporting period (coldest day).
#'
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmax`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' tmax <- c(5, 10, -3, 12, 4, 8, 22, -8, 7, 6)
#' ck_txn(tmax, dates)
ck_txn <- function(tmax, dates, period = "annual") {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  period <- validate_period(period)

  result <- aggregate_by_period(tmax, dates, period, min)
  build_result(result$periods, result$values, "txn", "\u00b0C", period)
}

#' Annual or Monthly Minimum of Daily Minimum Temperature (TNn)
#'
#' ETCCDI canonical index TNn. The minimum value of daily minimum
#' temperature (Tmin) within each reporting period (coldest night).
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmin`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' tmin <- c(-2, 3, -1, 5, -8, 0, 2, -12, 1, -1)
#' ck_tnn(tmin, dates)
ck_tnn <- function(tmin, dates, period = "annual") {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  period <- validate_period(period)

  result <- aggregate_by_period(tmin, dates, period, min)
  build_result(result$periods, result$values, "tnn", "\u00b0C", period)
}

#' Percentage of Cool Days (TX10p)
#'
#' ETCCDI canonical index TX10p. Percentage of days where daily Tmax
#' falls below the 10th percentile of the calendar-day distribution
#' from a reference period (default 1961 to 1990). The threshold is
#' computed using a 5-day window centred on each calendar day, pooled
#' across the reference period.
#'
#' Set `bootstrap = TRUE` to apply the Zhang et al. (2005) in-base
#' leave-one-out bootstrap, which removes the self-inclusion bias for
#' years inside the reference period. The bootstrap is computationally
#' expensive (roughly N^2 percentile fits for an N-year reference)
#' but is the canonical climdex.pcic / climpact behaviour and is
#' required for climate-change attribution work that spans the base.
#'
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmax`. Must contain
#'   data covering the reference period.
#' @param ref_start,ref_end Integer. Reference period boundary years
#'   (inclusive). Defaults to 1961 and 1990.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#' @param bootstrap Logical. If `TRUE`, apply the Zhang (2005) in-base
#'   bootstrap correction. Default `FALSE` for backward compatibility
#'   and speed.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @references Zhang, X., Hegerl, G. C., Zwiers, F. W., & Kenyon, J.
#'   (2005). Avoiding inhomogeneity in percentile-based indices of
#'   temperature extremes. *Journal of Climate*, 18(11), 1641-1651.
#'   \doi{10.1175/JCLI3366.1}.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_tx10p(tmax, dates))
ck_tx10p <- function(tmax, dates, ref_start = 1961L, ref_end = 1990L,
                     period = "annual", bootstrap = FALSE) {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  period <- validate_period(period)
  if (!is.logical(bootstrap) || length(bootstrap) != 1L) {
    cli::cli_abort("{.arg bootstrap} must be a single logical value.")
  }

  soft_ex <- .pct_exceedance_per_day(tmax, dates, percentile = 0.10,
                                     op = "<", ref_start, ref_end,
                                     bootstrap = bootstrap)
  agg <- .aggregate_pct_exceedance(soft_ex, dates, period)
  build_result(agg$periods, agg$values, "tx10p", "%", period)
}

#' Percentage of Cool Nights (TN10p)
#'
#' ETCCDI canonical index TN10p. Percentage of days where daily Tmin
#' falls below the 10th percentile of the calendar-day distribution
#' from a reference period (default 1961 to 1990). Computation follows
#' the same convention as [ck_tx10p()] and supports the same
#' `bootstrap` argument.
#'
#' @inheritParams ck_tx10p
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_tn10p(tmin, dates))
ck_tn10p <- function(tmin, dates, ref_start = 1961L, ref_end = 1990L,
                     period = "annual", bootstrap = FALSE) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  period <- validate_period(period)
  if (!is.logical(bootstrap) || length(bootstrap) != 1L) {
    cli::cli_abort("{.arg bootstrap} must be a single logical value.")
  }

  soft_ex <- .pct_exceedance_per_day(tmin, dates, percentile = 0.10,
                                     op = "<", ref_start, ref_end,
                                     bootstrap = bootstrap)
  agg <- .aggregate_pct_exceedance(soft_ex, dates, period)
  build_result(agg$periods, agg$values, "tn10p", "%", period)
}

#' Percentage of Warm Days (TX90p)
#'
#' ETCCDI canonical index TX90p. Percentage of days where daily Tmax
#' exceeds the 90th percentile of the calendar-day distribution from a
#' reference period (default 1961 to 1990). Computation follows the
#' same convention as [ck_tx10p()] and supports the same `bootstrap`
#' argument.
#'
#' @inheritParams ck_tx10p
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_tx90p(tmax, dates))
ck_tx90p <- function(tmax, dates, ref_start = 1961L, ref_end = 1990L,
                     period = "annual", bootstrap = FALSE) {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  period <- validate_period(period)
  if (!is.logical(bootstrap) || length(bootstrap) != 1L) {
    cli::cli_abort("{.arg bootstrap} must be a single logical value.")
  }

  soft_ex <- .pct_exceedance_per_day(tmax, dates, percentile = 0.90,
                                     op = ">", ref_start, ref_end,
                                     bootstrap = bootstrap)
  agg <- .aggregate_pct_exceedance(soft_ex, dates, period)
  build_result(agg$periods, agg$values, "tx90p", "%", period)
}

#' Percentage of Warm Nights (TN90p)
#'
#' ETCCDI canonical index TN90p. Percentage of days where daily Tmin
#' exceeds the 90th percentile of the calendar-day distribution from a
#' reference period (default 1961 to 1990). Computation follows the
#' same convention as [ck_tx10p()] and supports the same `bootstrap`
#' argument.
#'
#' @inheritParams ck_tn10p
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_tn90p(tmin, dates))
ck_tn90p <- function(tmin, dates, ref_start = 1961L, ref_end = 1990L,
                     period = "annual", bootstrap = FALSE) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  period <- validate_period(period)
  if (!is.logical(bootstrap) || length(bootstrap) != 1L) {
    cli::cli_abort("{.arg bootstrap} must be a single logical value.")
  }

  soft_ex <- .pct_exceedance_per_day(tmin, dates, percentile = 0.90,
                                     op = ">", ref_start, ref_end,
                                     bootstrap = bootstrap)
  agg <- .aggregate_pct_exceedance(soft_ex, dates, period)
  build_result(agg$periods, agg$values, "tn90p", "%", period)
}

#' Growing Season Length
#'
#' Compute the growing season length following the ETCCDI definition: the
#' number of days between the first occurrence of at least 6 consecutive days
#' with daily mean temperature above 5 degrees C and the first span of 6
#' consecutive days with Tmean below 5 degrees C after July 1 (Northern
#' Hemisphere) or January 1 (Southern Hemisphere). Calculated per year.
#'
#' @param tavg Numeric vector of daily mean temperatures (degrees C).
#' @param dates Date vector of the same length as `tavg`.
#' @param lat Numeric. Latitude in decimal degrees (used to determine
#'   hemisphere for end-of-season rule). Default 50 (Northern Hemisphere).
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:364
#' set.seed(42)
#' tavg <- sin(seq(0, 2 * pi, length.out = 365)) * 15 + 5
#' ck_growing_season(tavg, dates)
ck_growing_season <- function(tavg, dates, lat = 50) {
  validate_numeric(tavg, "tavg")
  validate_dates(dates, length(tavg))
  if (!is.numeric(lat) || length(lat) != 1) {
    cli::cli_abort("{.arg lat} must be a single numeric value.")
  }

  years <- as.integer(format(dates, "%Y"))
  unique_years <- unique(years)

  # Day-of-year after which to look for end of season
  # NH: after July 1 (doy 182); SH: after Jan 1 (doy 1)
  mid_doy <- if (lat >= 0) 182L else 1L

  values <- vapply(unique_years, function(yr) {
    idx <- which(years == yr)
    temps <- tavg[idx]
    doy <- as.integer(format(dates[idx], "%j"))
    n <- length(temps)
    if (n < 6) return(NA_real_)

    above5 <- temps > 5

    # Find first 6-day spell above 5C
    first_start <- NA_integer_
    for (i in seq_len(n - 5)) {
      if (all(above5[i:(i + 5)])) {
        first_start <- i
        break
      }
    }
    if (is.na(first_start)) return(0)

    # Find end of season: first 6-day spell below 5C after mid-year
    season_end <- n  # default: end of data
    for (i in seq_len(n - 5)) {
      if (doy[i] >= mid_doy && all(!above5[i:(i + 5)])) {
        season_end <- i - 1L
        break
      }
    }

    as.double(max(season_end - first_start + 1, 0))
  }, numeric(1))

  build_result(as.character(unique_years), values, "growing_season", "days", "annual")
}

#' Heating Degree Days
#'
#' Sum of `(base - Tavg)` for all days where daily average temperature is
#' below the base temperature (default 18 degrees C).
#'
#' @param tavg Numeric vector of daily average temperatures (degrees C).
#' @param dates Date vector of the same length as `tavg`.
#' @param base Numeric. Base temperature in degrees C (default 18).
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' tavg <- c(5, 10, 15, 20, 8, 12, 18, 3, 25, 7)
#' ck_heating_degree_days(tavg, dates)
ck_heating_degree_days <- function(tavg, dates, base = 18, period = "annual") {
  validate_numeric(tavg, "tavg")
  validate_dates(dates, length(tavg))
  period <- validate_period(period)

  hdd <- pmax(base - tavg, 0)
  result <- aggregate_by_period(hdd, dates, period, sum)
  build_result(result$periods, result$values, "heating_degree_days", "degree-days", period)
}

#' Cooling Degree Days
#'
#' Sum of `(Tavg - base)` for all days where daily average temperature is
#' above the base temperature (default 18 degrees C).
#'
#' @param tavg Numeric vector of daily average temperatures (degrees C).
#' @param dates Date vector of the same length as `tavg`.
#' @param base Numeric. Base temperature in degrees C (default 18).
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-07-01") + 0:9
#' tavg <- c(25, 30, 22, 20, 28, 19, 32, 17, 35, 27)
#' ck_cooling_degree_days(tavg, dates)
ck_cooling_degree_days <- function(tavg, dates, base = 18, period = "annual") {
  validate_numeric(tavg, "tavg")
  validate_dates(dates, length(tavg))
  period <- validate_period(period)

  cdd <- pmax(tavg - base, 0)
  result <- aggregate_by_period(cdd, dates, period, sum)
  build_result(result$periods, result$values, "cooling_degree_days", "degree-days", period)
}

#' Growing Degree Days
#'
#' Sum of `(Tavg - base)` for all days where daily average temperature is
#' above the base temperature (default 10 degrees C).
#'
#' @param tavg Numeric vector of daily average temperatures (degrees C).
#' @param dates Date vector of the same length as `tavg`.
#' @param base Numeric. Base temperature in degrees C (default 10).
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-07-01") + 0:9
#' tavg <- c(15, 20, 8, 12, 25, 9, 30, 11, 22, 18)
#' ck_growing_degree_days(tavg, dates)
ck_growing_degree_days <- function(tavg, dates, base = 10, period = "annual") {
  validate_numeric(tavg, "tavg")
  validate_dates(dates, length(tavg))
  period <- validate_period(period)

  gdd <- pmax(tavg - base, 0)
  result <- aggregate_by_period(gdd, dates, period, sum)
  build_result(result$periods, result$values, "growing_degree_days", "degree-days", period)
}

#' Diurnal Temperature Range
#'
#' Mean daily temperature range (Tmax - Tmin) per period.
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmin` and `tmax`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' tmin <- c(-2, 3, -1, 5, -3, 0, 2, -4, 1, -1)
#' tmax <- c(5, 10, 6, 12, 4, 8, 9, 3, 7, 6)
#' ck_diurnal_range(tmin, tmax, dates)
ck_diurnal_range <- function(tmin, tmax, dates, period = "annual") {
  validate_numeric(tmin, "tmin")
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmin))
  if (length(tmin) != length(tmax)) {
    cli::cli_abort("{.arg tmin} and {.arg tmax} must have the same length.")
  }

  dtr <- tmax - tmin
  result <- aggregate_by_period(dtr, dates, period, mean)
  build_result(result$periods, result$values, "diurnal_range", "\u00b0C", period)
}

#' Warm Spell Duration Index (WSDI)
#'
#' ETCCDI canonical index WSDI. Annual count of days in spans of at least
#' six consecutive days where daily Tmax exceeds the 90th percentile of the
#' calendar-day distribution from a reference period (default 1961 to 1990).
#' The threshold is computed using a 5-day window centred on each calendar
#' day, pooled across the reference period.
#'
#' Days inside qualifying spells are counted into the year they fall in;
#' a spell that crosses a year boundary contributes to both years
#' proportionally. Annual aggregation only.
#'
#' @inheritParams ck_tx10p
#' @param min_spell Integer. Minimum spell length in days (default 6,
#'   the ETCCDI standard).
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_wsdi(tmax, dates))
ck_wsdi <- function(tmax, dates, ref_start = 1961L, ref_end = 1990L,
                    min_spell = 6L) {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))

  thresholds <- .calendar_day_percentile(tmax, dates, 0.90,
                                         ref_start, ref_end, window = 5L)
  doy <- as.integer(format(dates, "%j"))
  above <- !is.na(tmax) & tmax > thresholds[doy]
  in_spell <- .find_spells(above, as.integer(min_spell))

  result <- count_by_period(in_spell, dates, "annual")
  build_result(result$periods, result$values, "wsdi", "days", "annual")
}

#' Cold Spell Duration Index (CSDI)
#'
#' ETCCDI canonical index CSDI. Annual count of days in spans of at least
#' six consecutive days where daily Tmin falls below the 10th percentile of
#' the calendar-day distribution from a reference period (default 1961 to
#' 1990). Cold-side counterpart to [ck_wsdi()].
#'
#' @inheritParams ck_tn10p
#' @param min_spell Integer. Minimum spell length in days (default 6,
#'   the ETCCDI standard).
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_csdi(tmin, dates))
ck_csdi <- function(tmin, dates, ref_start = 1961L, ref_end = 1990L,
                    min_spell = 6L) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))

  thresholds <- .calendar_day_percentile(tmin, dates, 0.10,
                                         ref_start, ref_end, window = 5L)
  doy <- as.integer(format(dates, "%j"))
  below <- !is.na(tmin) & tmin < thresholds[doy]
  in_spell <- .find_spells(below, as.integer(min_spell))

  result <- count_by_period(in_spell, dates, "annual")
  build_result(result$periods, result$values, "csdi", "days", "annual")
}

#' Warm Spell Days (Series-Quantile Approximation)
#'
#' Count the number of days in warm spells, where a warm spell is defined
#' as at least six consecutive days with Tmax above the `threshold`
#' quantile of the full input series. This is a quick approximation
#' driven by a single series-wide quantile, and does not require a
#' reference period.
#'
#' For the canonical ETCCDI WSDI definition (1961-1990 calendar-day base,
#' six-day spell rule), use [ck_wsdi()].
#'
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmax`.
#' @param threshold Numeric. Quantile threshold (default 0.9, i.e. 90th
#'   percentile).
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:364
#' set.seed(42)
#' tmax <- rnorm(365, mean = 20, sd = 5)
#' ck_warm_spell(tmax, dates)
ck_warm_spell <- function(tmax, dates, threshold = 0.9, period = "annual") {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  period <- validate_period(period)

  thresh_val <- stats::quantile(tmax, threshold, na.rm = TRUE)
  above <- tmax > thresh_val

  # Mark days that are part of spells >= 6 days
  in_spell <- logical(length(tmax))
  n <- length(tmax)
  i <- 1
  while (i <= n) {
    if (!is.na(above[i]) && above[i]) {
      start <- i
      while (i <= n && !is.na(above[i]) && above[i]) {
        i <- i + 1
      }
      spell_len <- i - start
      if (spell_len >= 6) {
        in_spell[start:(i - 1)] <- TRUE
      }
    } else {
      i <- i + 1
    }
  }

  result <- count_by_period(in_spell, dates, period)
  build_result(result$periods, result$values, "warm_spell", "days", period)
}

test_that("ck_frost_days counts days below 0", {
  dates <- as.Date("2024-01-01") + 0:9
  tmin <- c(-2, 3, -1, 5, -3, 0, 2, -4, 1, -1)
  result <- ck_frost_days(tmin, dates)
  expect_s3_class(result, "data.frame")
  expect_equal(result$value, 5)
  expect_equal(result$index, "frost_days")
  expect_equal(result$unit, "days")
})

test_that("ck_frost_days monthly aggregation works", {
  dates <- c(as.Date("2024-01-01") + 0:4, as.Date("2024-02-01") + 0:4)
  tmin <- c(-2, 3, -1, 5, -3, 0, 2, -4, 1, -1)
  result <- ck_frost_days(tmin, dates, period = "monthly")
  expect_equal(nrow(result), 2)
  expect_equal(result$value, c(3, 2))
})

test_that("ck_frost_days rejects non-numeric", {
  dates <- as.Date("2024-01-01") + 0:2
  expect_error(ck_frost_days("a", dates), "numeric")
})

test_that("ck_frost_days rejects mismatched lengths", {
  expect_error(ck_frost_days(c(1, 2), as.Date("2024-01-01")), "same length")
})

test_that("ck_ice_days counts days where tmax < 0", {
  dates <- as.Date("2024-01-01") + 0:4
  tmax <- c(-2, 3, -1, 5, -3)
  result <- ck_ice_days(tmax, dates)
  expect_equal(result$value, 3)
})

test_that("ck_summer_days counts days where tmax > 25", {
  dates <- as.Date("2024-07-01") + 0:4
  tmax <- c(22, 26, 28, 24, 30)
  result <- ck_summer_days(tmax, dates)
  expect_equal(result$value, 3)
})

test_that("ck_tropical_nights counts days where tmin > 20", {
  dates <- as.Date("2024-07-01") + 0:4
  tmin <- c(18, 21, 22, 19, 25)
  result <- ck_tropical_nights(tmin, dates)
  expect_equal(result$value, 3)
})

test_that("ck_growing_season returns days using 6-day spells and tavg", {
  # Create a year of data with clear warm season
  dates <- as.Date("2024-01-01") + 0:364
  tavg <- sin(seq(0, 2 * pi, length.out = 365)) * 15 + 5
  result <- ck_growing_season(tavg, dates)
  expect_s3_class(result, "data.frame")
  expect_true(result$value > 0)
  expect_equal(result$index, "growing_season")
})

test_that("ck_growing_season returns 0 when no 6-day spell", {
  dates <- as.Date("2024-01-01") + 0:29
  tavg <- rep(-10, 30)
  result <- ck_growing_season(tavg, dates)
  expect_equal(result$value, 0)
})

test_that("ck_growing_season uses 6-day spell not 5-day", {
  # Exactly 5 warm days should NOT trigger GSL
  dates <- as.Date("2024-01-01") + 0:364
  tavg <- rep(0, 365)
  tavg[100:104] <- 10  # only 5 warm days
  result <- ck_growing_season(tavg, dates)
  expect_equal(result$value, 0)

  # 6 warm days should trigger
  tavg[100:105] <- 10
  result <- ck_growing_season(tavg, dates)
  expect_true(result$value > 0)
})

test_that("ck_growing_season ends at first cold spell after July 1", {
  dates <- as.Date("2024-01-01") + 0:364
  # Warm from day 60 to day 250, then cold
  tavg <- rep(0, 365)
  tavg[60:250] <- 10
  result <- ck_growing_season(tavg, dates, lat = 50)
  # Season starts at day 60, ends when cold spell starts after July 1
  # Cold spell of 6+ days starts at day 251
  # End of season = day 250
  expect_equal(result$value, 250 - 60 + 1)
})

test_that("ck_heating_degree_days sums correctly", {
  dates <- as.Date("2024-01-01") + 0:4
  tavg <- c(10, 15, 20, 5, 18)
  result <- ck_heating_degree_days(tavg, dates, base = 18)
  # (18-10) + (18-15) + 0 + (18-5) + 0 = 8 + 3 + 13 = 24
  expect_equal(result$value, 24)
})

test_that("ck_cooling_degree_days sums correctly", {
  dates <- as.Date("2024-07-01") + 0:4
  tavg <- c(20, 25, 15, 30, 18)
  result <- ck_cooling_degree_days(tavg, dates, base = 18)
  # (20-18) + (25-18) + 0 + (30-18) + 0 = 2 + 7 + 12 = 21
  expect_equal(result$value, 21)
})

test_that("ck_growing_degree_days sums correctly", {
  dates <- as.Date("2024-07-01") + 0:4
  tavg <- c(15, 20, 8, 12, 25)
  result <- ck_growing_degree_days(tavg, dates, base = 10)
  # (15-10) + (20-10) + 0 + (12-10) + (25-10) = 5 + 10 + 2 + 15 = 32
  expect_equal(result$value, 32)
})

test_that("ck_diurnal_range computes mean range", {
  dates <- as.Date("2024-01-01") + 0:4
  tmin <- c(0, 5, 2, 3, 1)
  tmax <- c(10, 15, 12, 13, 11)
  result <- ck_diurnal_range(tmin, tmax, dates)
  expect_equal(result$value, 10)
  expect_equal(result$unit, "\u00b0C")
})

test_that("ck_diurnal_range rejects mismatched lengths", {
  dates <- as.Date("2024-01-01") + 0:4
  expect_error(ck_diurnal_range(1:5, 1:3, dates), "same length")
})

test_that("ck_warm_spell detects spells >= 6 days", {
  dates <- as.Date("2024-01-01") + 0:19
  # 90th percentile will be high; create a clear 7-day spell above it
  tmax <- c(rep(10, 13), rep(50, 7))
  result <- ck_warm_spell(tmax, dates, threshold = 0.5)
  expect_true(result$value >= 7)
})

test_that("ck_warm_spell returns 0 with no long spells", {
  dates <- as.Date("2024-01-01") + 0:9
  tmax <- c(30, 10, 30, 10, 30, 10, 30, 10, 30, 10)
  result <- ck_warm_spell(tmax, dates)
  expect_equal(result$value, 0)
})

test_that("all temperature functions handle NA values", {
  dates <- as.Date("2024-01-01") + 0:4
  tmin <- c(-2, NA, -1, 5, -3)
  result <- ck_frost_days(tmin, dates)
  expect_s3_class(result, "data.frame")
})

test_that("temperature functions reject non-Date dates", {
  expect_error(ck_frost_days(c(1, 2), c("a", "b")), "Date")
})

# --- Reference value tests ---

test_that("frost days at exactly 0C returns 0 (strict < 0)", {
  dates <- as.Date("2024-01-01")
  result <- ck_frost_days(c(0), dates)
  expect_equal(result$value, 0)
})

test_that("HDD with base=0 returns 0 when all temps >= 0", {
  dates <- as.Date("2024-01-01") + 0:4
  tavg <- c(0, 5, 10, 15, 20)
  result <- ck_heating_degree_days(tavg, dates, base = 0)
  expect_equal(result$value, 0)
})

test_that("CDD with base=0 returns sum of all temps", {
  dates <- as.Date("2024-01-01") + 0:4
  tavg <- c(5, 10, 15, 20, 25)
  result <- ck_cooling_degree_days(tavg, dates, base = 0)
  expect_equal(result$value, 75)
})

# ETCCDI extreme-value indices (TXx / TNx / TXn / TNn) ------------------------

test_that("ck_txx returns max of daily Tmax", {
  dates <- as.Date("2024-01-01") + 0:9
  tmax <- c(5, 10, 18, 12, 4, 8, 22, 3, 7, 6)
  result <- ck_txx(tmax, dates)
  expect_s3_class(result, "data.frame")
  expect_equal(result$value, 22)
  expect_equal(result$index, "txx")
  expect_equal(result$unit, "\u00b0C")
})

test_that("ck_txx monthly aggregation works", {
  dates <- c(as.Date("2024-01-01") + 0:4, as.Date("2024-02-01") + 0:4)
  tmax <- c(5, 10, 18, 12, 4, 8, 22, 3, 7, 6)
  result <- ck_txx(tmax, dates, period = "monthly")
  expect_equal(nrow(result), 2)
  expect_equal(result$value, c(18, 22))
})

test_that("ck_tnx returns max of daily Tmin (warmest night)", {
  dates <- as.Date("2024-07-01") + 0:9
  tmin <- c(15, 18, 22, 19, 14, 21, 23, 17, 20, 19)
  result <- ck_tnx(tmin, dates)
  expect_equal(result$value, 23)
  expect_equal(result$index, "tnx")
})

test_that("ck_txn returns min of daily Tmax (coldest day)", {
  dates <- as.Date("2024-01-01") + 0:9
  tmax <- c(5, 10, -3, 12, 4, 8, 22, -8, 7, 6)
  result <- ck_txn(tmax, dates)
  expect_equal(result$value, -8)
  expect_equal(result$index, "txn")
})

test_that("ck_tnn returns min of daily Tmin (coldest night)", {
  dates <- as.Date("2024-01-01") + 0:9
  tmin <- c(-2, 3, -1, 5, -8, 0, 2, -12, 1, -1)
  result <- ck_tnn(tmin, dates)
  expect_equal(result$value, -12)
  expect_equal(result$index, "tnn")
})

test_that("extreme-value functions reject non-numeric", {
  dates <- as.Date("2024-01-01") + 0:2
  expect_error(ck_txx("a", dates), "numeric")
  expect_error(ck_tnx("a", dates), "numeric")
  expect_error(ck_txn("a", dates), "numeric")
  expect_error(ck_tnn("a", dates), "numeric")
})

test_that("extreme-value functions reject mismatched lengths", {
  expect_error(ck_txx(c(1, 2), as.Date("2024-01-01")), "same length")
  expect_error(ck_tnn(c(1, 2), as.Date("2024-01-01")), "same length")
})

test_that("ETCCDI: txx >= tnx >= tnn and txx >= txn", {
  set.seed(42)
  dates <- as.Date("2024-01-01") + 0:364
  tmin <- rnorm(365, mean = 5, sd = 8)
  tmax <- tmin + abs(rnorm(365, mean = 8, sd = 3))
  expect_gte(ck_txx(tmax, dates)$value, ck_tnx(tmin, dates)$value)
  expect_gte(ck_tnx(tmin, dates)$value, ck_tnn(tmin, dates)$value)
  expect_gte(ck_txx(tmax, dates)$value, ck_txn(tmax, dates)$value)
})

# ETCCDI percentile-based indices (TX10p / TN10p / TX90p / TN90p) ------------

test_that("ck_tx10p returns approx 10% on uniform synthetic ref data", {
  set.seed(1)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  result <- ck_tx10p(tmax, dates, ref_start = 1961L, ref_end = 1962L)
  expect_s3_class(result, "data.frame")
  expect_equal(result$index[1], "tx10p")
  expect_equal(result$unit[1], "%")
  # Within the base period itself the long-run average should be near 10%.
  expect_true(all(result$value >= 0 & result$value <= 100))
  expect_true(abs(mean(result$value) - 10) < 5)
})

test_that("ck_tx90p returns approx 10% on uniform synthetic ref data", {
  set.seed(2)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  result <- ck_tx90p(tmax, dates, ref_start = 1961L, ref_end = 1962L)
  expect_true(abs(mean(result$value) - 10) < 5)
  expect_equal(result$index[1], "tx90p")
})

test_that("ck_tn10p and ck_tn90p run on Tmin and dispatch on percentile", {
  set.seed(3)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  r10 <- ck_tn10p(tmin, dates, ref_start = 1961L, ref_end = 1962L)
  r90 <- ck_tn90p(tmin, dates, ref_start = 1961L, ref_end = 1962L)
  expect_equal(r10$index[1], "tn10p")
  expect_equal(r90$index[1], "tn90p")
  expect_true(abs(mean(r10$value) - 10) < 5)
  expect_true(abs(mean(r90$value) - 10) < 5)
})

test_that("percentile functions error if no data in reference period", {
  dates <- seq(as.Date("2000-01-01"), as.Date("2000-12-31"), by = "day")
  tmax <- rnorm(length(dates), 15, 5)
  expect_error(
    ck_tx10p(tmax, dates, ref_start = 1961L, ref_end = 1990L),
    "reference period"
  )
})

test_that("percentile functions reject mismatched lengths", {
  expect_error(ck_tx10p(c(1, 2), as.Date("1961-01-01")), "same length")
  expect_error(ck_tn90p(c(1, 2), as.Date("1961-01-01")), "same length")
})

# Zhang 2005 in-base bootstrap (Phase D) -------------------------------------

test_that("ck_tx10p bootstrap=FALSE matches the unmodified threshold path", {
  set.seed(81)
  dates <- seq(as.Date("1961-01-01"), as.Date("1965-12-31"), by = "day")
  tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  result <- ck_tx10p(tmax, dates, ref_start = 1961L, ref_end = 1965L,
                     bootstrap = FALSE)
  expect_s3_class(result, "data.frame")
  expect_true(all(result$value >= 0 & result$value <= 100))
})

test_that("bootstrap correction nudges in-base years toward expected 10%", {
  # Ten-year iid sample: in-sample bias suppresses exceedance under the
  # standard threshold; bootstrap should move the mean closer to 10%.
  set.seed(82)
  dates <- seq(as.Date("1961-01-01"), as.Date("1970-12-31"), by = "day")
  tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  no_bs <- ck_tx10p(tmax, dates, ref_start = 1961L, ref_end = 1970L)
  bs    <- ck_tx10p(tmax, dates, ref_start = 1961L, ref_end = 1970L,
                    bootstrap = TRUE)
  expect_lt(abs(mean(bs$value) - 10), abs(mean(no_bs$value) - 10) + 1e-6)
})

test_that("bootstrap exceedance values stay within the unit interval after %", {
  set.seed(83)
  dates <- seq(as.Date("1961-01-01"), as.Date("1963-12-31"), by = "day")
  tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  bs <- ck_tn90p(tmin, dates, ref_start = 1961L, ref_end = 1963L,
                 bootstrap = TRUE)
  expect_true(all(bs$value >= 0 & bs$value <= 100))
})

test_that("bootstrap argument validation", {
  set.seed(84)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmax <- rnorm(length(dates), 15, 5)
  expect_error(
    ck_tx10p(tmax, dates, ref_start = 1961L, ref_end = 1962L,
             bootstrap = "yes"),
    "logical"
  )
  expect_error(
    ck_tx90p(tmax, dates, ref_start = 1961L, ref_end = 1962L,
             bootstrap = c(TRUE, FALSE)),
    "logical"
  )
})

# Calendar-day-base spell duration indices (CSDI / WSDI) ---------------------

test_that("ck_wsdi returns 0 when no 6-day exceedance spell", {
  set.seed(21)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  result <- ck_wsdi(tmax, dates, ref_start = 1961L, ref_end = 1962L)
  expect_s3_class(result, "data.frame")
  expect_equal(result$index[1], "wsdi")
  expect_equal(result$unit[1], "days")
  expect_true(all(result$value >= 0))
})

test_that("ck_wsdi captures injected long warm spell", {
  set.seed(22)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  # Inject 10-day hot spell in 1962 (outside ref period 1961 only).
  hot <- which(dates >= as.Date("1962-07-15") & dates <= as.Date("1962-07-24"))
  tmax[hot] <- tmax[hot] + 20
  result <- ck_wsdi(tmax, dates, ref_start = 1961L, ref_end = 1961L)
  yr_1962 <- result$value[format(result$period, "%Y") == "1962"]
  expect_gte(yr_1962, 10)
})

test_that("ck_csdi returns 0 when no 6-day under-threshold spell", {
  set.seed(23)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  result <- ck_csdi(tmin, dates, ref_start = 1961L, ref_end = 1962L)
  expect_equal(result$index[1], "csdi")
  expect_equal(result$unit[1], "days")
  expect_true(all(result$value >= 0))
})

test_that("ck_csdi captures injected long cold spell", {
  set.seed(24)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  # Inject 10-day cold spell in 1962 (outside ref period 1961 only).
  cold <- which(dates >= as.Date("1962-02-01") & dates <= as.Date("1962-02-10"))
  tmin[cold] <- tmin[cold] - 20
  result <- ck_csdi(tmin, dates, ref_start = 1961L, ref_end = 1961L)
  yr_1962 <- result$value[format(result$period, "%Y") == "1962"]
  expect_gte(yr_1962, 10)
})

test_that("ck_wsdi / ck_csdi reject mismatched lengths and missing reference", {
  expect_error(ck_wsdi(c(1, 2), as.Date("1961-01-01")), "same length")
  expect_error(ck_csdi(c(1, 2), as.Date("1961-01-01")), "same length")
  dates <- seq(as.Date("2000-01-01"), as.Date("2000-12-31"), by = "day")
  expect_error(
    ck_wsdi(rnorm(length(dates), 15, 5), dates, ref_start = 1961L, ref_end = 1990L),
    "reference period"
  )
})

test_that("ck_frost_days handles single observation", {
  result <- ck_frost_days(c(-1), as.Date("2024-01-01"))
  expect_equal(result$value, 1)
})

test_that("ck_frost_days with single non-frost day", {
  result <- ck_frost_days(c(5), as.Date("2024-01-01"))
  expect_equal(result$value, 0)
})

test_that("ck_frost_days handles all-NA input", {
  dates <- as.Date("2024-01-01") + 0:4
  result <- ck_frost_days(c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_), dates)
  expect_s3_class(result, "data.frame")
})

test_that("ck_frost_days handles cross-year boundary", {
  dates <- as.Date("2024-12-30") + 0:3
  tmin <- c(-5, -3, -2, -1)
  result <- ck_frost_days(tmin, dates, period = "annual")
  # 2024 has 2 days, 2025 has 2 days
  expect_equal(nrow(result), 2)
  expect_equal(result$value, c(2, 2))
})

test_that("ck_frost_days exactly at 0C boundary", {
  dates <- as.Date("2024-01-01") + 0:2
  tmin <- c(-0.001, 0, 0.001)
  result <- ck_frost_days(tmin, dates)
  # Only -0.001 is < 0
  expect_equal(result$value, 1)
})

test_that("temperature convert round-trip", {
  expect_equal(ck_convert_temp(ck_convert_temp(42, "C", "F"), "F", "C"),
               42, tolerance = 1e-10)
})

test_that("ck_dry_days with empty-like input (all wet)", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(5, 10, 15, 20, 25)
  result <- ck_dry_days(precip, dates)
  expect_equal(result$value, 0)
})

test_that("ck_total_precip handles all-zero", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- rep(0, 5)
  result <- ck_total_precip(precip, dates)
  expect_equal(result$value, 0)
})

test_that("ck_growing_season handles short data", {
  dates <- as.Date("2024-01-01") + 0:4
  tavg <- rep(20, 5)
  result <- ck_growing_season(tavg, dates)
  expect_true(is.na(result$value))  # < 6 days
})

test_that("ck_heating_degree_days single day", {
  result <- ck_heating_degree_days(c(10), as.Date("2024-01-01"), base = 18)
  expect_equal(result$value, 8)
})

test_that("ck_wind_chill handles NA in input", {
  result <- ck_wind_chill(tavg = c(-5, NA), wind_speed = c(20, 20))
  expect_true(is.na(result$value[2]))
})

test_that("ck_heat_index handles NA in input", {
  result <- ck_heat_index(tavg = c(35, NA), humidity = c(60, 60))
  expect_true(is.na(result$value[2]))
})

# Input validation and period aggregation helpers

#' Validate numeric vector input
#' @noRd
validate_numeric <- function(x, name) {

if (!is.numeric(x)) {
    cli::cli_abort("{.arg {name}} must be a numeric vector.")
  }
  invisible(x)
}

#' Validate dates vector
#' @noRd
validate_dates <- function(dates, n) {
  if (!inherits(dates, "Date")) {
    cli::cli_abort("{.arg dates} must be a {.cls Date} vector.")
  }
  if (length(dates) != n) {
    cli::cli_abort(
      "{.arg dates} must have the same length as the data vector ({n}), not {length(dates)}."
    )
  }
  invisible(dates)
}

#' Validate and parse period argument
#' @noRd
validate_period <- function(period) {
  period <- match.arg(period, c("annual", "monthly"))
  period
}

#' Extract period labels from dates
#' @noRd
get_periods <- function(dates, period) {
  if (period == "annual") {
    as.integer(format(dates, "%Y"))
  } else {
    format(dates, "%Y-%m")
  }
}

#' Convert period labels to Date objects
#' @noRd
period_to_date <- function(period_labels, period) {
  if (period == "annual") {
    as.Date(paste0(period_labels, "-01-01"))
  } else {
    as.Date(paste0(period_labels, "-01"))
  }
}

#' Build output data.frame
#' @noRd
build_result <- function(period_labels, values, index_name, unit, period) {
  data.frame(
    period = period_to_date(period_labels, period),
    value = values,
    index = index_name,
    unit = unit,
    stringsAsFactors = FALSE
  )
}

#' Aggregate by period applying a function
#' @noRd
aggregate_by_period <- function(x, dates, period, fun, na.rm = TRUE) {
  periods <- get_periods(dates, period)
  unique_periods <- unique(periods)
  values <- vapply(unique_periods, function(p) {
    fun(x[periods == p], na.rm = na.rm)
  }, numeric(1))
  list(periods = unique_periods, values = values)
}

#' Count by period where condition is TRUE
#' @noRd
count_by_period <- function(condition, dates, period) {
  periods <- get_periods(dates, period)
  unique_periods <- unique(periods)
  values <- vapply(unique_periods, function(p) {
    sum(condition[periods == p], na.rm = TRUE)
  }, numeric(1))
  list(periods = unique_periods, values = values)
}

#' Calendar-day percentile thresholds for a daily series
#'
#' Returns a length-366 vector of thresholds, one per day of year, computed
#' from the supplied reference window pooled across the reference period.
#' Used by the ETCCDI percentile indices (TX10p, TN10p, TX90p, TN90p, CSDI,
#' WSDI). Follows the ETCCDI +/-2-day window convention (a 5-day window
#' centred on each calendar day, wrapped at year boundaries). Does not
#' implement the Zhang et al. (2005) in-base bootstrap; thresholds are
#' applied directly, so values inside the reference period have a small
#' self-inclusion bias.
#'
#' @noRd
.calendar_day_percentile <- function(values, dates, percentile,
                                     ref_start = 1961L, ref_end = 1990L,
                                     window = 5L) {
  if (!is.numeric(percentile) || length(percentile) != 1L ||
      percentile <= 0 || percentile >= 1) {
    cli::cli_abort("{.arg percentile} must be a single number in (0, 1).")
  }
  if (window < 1L || window %% 2L != 1L) {
    cli::cli_abort("{.arg window} must be a positive odd integer.")
  }

  years <- as.integer(format(dates, "%Y"))
  doy <- as.integer(format(dates, "%j"))

  in_ref <- years >= ref_start & years <= ref_end
  if (!any(in_ref)) {
    cli::cli_abort(
      "No data in reference period {ref_start}-{ref_end}; supply data covering the period or pass {.arg ref_start} / {.arg ref_end}."
    )
  }

  ref_values <- values[in_ref]
  ref_doy <- doy[in_ref]

  half <- (window - 1L) %/% 2L
  thresholds <- rep(NA_real_, 366L)

  for (d in seq_len(366L)) {
    target_doys <- ((d - half - 1L):(d + half - 1L)) %% 366L + 1L
    pool <- ref_values[ref_doy %in% target_doys]
    if (length(pool) > 0L && !all(is.na(pool))) {
      thresholds[d] <- stats::quantile(pool, percentile, na.rm = TRUE,
                                       names = FALSE, type = 8L)
    }
  }
  thresholds
}

#' Mark days that fall inside spans of at least `min_spell` consecutive TRUE
#' values in a logical vector. Used by spell-duration indices (CSDI, WSDI).
#' @noRd
.find_spells <- function(condition, min_spell = 6L) {
  n <- length(condition)
  in_spell <- logical(n)
  i <- 1L
  while (i <= n) {
    if (!is.na(condition[i]) && condition[i]) {
      start <- i
      while (i <= n && !is.na(condition[i]) && condition[i]) {
        i <- i + 1L
      }
      if (i - start >= min_spell) {
        in_spell[start:(i - 1L)] <- TRUE
      }
    } else {
      i <- i + 1L
    }
  }
  in_spell
}

#' Huglin Heliothermal Index
#'
#' The Huglin index is used in viticulture to characterise the thermal
#' potential of a region for grape growing. It is computed over the growing
#' season (April 1 to September 30 in the Northern Hemisphere; October 1 to
#' March 31 in the Southern Hemisphere).
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmin`.
#' @param lat Numeric. Latitude in decimal degrees (used to determine
#'   hemisphere and day-length coefficient).
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @references Huglin, P. (1978). Nouveau mode d'evaluation des possibilites
#'   heliothermiques d'un milieu viticole. *Comptes Rendus de l'Academie
#'   d'Agriculture de France*, 64, 1117-1126.
#'
#' @export
#' @examples
#' dates <- seq(as.Date("2024-04-01"), as.Date("2024-09-30"), by = "day")
#' set.seed(42)
#' tmin <- rnorm(length(dates), mean = 12, sd = 3)
#' tmax <- tmin + runif(length(dates), 8, 15)
#' ck_huglin(tmin, tmax, dates, lat = 45)
ck_huglin <- function(tmin, tmax, dates, lat) {
  validate_numeric(tmin, "tmin")
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmin))
  if (length(tmin) != length(tmax)) {
    cli::cli_abort("{.arg tmin} and {.arg tmax} must have the same length.")
  }
  if (!is.numeric(lat) || length(lat) != 1) {
    cli::cli_abort("{.arg lat} must be a single numeric value.")
  }

  # Day-length coefficient (k) based on latitude (Huglin 1978)
  abs_lat <- abs(lat)
  k <- if (abs_lat <= 40) {
    1.02
  } else if (abs_lat >= 50) {
    1.06
  } else {
    1.02 + (abs_lat - 40) * 0.004
  }

  years <- as.integer(format(dates, "%Y"))
  unique_years <- unique(years)
  months <- as.integer(format(dates, "%m"))

  values <- vapply(unique_years, function(yr) {
    # Growing season months
    if (lat >= 0) {
      gs_mask <- years == yr & months >= 4 & months <= 9
    } else {
      # Southern hemisphere: Oct-Mar spans two calendar years
      gs_mask <- (years == yr & months >= 10) | (years == (yr + 1L) & months <= 3)
    }

    if (!any(gs_mask)) return(NA_real_)

    tavg_gs <- (tmin[gs_mask] + tmax[gs_mask]) / 2
    tmax_gs <- tmax[gs_mask]

    daily_hi <- pmax((tavg_gs - 10) + (tmax_gs - 10), 0) / 2 * k
    sum(daily_hi, na.rm = TRUE)
  }, numeric(1))

  build_result(as.character(unique_years), values, "huglin", "degree-days", "annual")
}

#' Winkler Index
#'
#' The Winkler index (also called growing degree days for viticulture)
#' accumulates daily mean temperature above 10 degrees C during the growing season
#' (April-October in NH, October-April in SH).
#'
#' @param tavg Numeric vector of daily average temperatures (degrees C).
#' @param dates Date vector of the same length as `tavg`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @references Amerine, M. A., & Winkler, A. J. (1944). Composition and
#'   quality of musts and wines of California grapes.
#'
#' @export
#' @examples
#' dates <- seq(as.Date("2024-04-01"), as.Date("2024-10-31"), by = "day")
#' set.seed(42)
#' tavg <- rnorm(length(dates), mean = 18, sd = 4)
#' ck_winkler(tavg, dates)
ck_winkler <- function(tavg, dates) {
  validate_numeric(tavg, "tavg")
  validate_dates(dates, length(tavg))

  years <- as.integer(format(dates, "%Y"))
  months <- as.integer(format(dates, "%m"))
  unique_years <- unique(years)

  values <- vapply(unique_years, function(yr) {
    gs_mask <- years == yr & months >= 4 & months <= 10
    if (!any(gs_mask)) return(NA_real_)
    sum(pmax(tavg[gs_mask] - 10, 0), na.rm = TRUE)
  }, numeric(1))

  build_result(as.character(unique_years), values, "winkler", "degree-days", "annual")
}

#' Branas Hydrothermal Index
#'
#' The Branas index combines temperature and precipitation during the
#' growing season to estimate disease pressure (especially downy mildew)
#' in vineyards. It is the sum of (monthly mean temperature) times
#' (monthly precipitation total) over the five months of the growing
#' season: April-August in the Northern Hemisphere; October-February
#' in the Southern Hemisphere. The Southern Hemisphere season spans
#' two calendar years and is reported under the year in which it
#' starts.
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param tavg Numeric vector of daily mean temperatures (degrees C).
#' @param dates Date vector of the same length as `precip`.
#' @param lat Numeric. Latitude in decimal degrees, used to select the
#'   hemisphere convention. Default 50 (Northern Hemisphere).
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @references Branas, J., Bernon, G., & Levadoux, L. (1946).
#'   Elements de viticulture generale.
#'
#' @export
#' @examples
#' dates <- seq(as.Date("2024-04-01"), as.Date("2024-08-31"), by = "day")
#' set.seed(42)
#' tavg <- rnorm(length(dates), mean = 12, sd = 3)
#' precip <- rgamma(length(dates), shape = 0.5, rate = 0.2)
#' ck_branas(precip, tavg, dates)
ck_branas <- function(precip, tavg, dates, lat = 50) {
  validate_numeric(precip, "precip")
  validate_numeric(tavg, "tavg")
  validate_dates(dates, length(precip))
  if (length(precip) != length(tavg)) {
    cli::cli_abort("{.arg precip} and {.arg tavg} must have the same length.")
  }
  if (!is.numeric(lat) || length(lat) != 1L) {
    cli::cli_abort("{.arg lat} must be a single numeric value.")
  }

  years <- as.integer(format(dates, "%Y"))
  month_labels <- format(dates, "%Y-%m")
  unique_years <- unique(years)

  # Per-year list of "year-month" labels covering the five-month
  # growing season for each hemisphere.
  season_months <- function(yr) {
    if (lat >= 0) {
      # NH: April through August of the same calendar year.
      paste0(yr, "-", sprintf("%02d", 4:8))
    } else {
      # SH: October-December of yr, January-February of yr + 1.
      c(paste0(yr,    "-", sprintf("%02d", 10:12)),
        paste0(yr + 1L, "-", sprintf("%02d",  1:2)))
    }
  }

  values <- vapply(unique_years, function(yr) {
    total <- 0
    for (m in season_months(yr)) {
      idx <- which(month_labels == m)
      if (length(idx) == 0) next
      monthly_t <- mean(tavg[idx], na.rm = TRUE)
      monthly_p <- sum(precip[idx], na.rm = TRUE)
      total <- total + monthly_t * monthly_p
    }
    total
  }, numeric(1))

  build_result(as.character(unique_years), values, "branas",
               "mm\u00b7\u00b0C", "annual")
}

#' First Frost Date
#'
#' Day of year of the first autumn frost (Tmin < 0 degrees C) in each
#' year. Hemisphere is selected by `lat`: in the Northern Hemisphere
#' (`lat >= 0`) the search starts at July 1 (DOY 183); in the Southern
#' Hemisphere it starts at March 1 (DOY 60), matching the autumn entry
#' for each.
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmin`.
#' @param lat Numeric. Latitude in decimal degrees, used to select the
#'   hemisphere convention. Default 50 (Northern Hemisphere).
#'
#' @return A data frame with columns `period`, `value` (day of year),
#'   `date` (the frost date), `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- seq(as.Date("2024-07-01"), as.Date("2024-12-31"), by = "day")
#' set.seed(42)
#' tmin <- 15 - seq_along(dates) * 0.15 + rnorm(length(dates), sd = 3)
#' ck_first_frost(tmin, dates)
ck_first_frost <- function(tmin, dates, lat = 50) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  if (!is.numeric(lat) || length(lat) != 1L) {
    cli::cli_abort("{.arg lat} must be a single numeric value.")
  }

  years <- as.integer(format(dates, "%Y"))
  doy <- as.integer(format(dates, "%j"))
  unique_years <- unique(years)

  # NH autumn: search after July 1 (DOY 183).
  # SH autumn: search after March 1 (DOY 60).
  cutoff <- if (lat >= 0) 183L else 60L

  rows <- lapply(unique_years, function(yr) {
    mask <- years == yr & doy >= cutoff & !is.na(tmin) & tmin < 0
    if (!any(mask)) {
      data.frame(period = as.Date(paste0(yr, "-01-01")),
                 value = NA_real_, date = as.Date(NA),
                 index = "first_frost", unit = "day of year",
                 stringsAsFactors = FALSE)
    } else {
      first_idx <- which(mask)[1]
      data.frame(period = as.Date(paste0(yr, "-01-01")),
                 value = as.double(doy[first_idx]),
                 date = dates[first_idx],
                 index = "first_frost", unit = "day of year",
                 stringsAsFactors = FALSE)
    }
  })
  do.call(rbind, rows)
}

#' Last Frost Date
#'
#' Day of year of the last spring frost (Tmin < 0 degrees C) in each
#' year. Hemisphere is selected by `lat`: in the Northern Hemisphere
#' the search runs up to July 1 (DOY 183); in the Southern Hemisphere
#' up to October 1 (DOY 274), matching the spring boundary for each.
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmin`.
#' @param lat Numeric. Latitude in decimal degrees, used to select the
#'   hemisphere convention. Default 50 (Northern Hemisphere).
#'
#' @return A data frame with columns `period`, `value` (day of year),
#'   `date` (the frost date), `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- seq(as.Date("2024-01-01"), as.Date("2024-06-30"), by = "day")
#' set.seed(42)
#' tmin <- -10 + seq_along(dates) * 0.12 + rnorm(length(dates), sd = 3)
#' ck_last_frost(tmin, dates)
ck_last_frost <- function(tmin, dates, lat = 50) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  if (!is.numeric(lat) || length(lat) != 1L) {
    cli::cli_abort("{.arg lat} must be a single numeric value.")
  }

  years <- as.integer(format(dates, "%Y"))
  doy <- as.integer(format(dates, "%j"))
  unique_years <- unique(years)

  # NH spring: search up to July 1 (DOY 183).
  # SH spring: search up to October 1 (DOY 274).
  cutoff <- if (lat >= 0) 183L else 274L

  rows <- lapply(unique_years, function(yr) {
    mask <- years == yr & doy < cutoff & !is.na(tmin) & tmin < 0
    if (!any(mask)) {
      data.frame(period = as.Date(paste0(yr, "-01-01")),
                 value = NA_real_, date = as.Date(NA),
                 index = "last_frost", unit = "day of year",
                 stringsAsFactors = FALSE)
    } else {
      last_idx <- max(which(mask))
      data.frame(period = as.Date(paste0(yr, "-01-01")),
                 value = as.double(doy[last_idx]),
                 date = dates[last_idx],
                 index = "last_frost", unit = "day of year",
                 stringsAsFactors = FALSE)
    }
  })
  do.call(rbind, rows)
}

test_that("ck_huglin returns annual data frame", {
  dates <- seq(as.Date("2024-04-01"), as.Date("2024-09-30"), by = "day")
  set.seed(42)
  tmin <- rnorm(length(dates), mean = 12, sd = 3)
  tmax <- tmin + runif(length(dates), 8, 15)
  result <- ck_huglin(tmin, tmax, dates, lat = 45)
  expect_s3_class(result, "data.frame")
  expect_equal(result$index, "huglin")
  expect_true(result$value > 0)
})

test_that("ck_huglin rejects mismatched lengths", {
  dates <- as.Date("2024-04-01") + 0:4
  expect_error(ck_huglin(1:5, 1:3, dates, lat = 45), "same length")
})

test_that("ck_huglin rejects non-numeric lat", {
  dates <- as.Date("2024-04-01") + 0:4
  expect_error(ck_huglin(1:5, 6:10, dates, lat = "a"), "numeric")
})

test_that("ck_winkler returns annual GDD", {
  dates <- seq(as.Date("2024-04-01"), as.Date("2024-10-31"), by = "day")
  set.seed(42)
  tavg <- rnorm(length(dates), mean = 18, sd = 4)
  result <- ck_winkler(tavg, dates)
  expect_s3_class(result, "data.frame")
  expect_equal(result$index, "winkler")
  expect_true(result$value > 0)
})

test_that("ck_winkler ignores temps below 10", {
  dates <- as.Date("2024-04-01") + 0:4
  tavg <- c(5, 8, 15, 20, 9)
  result <- ck_winkler(tavg, dates)
  # Only (15-10) + (20-10) = 15
  expect_equal(result$value, 15)
})

test_that("ck_branas computes hydrothermal index", {
  dates <- seq(as.Date("2024-04-01"), as.Date("2024-08-31"), by = "day")
  set.seed(42)
  tavg <- rnorm(length(dates), mean = 12, sd = 3)
  precip <- rgamma(length(dates), shape = 0.5, rate = 0.2)
  result <- ck_branas(precip, tavg, dates)
  expect_s3_class(result, "data.frame")
  expect_equal(result$index, "branas")
})

test_that("ck_branas rejects mismatched lengths", {
  dates <- as.Date("2024-04-01") + 0:4
  expect_error(ck_branas(1:5, 1:3, dates), "same length")
})

test_that("ck_first_frost finds first frost after July 1", {
  dates <- seq(as.Date("2024-07-01"), as.Date("2024-12-31"), by = "day")
  tmin <- rep(10, length(dates))
  # Set frost on Oct 15 (day index for Oct 15 from Jul 1 = 106)
  frost_date <- as.Date("2024-10-15")
  tmin[dates == frost_date] <- -2
  result <- ck_first_frost(tmin, dates)
  expect_equal(result$date, frost_date)
  expect_equal(result$value, as.integer(format(frost_date, "%j")))
})

test_that("ck_first_frost returns NA when no frost", {
  dates <- seq(as.Date("2024-07-01"), as.Date("2024-12-31"), by = "day")
  tmin <- rep(10, length(dates))
  result <- ck_first_frost(tmin, dates)
  expect_true(is.na(result$value))
})

test_that("ck_last_frost finds last frost before July 1", {
  dates <- seq(as.Date("2024-01-01"), as.Date("2024-06-30"), by = "day")
  tmin <- rep(10, length(dates))
  frost_date <- as.Date("2024-04-10")
  tmin[dates == frost_date] <- -1
  result <- ck_last_frost(tmin, dates)
  expect_equal(result$date, frost_date)
})

test_that("ck_last_frost returns NA when no frost", {
  dates <- seq(as.Date("2024-01-01"), as.Date("2024-06-30"), by = "day")
  tmin <- rep(10, length(dates))
  result <- ck_last_frost(tmin, dates)
  expect_true(is.na(result$value))
})

# --- Reference value tests ---

test_that("Huglin K coefficient at specific latitudes", {
  # K(40) = 1.02, K(45) = 1.04, K(50) = 1.06 (Huglin 1978)
  dates <- seq(as.Date("2024-04-01"), as.Date("2024-09-30"), by = "day")
  tmin <- rep(15, length(dates))
  tmax <- rep(25, length(dates))

  # At lat=40 vs lat=50, Huglin should differ by the K ratio
  r40 <- ck_huglin(tmin, tmax, dates, lat = 40)
  r45 <- ck_huglin(tmin, tmax, dates, lat = 45)
  r50 <- ck_huglin(tmin, tmax, dates, lat = 50)

  # Ratio should reflect K values: 1.04/1.02 and 1.06/1.02
  expect_equal(r45$value / r40$value, 1.04 / 1.02, tolerance = 0.001)
  expect_equal(r50$value / r40$value, 1.06 / 1.02, tolerance = 0.001)
})

test_that("Winkler with all temps exactly 10C returns 0", {
  dates <- seq(as.Date("2024-04-01"), as.Date("2024-10-31"), by = "day")
  tavg <- rep(10, length(dates))
  result <- ck_winkler(tavg, dates)
  expect_equal(result$value, 0)
})

test_that("Branas with known monthly means and precip", {
  # April only: mean temp = 15, total precip = 100
  # HI = 15 * 100 = 1500
  dates <- seq(as.Date("2024-04-01"), as.Date("2024-04-30"), by = "day")
  tavg <- rep(15, 30)
  precip <- rep(100 / 30, 30)  # totals to 100mm
  result <- ck_branas(precip, tavg, dates)
  expect_equal(result$value, 15 * 100, tolerance = 1)
})

# Hemisphere fixes -----------------------------------------------------------

test_that("ck_first_frost detects SH autumn frost (lat < 0)", {
  dates <- seq(as.Date("2024-01-01"), as.Date("2024-12-31"), by = "day")
  tmin <- rep(10, length(dates))
  # Inject SH autumn frost (DOY 140-142, May)
  tmin[140:142] <- -1
  expect_true(is.na(ck_first_frost(tmin, dates, lat = 50)$value))
  expect_equal(ck_first_frost(tmin, dates, lat = -33)$value, 140)
})

test_that("ck_last_frost detects SH spring frost (lat < 0)", {
  dates <- seq(as.Date("2024-01-01"), as.Date("2024-12-31"), by = "day")
  tmin <- rep(10, length(dates))
  # Inject SH spring frost (DOY 240-242, late August)
  tmin[240:242] <- -1
  expect_true(is.na(ck_last_frost(tmin, dates, lat = 50)$value))
  expect_equal(ck_last_frost(tmin, dates, lat = -33)$value, 242)
})

test_that("ck_first_frost / ck_last_frost reject non-numeric lat", {
  dates <- as.Date("2024-01-01") + 0:9
  tmin <- rep(0, 10)
  expect_error(ck_first_frost(tmin, dates, lat = "a"), "numeric")
  expect_error(ck_last_frost(tmin, dates, lat = c(1, 2)), "numeric")
})

test_that("ck_branas SH growing season spans Oct-Feb across calendar years", {
  dates <- seq(as.Date("2024-01-01"), as.Date("2025-12-31"), by = "day")
  set.seed(7)
  tavg <- rnorm(length(dates), mean = 12, sd = 3)
  precip <- rgamma(length(dates), shape = 0.5, rate = 0.2)
  nh <- ck_branas(precip, tavg, dates, lat = 50)
  sh <- ck_branas(precip, tavg, dates, lat = -33)
  expect_equal(nrow(nh), 2L)
  expect_equal(nrow(sh), 2L)
  expect_false(identical(nh$value, sh$value))
})

#' Maximum Consecutive Dry Days
#'
#' Maximum number of consecutive days with precipitation below a threshold.
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`.
#' @param threshold Numeric. Dry day threshold in mm (default 1).
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' precip <- c(0, 0, 5, 0, 0, 0, 2, 0, 0, 0)
#' ck_dry_days(precip, dates)
ck_dry_days <- function(precip, dates, threshold = 1, period = "annual") {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  period <- validate_period(period)

  periods <- get_periods(dates, period)
  unique_periods <- unique(periods)

  values <- vapply(unique_periods, function(p) {
    idx <- which(periods == p)
    dry <- precip[idx] < threshold
    max_consecutive(dry)
  }, numeric(1))

  build_result(unique_periods, values, "dry_days", "days", period)
}

#' Maximum Consecutive Wet Days
#'
#' Maximum number of consecutive days with precipitation at or above a
#' threshold.
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`.
#' @param threshold Numeric. Wet day threshold in mm (default 1).
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' precip <- c(5, 3, 0, 2, 8, 1, 0, 0, 4, 6)
#' ck_wet_days(precip, dates)
ck_wet_days <- function(precip, dates, threshold = 1, period = "annual") {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  period <- validate_period(period)

  periods <- get_periods(dates, period)
  unique_periods <- unique(periods)

  values <- vapply(unique_periods, function(p) {
    idx <- which(periods == p)
    wet <- precip[idx] >= threshold
    max_consecutive(wet)
  }, numeric(1))

  build_result(unique_periods, values, "wet_days", "days", period)
}

#' Total Wet-Day Precipitation (PRCPTOT)
#'
#' Annual or monthly total precipitation in wet days, where a wet day is
#' a day with precipitation at or above `wet_day_threshold` mm (default
#' 1 mm, the ETCCDI standard). This is the canonical ETCCDI 'PRCPTOT'
#' definition (Alexander et al. 2006; Zhang et al. 2011).
#'
#' Sub-threshold trace amounts are excluded. Pass
#' `wet_day_threshold = 0` to recover the previous behaviour of
#' summing all daily values.
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#' @param wet_day_threshold Numeric (mm). Days with precipitation
#'   strictly below this threshold are excluded from the sum. Default 1.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' precip <- c(0, 5, 3, 0, 8, 2, 0, 1, 4, 0)
#' ck_total_precip(precip, dates)
ck_total_precip <- function(precip, dates, period = "annual",
                            wet_day_threshold = 1) {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  period <- validate_period(period)
  if (!is.numeric(wet_day_threshold) || length(wet_day_threshold) != 1L ||
      wet_day_threshold < 0) {
    cli::cli_abort("{.arg wet_day_threshold} must be a single non-negative numeric value.")
  }

  # Apply ETCCDI wet-day filter: sum only days with precip >= threshold.
  is_wet <- !is.na(precip) & precip >= wet_day_threshold
  contribution <- ifelse(is_wet, precip, 0)

  result <- aggregate_by_period(contribution, dates, period, sum)
  build_result(result$periods, result$values, "total_precip", "mm", period)
}

#' Heavy Precipitation Days
#'
#' Count of days with precipitation at or above a threshold (default 10 mm).
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`.
#' @param threshold Numeric. Threshold in mm (default 10).
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' precip <- c(0, 5, 12, 0, 15, 2, 0, 11, 4, 0)
#' ck_heavy_precip(precip, dates)
ck_heavy_precip <- function(precip, dates, threshold = 10, period = "annual") {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  period <- validate_period(period)

  result <- count_by_period(precip >= threshold, dates, period)
  build_result(result$periods, result$values, "heavy_precip", "days", period)
}

#' Very Heavy Precipitation Days
#'
#' Count of days with precipitation at or above a threshold (default 20 mm).
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`.
#' @param threshold Numeric. Threshold in mm (default 20).
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' precip <- c(0, 5, 22, 0, 15, 25, 0, 11, 4, 30)
#' ck_very_heavy_precip(precip, dates)
ck_very_heavy_precip <- function(precip, dates, threshold = 20, period = "annual") {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  period <- validate_period(period)

  result <- count_by_period(precip >= threshold, dates, period)
  build_result(result$periods, result$values, "very_heavy_precip", "days", period)
}

#' Maximum 1-Day Precipitation
#'
#' Maximum precipitation recorded in a single day per period.
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' precip <- c(0, 5, 22, 0, 15, 25, 0, 11, 4, 30)
#' ck_max_1day_precip(precip, dates)
ck_max_1day_precip <- function(precip, dates, period = "annual") {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  period <- validate_period(period)

  result <- aggregate_by_period(precip, dates, period, max)
  build_result(result$periods, result$values, "max_1day_precip", "mm", period)
}

#' Maximum 5-Day Precipitation
#'
#' Maximum precipitation total over any 5 consecutive days per period.
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' precip <- c(0, 5, 22, 0, 15, 25, 0, 11, 4, 30)
#' ck_max_5day_precip(precip, dates)
ck_max_5day_precip <- function(precip, dates, period = "annual") {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  period <- validate_period(period)

  periods <- get_periods(dates, period)
  unique_periods <- unique(periods)

  values <- vapply(unique_periods, function(p) {
    idx <- which(periods == p)
    x <- precip[idx]
    n <- length(x)
    if (n < 5) return(sum(x, na.rm = TRUE))
    rolling5 <- vapply(seq_len(n - 4), function(i) {
      sum(x[i:(i + 4)], na.rm = TRUE)
    }, numeric(1))
    max(rolling5, na.rm = TRUE)
  }, numeric(1))

  build_result(unique_periods, values, "max_5day_precip", "mm", period)
}

#' Precipitation Intensity (SDII)
#'
#' Mean precipitation on wet days (days with precipitation >= 1 mm).
#' Also known as the Simple Daily Intensity Index.
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-01-01") + 0:9
#' precip <- c(0, 5, 12, 0, 15, 2, 0, 11, 4, 0)
#' ck_precip_intensity(precip, dates)
ck_precip_intensity <- function(precip, dates, period = "annual") {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  period <- validate_period(period)

  periods <- get_periods(dates, period)
  unique_periods <- unique(periods)

  values <- vapply(unique_periods, function(p) {
    pp <- precip[periods == p]
    pp <- pp[!is.na(pp)]
    if (length(pp) == 0L) return(NA_real_)
    wet <- pp[pp >= 1]
    if (length(wet) == 0L) return(0)
    mean(wet)
  }, numeric(1))

  build_result(unique_periods, values, "precip_intensity", "mm/day", period)
}

#' Very Wet Days Total (R95p)
#'
#' ETCCDI canonical index R95p. Annual total precipitation on days where
#' daily precipitation exceeds the 95th percentile of wet-day precipitation
#' in a reference period (default 1961 to 1990). A wet day is one with
#' precipitation at or above 1 mm. The threshold is a single value derived
#' from all wet days in the reference period (not calendar-day specific).
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`. Must contain
#'   data covering the reference period.
#' @param ref_start,ref_end Integer. Reference period boundary years
#'   (inclusive). Defaults to 1961 and 1990.
#' @param period Character. Aggregation period: `"annual"` (default) or
#'   `"monthly"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' precip <- pmax(rgamma(length(dates), shape = 0.4, scale = 8) - 1, 0)
#' tail(ck_r95p(precip, dates))
ck_r95p <- function(precip, dates, ref_start = 1961L, ref_end = 1990L,
                    period = "annual") {
  .precip_pct_total(precip, dates, 0.95, ref_start, ref_end, period, "r95p")
}

#' Extremely Wet Days Total (R99p)
#'
#' ETCCDI canonical index R99p. Annual total precipitation on days where
#' daily precipitation exceeds the 99th percentile of wet-day precipitation
#' in a reference period (default 1961 to 1990). Same convention as
#' [ck_r95p()].
#'
#' @inheritParams ck_r95p
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' precip <- pmax(rgamma(length(dates), shape = 0.4, scale = 8) - 1, 0)
#' tail(ck_r99p(precip, dates))
ck_r99p <- function(precip, dates, ref_start = 1961L, ref_end = 1990L,
                    period = "annual") {
  .precip_pct_total(precip, dates, 0.99, ref_start, ref_end, period, "r99p")
}

#' Implementation backbone for R95p / R99p
#' @noRd
.precip_pct_total <- function(precip, dates, percentile,
                              ref_start, ref_end, period, index_name) {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  period <- validate_period(period)

  years <- as.integer(format(dates, "%Y"))
  in_ref <- years >= ref_start & years <= ref_end
  if (!any(in_ref)) {
    cli::cli_abort(
      "No data in reference period {ref_start}-{ref_end}; supply data covering the period or pass {.arg ref_start} / {.arg ref_end}."
    )
  }

  ref_wet <- precip[in_ref]
  ref_wet <- ref_wet[!is.na(ref_wet) & ref_wet >= 1]
  if (length(ref_wet) == 0L) {
    cli::cli_abort(
      "No wet days (precip >= 1 mm) in reference period {ref_start}-{ref_end}."
    )
  }
  threshold <- stats::quantile(ref_wet, percentile, names = FALSE,
                               na.rm = TRUE, type = 8L)

  is_wet <- !is.na(precip) & precip >= 1
  exceeds <- is_wet & precip > threshold

  contribution <- ifelse(exceeds, precip, 0)
  result <- aggregate_by_period(contribution, dates, period, sum)
  build_result(result$periods, result$values, index_name, "mm", period)
}

#' Max consecutive TRUE values in a logical vector
#' @noRd
max_consecutive <- function(x) {
  if (length(x) == 0 || all(is.na(x))) return(0)
  x[is.na(x)] <- FALSE
  if (!any(x)) return(0)
  r <- rle(x)
  max(r$lengths[r$values])
}

test_that("ck_dry_days finds max consecutive dry days", {
  dates <- as.Date("2024-01-01") + 0:9
  precip <- c(0, 0, 5, 0, 0, 0, 2, 0, 0, 0)
  result <- ck_dry_days(precip, dates)
  expect_equal(result$value, 3)
  expect_equal(result$index, "dry_days")
})

test_that("ck_wet_days finds max consecutive wet days", {
  dates <- as.Date("2024-01-01") + 0:9
  precip <- c(5, 3, 0, 2, 8, 1, 0, 0, 4, 6)
  result <- ck_wet_days(precip, dates)
  expect_equal(result$value, 3)
})

test_that("ck_dry_days with custom threshold", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(0.5, 0.3, 0.8, 2, 0.1)
  result <- ck_dry_days(precip, dates, threshold = 1)
  expect_equal(result$value, 3)
})

test_that("ck_total_precip sums correctly", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(0, 5, 3, 0, 8)
  result <- ck_total_precip(precip, dates)
  expect_equal(result$value, 16)
  expect_equal(result$unit, "mm")
})

test_that("ck_heavy_precip counts correctly", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(0, 5, 12, 0, 15)
  result <- ck_heavy_precip(precip, dates, threshold = 10)
  expect_equal(result$value, 2)
})

test_that("ck_very_heavy_precip counts correctly", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(0, 5, 22, 0, 25)
  result <- ck_very_heavy_precip(precip, dates, threshold = 20)
  expect_equal(result$value, 2)
})

test_that("ck_max_1day_precip returns max", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(0, 5, 22, 0, 15)
  result <- ck_max_1day_precip(precip, dates)
  expect_equal(result$value, 22)
})

test_that("ck_max_5day_precip returns max rolling sum", {
  dates <- as.Date("2024-01-01") + 0:9
  precip <- c(0, 1, 2, 3, 4, 10, 0, 0, 0, 0)
  result <- ck_max_5day_precip(precip, dates)
  # 1+2+3+4+10 = 20 or 2+3+4+10+0 = 19, max is 20
  expect_equal(result$value, 20)
})

test_that("ck_max_5day_precip with fewer than 5 days", {
  dates <- as.Date("2024-01-01") + 0:2
  precip <- c(5, 3, 8)
  result <- ck_max_5day_precip(precip, dates)
  expect_equal(result$value, 16)
})

test_that("ck_precip_intensity computes mean of wet days", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(0, 5, 0, 15, 0)
  result <- ck_precip_intensity(precip, dates)
  expect_equal(result$value, 10)
  expect_equal(result$unit, "mm/day")
})

test_that("ck_precip_intensity returns 0 when no wet days", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(0, 0.5, 0, 0.3, 0)
  result <- ck_precip_intensity(precip, dates)
  expect_equal(result$value, 0)
})

test_that("monthly aggregation works for precipitation", {
  dates <- c(as.Date("2024-01-01") + 0:4, as.Date("2024-02-01") + 0:4)
  precip <- c(0, 5, 3, 0, 8, 10, 0, 0, 2, 1)
  result <- ck_total_precip(precip, dates, period = "monthly")
  expect_equal(nrow(result), 2)
  expect_equal(result$value, c(16, 13))
})

test_that("max_consecutive handles all-FALSE", {
  expect_equal(climatekit:::max_consecutive(c(FALSE, FALSE)), 0)
})

test_that("max_consecutive handles all-TRUE", {
  expect_equal(climatekit:::max_consecutive(c(TRUE, TRUE, TRUE)), 3)
})

test_that("max_consecutive handles empty", {
  expect_equal(climatekit:::max_consecutive(logical(0)), 0)
})

test_that("max_consecutive handles NA", {
  expect_equal(climatekit:::max_consecutive(c(TRUE, NA, TRUE)), 1)
})

# --- Reference value tests ---

test_that("dry days with all-zero precip returns full length", {
  dates <- as.Date("2024-01-01") + 0:9
  precip <- rep(0, 10)
  result <- ck_dry_days(precip, dates)
  expect_equal(result$value, 10)
})

test_that("wet days with all precip >= 1 returns full length", {
  dates <- as.Date("2024-01-01") + 0:9
  precip <- rep(5, 10)
  result <- ck_wet_days(precip, dates)
  expect_equal(result$value, 10)
})

test_that("max 5-day precip with exactly 5 days", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(1, 2, 3, 4, 5)
  result <- ck_max_5day_precip(precip, dates)
  expect_equal(result$value, 15)
})

test_that("precip = exactly 1.0 mm counts as wet, not dry", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(1.0, 1.0, 1.0, 1.0, 1.0)
  result_wet <- ck_wet_days(precip, dates)
  result_dry <- ck_dry_days(precip, dates)
  expect_equal(result_wet$value, 5)
  expect_equal(result_dry$value, 0)
})

# ETCCDI percentile-based precip indices (R95p / R99p) -----------------------

test_that("ck_r95p sums precip on extreme wet days, returns mm", {
  set.seed(11)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  precip <- pmax(rgamma(length(dates), shape = 0.4, scale = 8) - 1, 0)
  result <- ck_r95p(precip, dates, ref_start = 1961L, ref_end = 1962L)
  expect_s3_class(result, "data.frame")
  expect_equal(result$index[1], "r95p")
  expect_equal(result$unit[1], "mm")
  expect_true(all(result$value >= 0))
  # R95p must be <= total annual precip on wet days
  total <- ck_total_precip(precip, dates)
  expect_true(all(result$value <= total$value + 1e-6))
})

test_that("ck_r99p sums even smaller fraction than ck_r95p", {
  set.seed(12)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  precip <- pmax(rgamma(length(dates), shape = 0.4, scale = 8) - 1, 0)
  r95 <- ck_r95p(precip, dates, ref_start = 1961L, ref_end = 1962L)
  r99 <- ck_r99p(precip, dates, ref_start = 1961L, ref_end = 1962L)
  expect_equal(r99$index[1], "r99p")
  # R99p <= R95p (smaller fraction of days exceed the higher threshold)
  expect_true(all(r99$value <= r95$value + 1e-6))
})

test_that("ck_r95p errors on no reference data", {
  dates <- seq(as.Date("2000-01-01"), as.Date("2000-12-31"), by = "day")
  precip <- rep(0.1, length(dates))
  expect_error(
    ck_r95p(precip, dates, ref_start = 1961L, ref_end = 1990L),
    "reference period"
  )
})

test_that("ck_r95p errors on dry reference period", {
  dates <- seq(as.Date("1961-01-01"), as.Date("1961-12-31"), by = "day")
  precip <- rep(0, length(dates))
  expect_error(
    ck_r95p(precip, dates, ref_start = 1961L, ref_end = 1961L),
    "wet days"
  )
})

# Bug fixes ------------------------------------------------------------------

test_that("ck_total_precip applies ETCCDI wet-day filter (PRCPTOT)", {
  dates <- as.Date("2024-01-01") + 0:5
  # Three trace days (<1mm) and three wet days; ETCCDI excludes the trace days.
  precip <- c(0.5, 0.3, 0.7, 5, 10, 15)
  expect_equal(ck_total_precip(precip, dates)$value, 30)
})

test_that("ck_total_precip wet_day_threshold = 0 recovers raw sum", {
  dates <- as.Date("2024-01-01") + 0:5
  precip <- c(0.5, 0.3, 0.7, 5, 10, 15)
  expect_equal(
    ck_total_precip(precip, dates, wet_day_threshold = 0)$value,
    sum(precip)
  )
})

test_that("ck_total_precip rejects negative threshold", {
  dates <- as.Date("2024-01-01") + 0:2
  expect_error(
    ck_total_precip(c(1, 2, 3), dates, wet_day_threshold = -1),
    "non-negative"
  )
})

test_that("ck_precip_intensity returns NA for all-NA period", {
  dates <- as.Date("2024-01-01") + 0:4
  expect_true(is.na(ck_precip_intensity(rep(NA_real_, 5), dates)$value))
})

test_that("ck_precip_intensity ignores NAs in mixed period", {
  dates <- as.Date("2024-01-01") + 0:4
  precip <- c(NA, 5, NA, 15, NA)
  result <- ck_precip_intensity(precip, dates)
  expect_equal(result$value, 10)
})

# Excess Heat Factor (EHF) following Nairn & Fawcett (2013).
#
# EHF combines a "significance" component (excess of recent 3-day mean
# daily temperature over the long-run 95th percentile from a reference
# period) with an "acclimatisation" component (excess of recent 3-day
# mean over the previous 30-day mean). EHF > 0 indicates heatwave
# conditions; larger values indicate more severe / less-acclimatised
# events.
#
#   DMT(i)        = (TX(i) + TN(i)) / 2
#   T(i, 3)       = mean(DMT(i-2 .. i))
#   T(i, 30)      = mean(DMT(i-32 .. i-3))
#   T95           = 95th percentile of DMT over the reference period
#   EHIsig(i)     = T(i, 3) - T95
#   EHIaccl(i)    = T(i, 3) - T(i, 30)
#   EHF(i)        = max(EHIsig(i), 1) * EHIaccl(i)
#
# Reference: Nairn JR, Fawcett RJB (2013). Defining heatwaves: heatwave
# defined as a heat-impact event servicing all community and business
# sectors in Australia. CAWCR Technical Report No. 060.

#' Compute daily Excess Heat Factor (Nairn & Fawcett 2013)
#'
#' Internal helper that returns the daily EHF time series alongside
#' its `T(i, 3)` and `T95` building blocks. Used by [ck_ehf()] and
#' available for users who need event-level rather than annual output.
#' @noRd
.ck_ehf_daily <- function(tmax, tmin, dates,
                          ref_start = 1961L, ref_end = 1990L) {
  validate_numeric(tmax, "tmax")
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmax))
  if (length(tmax) != length(tmin)) {
    cli::cli_abort("{.arg tmax} and {.arg tmin} must have the same length.")
  }

  dmt <- (tmax + tmin) / 2
  n <- length(dmt)

  # T(i, 3): trailing 3-day mean of DMT (i-2..i).
  t3 <- rep(NA_real_, n)
  for (i in 3:n) {
    win <- dmt[(i - 2L):i]
    if (all(is.na(win))) next
    t3[i] <- mean(win, na.rm = TRUE)
  }

  # T(i, 30): mean of DMT over the 30 days BEFORE the recent 3-day window
  # (i-32..i-3). Requires at least 33 days of history.
  t30 <- rep(NA_real_, n)
  for (i in 33:n) {
    win <- dmt[(i - 32L):(i - 3L)]
    if (all(is.na(win))) next
    t30[i] <- mean(win, na.rm = TRUE)
  }

  # T95: 95th percentile of DMT over the reference period.
  years <- as.integer(format(dates, "%Y"))
  in_ref <- years >= ref_start & years <= ref_end
  if (!any(in_ref)) {
    cli::cli_abort(
      "No data in reference period {ref_start}-{ref_end}; supply data covering the period or pass {.arg ref_start} / {.arg ref_end}."
    )
  }
  t95 <- stats::quantile(dmt[in_ref], 0.95, na.rm = TRUE,
                         names = FALSE, type = 8L)

  ehi_sig  <- t3 - t95
  ehi_accl <- t3 - t30
  ehf <- pmax(ehi_sig, 1) * ehi_accl

  list(dates = dates, dmt = dmt, t3 = t3, t30 = t30, t95 = t95,
       ehi_sig = ehi_sig, ehi_accl = ehi_accl, ehf = ehf)
}

#' Excess Heat Factor (EHF, Nairn & Fawcett 2013)
#'
#' Annual summary of the daily Excess Heat Factor heatwave intensity
#' metric. EHF combines a 3-day mean daily temperature anomaly above
#' the 95th percentile of the reference period with an acclimatisation
#' term (3-day mean minus previous 30-day mean). Positive EHF days
#' indicate heatwave conditions; larger values indicate more severe
#' or less-acclimatised events. This is the operational heatwave
#' metric used by the Australian Bureau of Meteorology.
#'
#' Three annual summaries are exposed via the `stat` argument:
#' \itemize{
#'   \item \code{"max"} (default): peak EHF in the year. Strongest
#'         single-day intensity.
#'   \item \code{"n_positive"}: count of days with `EHF > 0`. A
#'         frequency-of-heatwave-conditions measure.
#'   \item \code{"sum_positive"}: sum of EHF on days with `EHF > 0`.
#'         A severity-weighted total.
#' }
#'
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param dates Date vector of the same length as `tmax`. Must contain
#'   data covering the reference period.
#' @param ref_start,ref_end Integer. Reference period boundary years
#'   (inclusive). Defaults to 1961 and 1990.
#' @param stat One of `"max"`, `"n_positive"`, or `"sum_positive"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @references
#' Nairn, J. R., & Fawcett, R. J. B. (2013). Defining heatwaves:
#' heatwave defined as a heat-impact event servicing all community and
#' business sectors in Australia. *CAWCR Technical Report No. 060*.
#'
#' Perkins, S. E., & Alexander, L. V. (2013). On the measurement of
#' heatwaves. *Journal of Climate*, 26(13), 4500-4517.
#' \doi{10.1175/JCLI-D-12-00383.1}.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' s <- 2 * pi * as.integer(format(dates, "%j")) / 365
#' tmax <- 20 + 10 * sin(s) + rnorm(length(dates))
#' tmin <- 10 +  8 * sin(s) + rnorm(length(dates))
#' tail(ck_ehf(tmax, tmin, dates))
ck_ehf <- function(tmax, tmin, dates,
                   ref_start = 1961L, ref_end = 1990L,
                   stat = c("max", "n_positive", "sum_positive")) {
  stat <- match.arg(stat)

  daily <- .ck_ehf_daily(tmax, tmin, dates, ref_start, ref_end)
  ehf <- daily$ehf

  years <- as.integer(format(dates, "%Y"))
  unique_years <- unique(years)

  values <- vapply(unique_years, function(yr) {
    sel <- years == yr
    e <- ehf[sel]
    e <- e[!is.na(e)]
    if (length(e) == 0L) return(NA_real_)
    switch(stat,
      "max"          = max(e),
      "n_positive"   = as.numeric(sum(e > 0)),
      "sum_positive" = sum(e[e > 0])
    )
  }, numeric(1))

  unit <- switch(stat,
    "max"          = "\u00b0C^2",
    "n_positive"   = "days",
    "sum_positive" = "\u00b0C^2"
  )

  build_result(as.character(unique_years), values,
               paste0("ehf_", stat), unit, "annual")
}

# Tests for ck_ehf (Excess Heat Factor, Nairn & Fawcett 2013).

make_climate_two_years <- function(seed = 71) {
  set.seed(seed)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  s <- 2 * pi * as.integer(format(dates, "%j")) / 365
  tmax <- 20 + 10 * sin(s) + rnorm(length(dates))
  tmin <- 10 +  8 * sin(s) + rnorm(length(dates))
  list(dates = dates, tmax = tmax, tmin = tmin)
}

test_that("ck_ehf returns annual data frame with expected shape", {
  d <- make_climate_two_years()
  result <- ck_ehf(d$tmax, d$tmin, d$dates,
                   ref_start = 1961L, ref_end = 1961L)
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2L)
  expect_equal(result$index[1], "ehf_max")
  expect_true(all(c("period", "value", "index", "unit") %in% names(result)))
})

test_that("ck_ehf 'n_positive' counts heatwave-condition days", {
  d <- make_climate_two_years(seed = 72)
  hot <- which(d$dates >= as.Date("1962-07-15") & d$dates <= as.Date("1962-07-25"))
  d$tmax[hot] <- d$tmax[hot] + 15
  d$tmin[hot] <- d$tmin[hot] + 10
  result <- ck_ehf(d$tmax, d$tmin, d$dates,
                   ref_start = 1961L, ref_end = 1961L,
                   stat = "n_positive")
  yr_1962 <- result$value[format(result$period, "%Y") == "1962"]
  expect_gte(yr_1962, 5)
  expect_equal(result$index[1], "ehf_n_positive")
  expect_equal(result$unit[1], "days")
})

test_that("ck_ehf 'sum_positive' >= max EHF on year of injected heatwave", {
  d <- make_climate_two_years(seed = 73)
  hot <- which(d$dates >= as.Date("1962-07-15") & d$dates <= as.Date("1962-07-25"))
  d$tmax[hot] <- d$tmax[hot] + 15
  d$tmin[hot] <- d$tmin[hot] + 10
  rmax <- ck_ehf(d$tmax, d$tmin, d$dates,
                 ref_start = 1961L, ref_end = 1961L, stat = "max")
  rsum <- ck_ehf(d$tmax, d$tmin, d$dates,
                 ref_start = 1961L, ref_end = 1961L, stat = "sum_positive")
  yr <- format(rmax$period, "%Y") == "1962"
  expect_gte(rsum$value[yr], rmax$value[yr])
})

test_that("ck_ehf rejects bad stat", {
  d <- make_climate_two_years(seed = 74)
  expect_error(
    ck_ehf(d$tmax, d$tmin, d$dates, ref_start = 1961L, ref_end = 1961L,
           stat = "bogus"),
    "should be one of"
  )
})

test_that("ck_ehf rejects mismatched lengths and missing reference data", {
  expect_error(ck_ehf(c(1, 2), c(1, 2, 3), as.Date("1961-01-01") + 0:1),
               "same length")
  dates <- seq(as.Date("2000-01-01"), as.Date("2000-12-31"), by = "day")
  tmax <- rnorm(length(dates), 20, 5)
  tmin <- rnorm(length(dates), 10, 4)
  expect_error(ck_ehf(tmax, tmin, dates, ref_start = 1961L, ref_end = 1990L),
               "reference period")
})

test_that("ck_compute dispatches ehf", {
  d <- make_climate_two_years(seed = 75)
  df <- data.frame(dates = d$dates, tmax = d$tmax, tmin = d$tmin)
  result <- ck_compute(df, "ehf", ref_start = 1961L, ref_end = 1961L)
  expect_s3_class(result, "data.frame")
})

#' Standardized Precipitation Index (SPI)
#'
#' Compute the SPI by fitting a parametric distribution to rolling monthly
#' precipitation accumulations and transforming to standard normal
#' deviates. Two distributions are supported: the two-parameter gamma
#' (default; WMO-1090 standard) and the three-parameter Pearson III. The
#' Pearson III tail is heavier and is preferred in arid regions where the
#' wet-day distribution is highly skewed (Stagge et al. 2015).
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param dates Date vector of the same length as `precip`.
#' @param scale Integer. Accumulation period in months (default 3).
#' @param distribution Character. Either `"gamma"` (default) or
#'   `"pearsonIII"`.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @references
#' McKee, T. B., Doesken, N. J., & Kleist, J. (1993). The relationship of
#' drought frequency and duration to time scales.
#'
#' Stagge, J. H., Tallaksen, L. M., Gudmundsson, L., Van Loon, A. F., &
#' Stahl, K. (2015). Candidate distributions for climatological drought
#' indices (SPI and SPEI). *International Journal of Climatology*, 35(13),
#' 4027-4040. \doi{10.1002/joc.4267}.
#'
#' @export
#' @examples
#' dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
#' set.seed(42)
#' precip <- rgamma(length(dates), shape = 0.5, rate = 0.1)
#' ck_spi(precip, dates, scale = 3)
ck_spi <- function(precip, dates, scale = 3,
                   distribution = c("gamma", "pearsonIII")) {
  validate_numeric(precip, "precip")
  validate_dates(dates, length(precip))
  if (!is.numeric(scale) || length(scale) != 1 || scale < 1) {
    cli::cli_abort("{.arg scale} must be a positive integer.")
  }
  scale <- as.integer(scale)
  distribution <- match.arg(distribution)

  monthly <- .monthly_totals(precip, dates)

  n <- nrow(monthly)
  if (n < scale) {
    cli::cli_abort("Not enough months ({n}) for scale {scale}.")
  }

  accum <- rep(NA_real_, n)
  for (i in scale:n) {
    accum[i] <- sum(monthly$total[(i - scale + 1):i], na.rm = TRUE)
  }

  cal_month <- as.integer(format(monthly$month, "%m"))
  spi_values <- rep(NA_real_, n)
  fit_fun <- switch(distribution,
    "gamma"      = .gamma_to_normal,
    "pearsonIII" = .pearson3_to_normal
  )
  for (m in 1:12) {
    idx <- which(cal_month == m & !is.na(accum))
    if (length(idx) >= 3) {
      spi_values[idx] <- fit_fun(accum[idx])
    }
  }

  result <- data.frame(
    period = monthly$month,
    value = spi_values,
    index = "spi",
    unit = "dimensionless",
    stringsAsFactors = FALSE
  )
  result[!is.na(result$value), ]
}

#' Standardized Precipitation-Evapotranspiration Index (SPEI)
#'
#' Compute the SPEI by fitting a log-logistic distribution to the monthly
#' climatic water balance (precipitation minus potential evapotranspiration)
#' accumulated over a rolling window.
#'
#' @param precip Numeric vector of daily precipitation (mm).
#' @param pet Numeric vector of daily potential evapotranspiration (mm).
#' @param dates Date vector of the same length as `precip` and `pet`.
#' @param scale Integer. Accumulation period in months (default 3).
#' @param distribution Character. Either `"log-logistic"` (default,
#'   Vicente-Serrano et al. 2010) or `"gev"` (Generalised Extreme
#'   Value, fitted by L-moments; preferred for water-balance series
#'   with heavy upper or lower tails).
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @references Vicente-Serrano, S. M., Begueria, S., & Lopez-Moreno, J. I.
#'   (2010). A multiscalar drought index sensitive to global warming: the
#'   Standardized Precipitation Evapotranspiration Index. *Journal of Climate*,
#'   23(7), 1696-1718.
#'
#' @export
#' @examples
#' dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
#' set.seed(42)
#' precip <- rgamma(length(dates), shape = 0.5, rate = 0.1)
#' pet <- rep(3, length(dates))
#' ck_spei(precip, pet, dates, scale = 3)
ck_spei <- function(precip, pet, dates, scale = 3,
                    distribution = c("log-logistic", "gev")) {
  validate_numeric(precip, "precip")
  validate_numeric(pet, "pet")
  validate_dates(dates, length(precip))
  if (length(precip) != length(pet)) {
    cli::cli_abort("{.arg precip} and {.arg pet} must have the same length.")
  }
  if (!is.numeric(scale) || length(scale) != 1 || scale < 1) {
    cli::cli_abort("{.arg scale} must be a positive integer.")
  }
  scale <- as.integer(scale)
  distribution <- match.arg(distribution)

  wb <- precip - pet
  monthly <- .monthly_totals(wb, dates)

  n <- nrow(monthly)
  if (n < scale) {
    cli::cli_abort("Not enough months ({n}) for scale {scale}.")
  }

  accum <- rep(NA_real_, n)
  for (i in scale:n) {
    accum[i] <- sum(monthly$total[(i - scale + 1):i], na.rm = TRUE)
  }

  cal_month <- as.integer(format(monthly$month, "%m"))
  spei_values <- rep(NA_real_, n)
  fit_fun <- switch(distribution,
    "log-logistic" = .loglogistic_to_normal,
    "gev"          = .gev_to_normal
  )
  for (m in 1:12) {
    idx <- which(cal_month == m & !is.na(accum))
    if (length(idx) >= 3) {
      spei_values[idx] <- fit_fun(accum[idx])
    }
  }

  result <- data.frame(
    period = monthly$month,
    value = spei_values,
    index = "spei",
    unit = "dimensionless",
    stringsAsFactors = FALSE
  )
  result[!is.na(result$value), ]
}

#' Potential Evapotranspiration (Hargreaves Method)
#'
#' Estimate daily PET using the Hargreaves-Samani equation, which requires
#' only daily temperature extremes and latitude.
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param tmax Numeric vector of daily maximum temperatures (degrees C).
#' @param lat Numeric. Latitude in decimal degrees.
#' @param dates Date vector of the same length as `tmin`.
#'
#' @return A data frame with columns `date`, `value`, `index`, and `unit`.
#'
#' @references Hargreaves, G. H., & Samani, Z. A. (1985). Reference crop
#'   evapotranspiration from temperature. *Applied Engineering in Agriculture*,
#'   1(2), 96-99.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-07-01") + 0:9
#' tmin <- c(15, 16, 14, 17, 15, 13, 16, 14, 15, 16)
#' tmax <- c(30, 32, 28, 33, 31, 27, 34, 29, 30, 32)
#' ck_pet(tmin, tmax, lat = 45, dates = dates)
ck_pet <- function(tmin, tmax, lat, dates) {
  validate_numeric(tmin, "tmin")
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmin))
  if (length(tmin) != length(tmax)) {
    cli::cli_abort("{.arg tmin} and {.arg tmax} must have the same length.")
  }
  if (!is.numeric(lat) || length(lat) != 1) {
    cli::cli_abort("{.arg lat} must be a single numeric value.")
  }

  tavg <- (tmin + tmax) / 2
  td <- tmax - tmin
  td[td < 0] <- 0

  # Day of year
  doy <- as.integer(format(dates, "%j"))

  # Extraterrestrial radiation (Ra) in mm/day equivalent
  ra <- .extraterrestrial_radiation(lat, doy)

  # Hargreaves equation
  pet <- 0.0023 * ra * sqrt(td) * (tavg + 17.8)
  pet[pet < 0] <- 0

  data.frame(
    date = dates,
    value = pet,
    index = "pet",
    unit = "mm",
    stringsAsFactors = FALSE
  )
}

#' Reference Evapotranspiration (FAO-56 Penman-Monteith)
#'
#' Compute reference evapotranspiration ETo using the FAO-56
#' Penman-Monteith equation (Allen et al. 1998), the international
#' standard for ETo estimation. Required inputs are daily Tmin and
#' Tmax; optional inputs (humidity, wind speed, incoming solar
#' radiation, elevation) increase accuracy. Where humidity, wind, or
#' solar radiation are missing, FAO-56 fallback estimators are used.
#'
#' Inputs and units:
#' \itemize{
#'   \item `tmin`, `tmax`: daily minimum and maximum temperature (degrees C).
#'   \item `lat`: latitude in decimal degrees, used for the
#'         extraterrestrial-radiation calculation.
#'   \item `elev`: elevation above sea level in metres (default 0).
#'   \item `wind`: 2-metre wind speed (m/s). Default 2 m/s (the
#'         FAO-56 fallback value when wind data are unavailable).
#'   \item `rh_min`, `rh_max`: minimum and maximum daily relative
#'         humidity (\%). If both are NULL (default), actual vapour
#'         pressure is estimated as `e0(tmin)` (FAO-56 Eq. 48).
#'   \item `rs`: daily incoming solar radiation (MJ m^-2 day^-1).
#'         If NULL, estimated by Hargreaves-Samani: `Rs = krs * Ra
#'         * sqrt(Tmax - Tmin)` with `krs = 0.16` for inland sites.
#'   \item `albedo`: surface albedo (default 0.23, the FAO grass
#'         reference).
#' }
#'
#' @param tmin Numeric vector of daily minimum temperatures (degrees C).
#' @param tmax Numeric vector of daily maximum temperatures (degrees C),
#'   same length as `tmin`.
#' @param lat Numeric. Latitude in decimal degrees.
#' @param dates Date vector of the same length as `tmin`.
#' @param elev Numeric. Elevation above sea level in metres
#'   (default 0).
#' @param wind Numeric vector or single value. 2-m wind speed (m/s).
#'   Default `2` (FAO-56 fallback for unmeasured wind).
#' @param rh_min,rh_max Optional numeric vectors of daily minimum and
#'   maximum relative humidity in percent. Both must be supplied
#'   together; otherwise vapour pressure falls back to `e0(tmin)`.
#' @param rs Optional numeric vector of daily incoming solar
#'   radiation (MJ m^-2 day^-1). If `NULL`, Hargreaves-Samani
#'   estimate is used.
#' @param albedo Numeric. Surface albedo (default 0.23).
#' @param krs Numeric. Hargreaves-Samani coefficient for the Rs
#'   fallback (default 0.16 for inland sites; 0.19 for coastal).
#'
#' @return A data frame with columns `date`, `value`, `index`
#'   (`"pet_pm"`), and `unit`.
#'
#' @references Allen, R. G., Pereira, L. S., Raes, D., & Smith, M.
#'   (1998). Crop evapotranspiration: guidelines for computing crop
#'   water requirements. *FAO Irrigation and Drainage Paper 56*.
#'
#' @export
#' @examples
#' dates <- as.Date("2024-07-01") + 0:9
#' tmin <- c(15, 16, 14, 17, 15, 13, 16, 14, 15, 16)
#' tmax <- c(30, 32, 28, 33, 31, 27, 34, 29, 30, 32)
#' ck_pet_pm(tmin, tmax, lat = 45, dates = dates)
ck_pet_pm <- function(tmin, tmax, lat, dates,
                      elev = 0, wind = 2,
                      rh_min = NULL, rh_max = NULL, rs = NULL,
                      albedo = 0.23, krs = 0.16) {
  validate_numeric(tmin, "tmin")
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmin))
  if (length(tmin) != length(tmax)) {
    cli::cli_abort("{.arg tmin} and {.arg tmax} must have the same length.")
  }
  if (!is.numeric(lat) || length(lat) != 1L) {
    cli::cli_abort("{.arg lat} must be a single numeric value.")
  }
  if (!is.numeric(elev) || length(elev) != 1L) {
    cli::cli_abort("{.arg elev} must be a single numeric value.")
  }
  if (!is.numeric(albedo) || length(albedo) != 1L ||
      albedo < 0 || albedo > 1) {
    cli::cli_abort("{.arg albedo} must be a single number in [0, 1].")
  }
  if (!is.numeric(krs) || length(krs) != 1L || krs <= 0) {
    cli::cli_abort("{.arg krs} must be a single positive numeric value.")
  }

  n <- length(tmin)

  # Wind: scalar or vector of length n
  if (length(wind) == 1L) wind <- rep(as.numeric(wind), n)
  validate_numeric(wind, "wind")
  if (length(wind) != n) {
    cli::cli_abort("{.arg wind} must be length 1 or length {n}.")
  }

  tmean <- (tmin + tmax) / 2

  # Saturation vapour pressure: e0(T) = 0.6108 * exp(17.27 * T / (T + 237.3)) (kPa)
  e0 <- function(T) 0.6108 * exp(17.27 * T / (T + 237.3))
  e0_tmax <- e0(tmax)
  e0_tmin <- e0(tmin)
  es <- (e0_tmax + e0_tmin) / 2  # FAO-56 Eq. 12

  # Slope of vapour pressure curve at Tmean (kPa / degC, FAO-56 Eq. 13)
  delta <- 4098 * e0(tmean) / (tmean + 237.3)^2

  # Atmospheric pressure (kPa, FAO-56 Eq. 7)
  P <- 101.3 * ((293 - 0.0065 * elev) / 293)^5.26
  # Psychrometric constant (kPa / degC, FAO-56 Eq. 8)
  gamma <- 0.000665 * P

  # Actual vapour pressure
  if (!is.null(rh_min) && !is.null(rh_max)) {
    validate_numeric(rh_min, "rh_min")
    validate_numeric(rh_max, "rh_max")
    if (length(rh_min) != n || length(rh_max) != n) {
      cli::cli_abort("{.arg rh_min} and {.arg rh_max} must be length {n}.")
    }
    ea <- (e0_tmin * rh_max + e0_tmax * rh_min) / 200  # FAO-56 Eq. 17
  } else {
    # FAO-56 Eq. 48 fallback: assume Tdew = Tmin
    ea <- e0_tmin
  }

  # Extraterrestrial radiation (MJ/m^2/day)
  doy <- as.integer(format(dates, "%j"))
  ra_mm <- .extraterrestrial_radiation(lat, doy)  # mm/day
  ra_mj <- ra_mm * 2.45                          # convert to MJ/m^2/day

  # Solar radiation: supplied or Hargreaves-Samani estimate
  if (is.null(rs)) {
    td <- pmax(tmax - tmin, 0)
    rs_calc <- krs * ra_mj * sqrt(td)
  } else {
    validate_numeric(rs, "rs")
    if (length(rs) != n) {
      cli::cli_abort("{.arg rs} must be length {n}.")
    }
    rs_calc <- rs
  }

  # Clear-sky solar radiation Rso (FAO-56 Eq. 37)
  rso <- (0.75 + 2e-5 * elev) * ra_mj
  # Cap rs/rso at 1 to keep the longwave correction term sensible
  ratio <- pmin(rs_calc / pmax(rso, 1e-6), 1)

  # Net shortwave radiation (MJ/m^2/day, FAO-56 Eq. 38)
  rns <- (1 - albedo) * rs_calc

  # Net longwave radiation (MJ/m^2/day, FAO-56 Eq. 39)
  sigma_sb <- 4.903e-9
  rnl <- sigma_sb *
    ((tmax + 273.16)^4 + (tmin + 273.16)^4) / 2 *
    (0.34 - 0.14 * sqrt(pmax(ea, 0))) *
    (1.35 * ratio - 0.35)

  rn <- rns - rnl

  # FAO-56 Penman-Monteith (Eq. 6); G = 0 for daily timestep
  numerator <- 0.408 * delta * rn +
               gamma * (900 / (tmean + 273)) * wind * (es - ea)
  denominator <- delta + gamma * (1 + 0.34 * wind)
  pet <- numerator / denominator
  pet[!is.finite(pet) | pet < 0] <- 0

  data.frame(
    date  = dates,
    value = pet,
    index = "pet_pm",
    unit  = "mm",
    stringsAsFactors = FALSE
  )
}

#' Compute extraterrestrial radiation (Ra) in mm/day equivalent
#' @noRd
.extraterrestrial_radiation <- function(lat, doy) {
  # Solar constant in MJ/m2/min
  gsc <- 0.0820

  # Latitude in radians
  phi <- lat * pi / 180

  # Solar declination
  delta <- 0.409 * sin(2 * pi / 365 * doy - 1.39)

  # Sunset hour angle
  ws <- acos(-tan(phi) * tan(delta))
  # Clamp to valid range

  ws[is.nan(ws)] <- if (lat >= 0) pi else 0

  # Inverse relative distance Earth-Sun

  dr <- 1 + 0.033 * cos(2 * pi / 365 * doy)

  # Ra in MJ/m2/day
  ra_mj <- (24 * 60 / pi) * gsc * dr *
    (ws * sin(phi) * sin(delta) + cos(phi) * cos(delta) * sin(ws))

  # Convert MJ/m2/day to mm/day (divide by latent heat of vaporization 2.45)
  ra_mj / 2.45
}

#' Aggregate daily values to monthly totals
#' @noRd
.monthly_totals <- function(x, dates) {
  month_labels <- format(dates, "%Y-%m")
  unique_months <- unique(month_labels)
  totals <- vapply(unique_months, function(m) {
    sum(x[month_labels == m], na.rm = TRUE)
  }, numeric(1))
  data.frame(
    month = as.Date(paste0(unique_months, "-01")),
    total = totals,
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

#' Fit gamma distribution (MLE) and transform to standard normal
#'
#' Uses Maximum Likelihood Estimation via Thom's (1958) approximation,
#' as specified by the WMO User Guide (WMO-No. 1090) and consistent with
#' the SPEI R package.
#' @noRd
.gamma_to_normal <- function(x) {
  result <- rep(NA_real_, length(x))
  valid <- !is.na(x)
  xv <- x[valid]

  if (length(xv) < 3) return(result)

  # Proportion of zeros
  q <- sum(xv == 0) / length(xv)
  xv_pos <- xv[xv > 0]

  if (length(xv_pos) < 2) return(result)

  # MLE via Thom's (1958) approximation
  xbar <- mean(xv_pos)
  A <- log(xbar) - mean(log(xv_pos))
  if (A <= 0) return(result)

  alpha <- (1 + sqrt(1 + 4 * A / 3)) / (4 * A)
  beta <- xbar / alpha

  # Transform to standard normal
  cdf <- rep(NA_real_, length(xv))
  for (i in seq_along(xv)) {
    if (xv[i] == 0) {
      cdf[i] <- q
    } else {
      cdf[i] <- q + (1 - q) * stats::pgamma(xv[i], shape = alpha, scale = beta)
    }
  }
  # Clamp to avoid Inf
  cdf[cdf <= 0] <- 0.001
  cdf[cdf >= 1] <- 0.999

  result[valid] <- stats::qnorm(cdf)
  result
}

#' Fit Pearson III distribution (method of moments) and transform to standard normal
#'
#' Method-of-moments estimator for the three-parameter Pearson III. Used as
#' an alternative SPI fit for arid-region or heavy-tailed precipitation
#' regimes where the gamma fit is poor (Stagge et al. 2015).
#' @noRd
.pearson3_to_normal <- function(x) {
  result <- rep(NA_real_, length(x))
  valid <- !is.na(x)
  xv <- x[valid]
  n <- length(xv)

  if (n < 3L) return(result)

  mu <- mean(xv)
  sigma <- stats::sd(xv)
  if (!is.finite(sigma) || sigma <= 0) return(result)

  m3 <- mean((xv - mu)^3)
  gamma_skew <- m3 / sigma^3

  if (!is.finite(gamma_skew) || abs(gamma_skew) < 1e-6) {
    # Approximately normal: fall back to standardisation.
    cdf <- stats::pnorm((xv - mu) / sigma)
  } else {
    alpha <- 4 / gamma_skew^2
    beta  <- sigma * abs(gamma_skew) / 2
    xi    <- mu - alpha * sign(gamma_skew) * beta

    if (!is.finite(alpha) || alpha <= 0 || !is.finite(beta) || beta <= 0) {
      return(result)
    }

    if (gamma_skew > 0) {
      cdf <- stats::pgamma((xv - xi) / beta, shape = alpha)
    } else {
      cdf <- 1 - stats::pgamma((xi - xv) / beta, shape = alpha)
    }
  }

  cdf[is.na(cdf)] <- 0.5
  cdf[cdf <= 0] <- 0.001
  cdf[cdf >= 1] <- 0.999

  result[valid] <- stats::qnorm(cdf)
  result
}

#' Fit Generalised Extreme Value distribution (L-moments) and transform to standard normal
#'
#' Hosking (1985, 1990) L-moment estimator. Used as an alternative SPEI
#' fit when the climatic water balance has long-tailed extremes that the
#' log-logistic does not capture.
#' @noRd
.gev_to_normal <- function(x) {
  result <- rep(NA_real_, length(x))
  valid <- !is.na(x)
  xv <- x[valid]
  n <- length(xv)

  if (n < 4L) return(result)

  xs <- sort(xv)
  ranks <- seq_len(n)
  b0 <- mean(xs)
  b1 <- sum(((ranks - 1) / (n - 1)) * xs) / n
  b2 <- sum(((ranks - 1) * (ranks - 2) / ((n - 1) * (n - 2))) * xs) / n

  l1 <- b0
  l2 <- 2 * b1 - b0
  l3 <- 6 * b2 - 6 * b1 + b0

  if (!is.finite(l2) || l2 <= 0) {
    cli::cli_warn("SPEI GEV fit failed: L-moment l2 <= 0. Returning NAs.")
    return(result)
  }
  t3 <- l3 / l2
  if (!is.finite(t3) || abs(t3) >= 1) {
    cli::cli_warn("SPEI GEV fit failed: L-skewness out of range. Returning NAs.")
    return(result)
  }

  # Hosking 1985 approximation; k follows Hosking sign convention
  # (opposite of the textbook xi).
  c_const <- 2 / (3 + t3) - log(2) / log(3)
  k <- 7.8590 * c_const + 2.9554 * c_const^2

  if (abs(k) < 1e-6) {
    # Gumbel limit: F(x) = exp(-exp(-(x - u) / alpha))
    alpha <- l2 / log(2)
    if (!is.finite(alpha) || alpha <= 0) return(result)
    u <- l1 - alpha * 0.5772156649  # Euler-Mascheroni
    cdf <- exp(-exp(-(xv - u) / alpha))
  } else {
    g1 <- gamma(1 + k)
    if (!is.finite(g1)) return(result)
    alpha <- l2 * k / ((1 - 2^(-k)) * g1)
    if (!is.finite(alpha) || alpha <= 0) return(result)
    u <- l1 - alpha * (1 - g1) / k
    z <- 1 - k * (xv - u) / alpha
    z[z <= 0] <- NA_real_
    cdf <- exp(-z^(1 / k))
  }

  cdf[is.na(cdf)] <- 0.5
  cdf[cdf <= 0] <- 0.001
  cdf[cdf >= 1] <- 0.999

  result[valid] <- stats::qnorm(cdf)
  result
}

#' Fit log-logistic distribution and transform to standard normal
#'
#' Uses unbiased probability weighted moments (Hosking 1990) to estimate
#' log-logistic parameters, consistent with Vicente-Serrano et al. (2010).
#' @noRd
.loglogistic_to_normal <- function(x) {
  result <- rep(NA_real_, length(x))
  valid <- !is.na(x)
  xv <- x[valid]
  n <- length(xv)

  if (n < 3) return(result)

  # Shift to positive for PWM estimation
  xv_shifted <- xv - min(xv) + 1
  xv_sorted <- sort(xv_shifted)

  # Unbiased PWMs (Hosking 1990)
  ranks <- seq_len(n)
  b0 <- mean(xv_sorted)
  b1 <- sum(((ranks - 1) / (n - 1)) * xv_sorted) / n
  b2 <- sum(((ranks - 1) * (ranks - 2) / ((n - 1) * (n - 2))) *
              xv_sorted) / n

  # L-moments
  l1 <- b0
  l2 <- 2 * b1 - b0
  l3 <- 6 * b2 - 6 * b1 + b0

  # Log-logistic parameters from L-moments
  if (l2 <= 0) {
    cli::cli_warn("SPEI fitting failed: L-moment l2 <= 0. Returning NAs.")
    return(result)
  }
  t3 <- l3 / l2  # L-skewness

  if (abs(t3) >= 1) {
    cli::cli_warn("SPEI fitting failed: L-skewness out of range. Returning NAs.")
    return(result)
  }

  beta_ll <- t3 * pi / (3 * sin(t3 * pi / 3))

  if (is.na(beta_ll) || !is.finite(beta_ll) || beta_ll <= 0) {
    cli::cli_warn("SPEI fitting failed: invalid shape parameter. Returning NAs.")
    return(result)
  }

  g1 <- gamma(1 + 1 / beta_ll)
  g2 <- gamma(1 - 1 / beta_ll)
  if (!is.finite(g1) || !is.finite(g2) || g1 * g2 == 0) {
    cli::cli_warn("SPEI fitting failed: gamma function overflow. Returning NAs.")
    return(result)
  }

  alpha_ll <- l2 / (g1 * g2)
  xi <- l1 - alpha_ll * g1 * g2

  if (is.na(alpha_ll) || alpha_ll <= 0) {
    cli::cli_warn("SPEI fitting failed: invalid scale parameter. Returning NAs.")
    return(result)
  }

  # Compute CDF for each observation (in original order)
  cdf <- vapply(seq_along(xv), function(i) {
    a <- (xv_shifted[i] - xi) / alpha_ll
    if (a <= 0) 0.001 else 1 / (1 + a^(-beta_ll))
  }, numeric(1))

  cdf[cdf <= 0] <- 0.001
  cdf[cdf >= 1] <- 0.999
  result[valid] <- stats::qnorm(cdf)
  result
}

test_that("ck_spi returns data frame with correct structure", {
  dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
  set.seed(42)
  precip <- rgamma(length(dates), shape = 0.5, rate = 0.1)
  result <- ck_spi(precip, dates, scale = 3)
  expect_s3_class(result, "data.frame")
  expect_true(all(c("period", "value", "index", "unit") %in% names(result)))
  expect_equal(unique(result$index), "spi")
  expect_equal(unique(result$unit), "dimensionless")
})

test_that("ck_spi values are roughly standard normal", {
  dates <- seq(as.Date("2015-01-01"), as.Date("2023-12-31"), by = "day")
  set.seed(123)
  precip <- rgamma(length(dates), shape = 2, rate = 0.5)
  result <- ck_spi(precip, dates, scale = 3)
  # Mean should be near 0, sd near 1

  expect_true(abs(mean(result$value, na.rm = TRUE)) < 0.5)
  expect_true(sd(result$value, na.rm = TRUE) > 0.3)
  expect_true(sd(result$value, na.rm = TRUE) < 2)
})

test_that("ck_spi rejects too-short data", {
  dates <- as.Date("2024-01-01") + 0:29
  precip <- rep(1, 30)
  expect_error(ck_spi(precip, dates, scale = 6), "Not enough months")
})

test_that("ck_spi rejects invalid scale", {
  dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
  precip <- rep(1, length(dates))
  expect_error(ck_spi(precip, dates, scale = -1), "positive integer")
})

test_that("ck_spei returns data frame", {
  dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
  set.seed(42)
  precip <- rgamma(length(dates), shape = 0.5, rate = 0.1)
  pet <- rep(3, length(dates))
  result <- ck_spei(precip, pet, dates, scale = 3)
  expect_s3_class(result, "data.frame")
  expect_equal(unique(result$index), "spei")
})

test_that("ck_spei rejects mismatched lengths", {
  dates <- as.Date("2020-01-01") + 0:99
  expect_error(ck_spei(rep(1, 100), rep(1, 50), dates), "same length")
})

test_that("ck_pet returns daily PET values", {
  dates <- as.Date("2024-07-01") + 0:9
  tmin <- c(15, 16, 14, 17, 15, 13, 16, 14, 15, 16)
  tmax <- c(30, 32, 28, 33, 31, 27, 34, 29, 30, 32)
  result <- ck_pet(tmin, tmax, lat = 45, dates = dates)
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 10)
  expect_true(all(result$value >= 0))
  expect_equal(result$index[1], "pet")
  expect_true("date" %in% names(result))
})

test_that("ck_pet rejects mismatched lengths", {
  dates <- as.Date("2024-07-01") + 0:4
  expect_error(ck_pet(1:5, 1:3, lat = 45, dates = dates), "same length")
})

test_that("ck_pet rejects non-numeric lat", {
  dates <- as.Date("2024-07-01") + 0:4
  expect_error(ck_pet(1:5, 6:10, lat = "a", dates = dates), "numeric")
})

test_that("ck_pet at equator produces positive values", {
  dates <- as.Date("2024-06-21") + 0:4
  tmin <- rep(20, 5)
  tmax <- rep(35, 5)
  result <- ck_pet(tmin, tmax, lat = 0, dates = dates)
  expect_true(all(result$value > 0))
})

test_that("monthly totals helper works", {
  dates <- seq(as.Date("2024-01-01"), as.Date("2024-03-31"), by = "day")
  x <- rep(1, length(dates))
  result <- climatekit:::.monthly_totals(x, dates)
  expect_equal(nrow(result), 3)
  expect_equal(result$total[1], 31)  # January
  expect_equal(result$total[2], 29)  # February (2024 is leap year)
  expect_equal(result$total[3], 31)  # March
})

# --- Reference value tests ---

test_that("PET Hargreaves reference value for lat=45, July 15", {
  # Hand calculation: Tmin=15, Tmax=30, Tavg=22.5, TD=15
  # DOY=197 (July 15, 2024)
  # PET = 0.0023 * Ra * sqrt(15) * (22.5 + 17.8)
  dates <- as.Date("2024-07-15")
  result <- ck_pet(tmin = 15, tmax = 30, lat = 45, dates = dates)
  # PET should be a reasonable value (3-8 mm/day for mid-latitude summer)
  expect_true(result$value > 2)
  expect_true(result$value < 10)
})

test_that("SPI with known gamma-distributed data", {
  # Generate 5 years of data from known gamma
  dates <- seq(as.Date("2018-01-01"), as.Date("2022-12-31"), by = "day")
  set.seed(42)
  precip <- rgamma(length(dates), shape = 2, rate = 0.5)
  result <- ck_spi(precip, dates, scale = 3)
  # SPI values should be roughly standard normal
  vals <- result$value[!is.na(result$value)]
  expect_true(abs(mean(vals)) < 0.5)
  expect_true(sd(vals) > 0.5)
  expect_true(sd(vals) < 2.0)
})

test_that("SPI fits per calendar month", {
  # With seasonal precipitation (wet summers, dry winters),
  # per-month fitting should produce values closer to 0 within each month
  dates <- seq(as.Date("2010-01-01"), as.Date("2023-12-31"), by = "day")
  set.seed(99)
  doy <- as.integer(format(dates, "%j"))
  # Strong seasonal signal: summer wet, winter dry
  seasonal <- pmax(0, sin(2 * pi * doy / 365) * 8 + 2)
  precip <- rgamma(length(dates), shape = 1, rate = 1) * seasonal
  result <- ck_spi(precip, dates, scale = 3)

  # Per-month fitting: January SPIs should be near 0 as a group
  jan_vals <- result$value[format(result$period, "%m") == "01"]
  if (length(jan_vals) >= 3) {
    expect_true(abs(mean(jan_vals, na.rm = TRUE)) < 1.0)
  }
})

test_that("SPI handles all-zero precipitation gracefully", {
  dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
  precip <- rep(0, length(dates))
  result <- ck_spi(precip, dates, scale = 3)
  # Should return NAs or valid values without error
  expect_s3_class(result, "data.frame")
})

test_that("SPEI warns on fitting failure rather than silently falling back", {
  # With constant water balance, fitting should fail and warn
  dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
  precip <- rep(5, length(dates))
  pet <- rep(3, length(dates))
  # Constant difference -> zero variance -> should warn about fitting failure
  expect_warning(
    result <- ck_spei(precip, pet, dates, scale = 3),
    "SPEI fitting failed"
  )
  expect_s3_class(result, "data.frame")
})

# Distribution choice (Phase C) ----------------------------------------------

test_that("ck_spi pearsonIII produces standardised output", {
  dates <- seq(as.Date("1961-01-01"), as.Date("1990-12-31"), by = "day")
  set.seed(101)
  precip <- rgamma(length(dates), shape = 0.5, rate = 0.1)
  g <- ck_spi(precip, dates, scale = 3, distribution = "gamma")
  p3 <- ck_spi(precip, dates, scale = 3, distribution = "pearsonIII")
  expect_s3_class(p3, "data.frame")
  # Standardised: mean near 0, sd near 1
  expect_lt(abs(mean(p3$value, na.rm = TRUE)), 0.3)
  expect_lt(abs(sd(p3$value, na.rm = TRUE) - 1), 0.3)
  # gamma and pearsonIII should agree to within reasonable tolerance
  expect_gt(cor(g$value, p3$value, use = "complete.obs"), 0.9)
})

test_that("ck_spei gev produces standardised output", {
  dates <- seq(as.Date("1961-01-01"), as.Date("1990-12-31"), by = "day")
  set.seed(102)
  precip <- rgamma(length(dates), shape = 0.5, rate = 0.1)
  pet <- rep(3, length(dates))
  gev <- ck_spei(precip, pet, dates, scale = 3, distribution = "gev")
  expect_s3_class(gev, "data.frame")
  expect_lt(abs(mean(gev$value, na.rm = TRUE)), 0.3)
  expect_lt(abs(sd(gev$value, na.rm = TRUE) - 1), 0.3)
})

test_that("invalid distribution argument is rejected", {
  dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
  precip <- rep(1, length(dates))
  pet <- rep(0.5, length(dates))
  expect_error(ck_spi(precip, dates, distribution = "bogus"), "should be one of")
  expect_error(ck_spei(precip, pet, dates, distribution = "bogus"), "should be one of")
})

# FAO-56 Penman-Monteith PET (Phase E) ---------------------------------------

test_that("ck_pet_pm returns sensible mid-summer mid-latitude values", {
  dates <- as.Date("2024-07-01") + 0:9
  tmin <- c(15, 16, 14, 17, 15, 13, 16, 14, 15, 16)
  tmax <- c(30, 32, 28, 33, 31, 27, 34, 29, 30, 32)
  result <- ck_pet_pm(tmin, tmax, lat = 45, dates = dates)
  expect_s3_class(result, "data.frame")
  expect_equal(result$index[1], "pet_pm")
  expect_equal(result$unit[1], "mm")
  # Range check: typical July ETo at 45N is 3-7 mm/day.
  expect_true(all(result$value >= 2 & result$value <= 8))
})

test_that("ck_pet_pm respects elevation (higher = lower P -> nudges PET)", {
  dates <- as.Date("2024-07-01") + 0:9
  tmin <- rep(15, 10)
  tmax <- rep(30, 10)
  sea  <- ck_pet_pm(tmin, tmax, lat = 45, dates = dates, elev = 0)
  high <- ck_pet_pm(tmin, tmax, lat = 45, dates = dates, elev = 2000)
  expect_false(identical(sea$value, high$value))
})

test_that("ck_pet_pm honours supplied solar radiation rs", {
  dates <- as.Date("2024-07-01") + 0:4
  tmin <- rep(15, 5)
  tmax <- rep(30, 5)
  est <- ck_pet_pm(tmin, tmax, lat = 45, dates = dates)
  fix <- ck_pet_pm(tmin, tmax, lat = 45, dates = dates,
                   rs = rep(20, 5))
  expect_false(identical(est$value, fix$value))
})

test_that("ck_pet_pm rejects invalid arguments", {
  dates <- as.Date("2024-07-01") + 0:4
  tmin <- rep(15, 5)
  tmax <- rep(30, 5)
  expect_error(ck_pet_pm(tmin, tmax, lat = "x", dates = dates), "numeric")
  expect_error(ck_pet_pm(tmin, c(30, 30), lat = 45, dates = dates),
               "same length")
  expect_error(ck_pet_pm(tmin, tmax, lat = 45, dates = dates, albedo = 1.5),
               "\\[0, 1\\]")
})

# Gridded / netCDF support.
#
# climatekit's core functions take vectors and return tidy data frames.
# These helpers let users apply any ck_* function over a SpatRaster
# (one time series per grid cell) or read a netCDF file as a
# SpatRaster. The hard work is delegated to terra and ncdf4, both in
# Suggests, so loading climatekit alone does not pull them in.

#' Read a netCDF File as a SpatRaster
#'
#' Convenience wrapper that delegates to [terra::rast()]. Reads the
#' file at `path` and returns a SpatRaster, optionally restricted to a
#' single variable. terra and ncdf4 must be installed (both are listed
#' in `Suggests:`).
#'
#' @param path Character. Path to a netCDF file.
#' @param var Character or `NULL`. Variable to extract. If `NULL`, the
#'   default behaviour of [terra::rast()] applies.
#'
#' @return A SpatRaster (one layer per time step in the netCDF file).
#'
#' @export
#' @examples
#' \dontrun{
#'   r <- ck_from_netcdf("tas_day.nc", var = "tas")
#'   terra::nlyr(r)  # number of daily layers
#' }
ck_from_netcdf <- function(path, var = NULL) {
  if (!requireNamespace("terra", quietly = TRUE)) {
    cli::cli_abort(c(
      "Package {.pkg terra} is required for {.fn ck_from_netcdf}.",
      "i" = "Install it with {.code install.packages(\"terra\")}."
    ))
  }
  if (!requireNamespace("ncdf4", quietly = TRUE)) {
    cli::cli_warn(c(
      "Package {.pkg ncdf4} is recommended for reading 'netCDF' files.",
      "i" = "{.pkg terra}'s GDAL bindings can read .nc on many systems, but",
      "i" = "if {.fn terra::rast} fails on your file, install it with",
      "i" = "{.code install.packages(\"ncdf4\")}."
    ))
  }
  if (!is.character(path) || length(path) != 1L) {
    cli::cli_abort("{.arg path} must be a single character string.")
  }
  if (!file.exists(path)) {
    cli::cli_abort("File not found: {.path {path}}.")
  }
  if (is.null(var)) {
    terra::rast(path)
  } else {
    terra::rast(path, subds = var)
  }
}

#' Apply a climatekit Index Function Over a SpatRaster
#'
#' Compute a daily climatekit index function for every cell in a
#' SpatRaster `x` whose layers represent successive days, and return
#' the per-period results as a SpatRaster. The supplied function `fun`
#' must accept a numeric vector and a Date vector and return a data
#' frame with `period` and `value` columns (the standard climatekit
#' shape).
#'
#' All cells must share the same `dates`. Cells that are entirely `NA`
#' are returned as `NA`. Run-time scales linearly with the number of
#' cells; for very large grids consider sub-setting first.
#'
#' @param x A SpatRaster. Layers correspond one-to-one with `dates`.
#' @param fun A `ck_*` function (or any function with the same
#'   signature: numeric vector + Date vector + optional named
#'   arguments, returning a data frame with `period` and `value`).
#' @param dates Date vector of length `terra::nlyr(x)`.
#' @param ... Additional named arguments forwarded to `fun` (for
#'   example `period = "annual"`, `ref_start`, `ref_end`).
#'
#' @return A SpatRaster with one layer per output period (layer names
#'   are the period labels).
#'
#' @export
#' @examples
#' \donttest{
#'   if (requireNamespace("terra", quietly = TRUE)) {
#'     dates <- seq(as.Date("2024-01-01"), as.Date("2024-12-31"), by = "day")
#'     n <- length(dates)
#'     # Tiny 2x2 SpatRaster of synthetic daily Tmax
#'     r <- terra::rast(nrows = 2, ncols = 2, nlyrs = n,
#'                      xmin = 0, xmax = 2, ymin = 0, ymax = 2)
#'     set.seed(1)
#'     for (i in seq_len(n)) {
#'       terra::values(r[[i]]) <- rnorm(4, 15, 5)
#'     }
#'     txx_r <- ck_apply_grid(r, ck_txx, dates = dates, period = "annual")
#'     terra::nlyr(txx_r)
#'   }
#' }
ck_apply_grid <- function(x, fun, dates, ...) {
  if (!requireNamespace("terra", quietly = TRUE)) {
    cli::cli_abort(c(
      "Package {.pkg terra} is required for {.fn ck_apply_grid}.",
      "i" = "Install it with {.code install.packages(\"terra\")}."
    ))
  }
  if (!inherits(x, "SpatRaster")) {
    cli::cli_abort("{.arg x} must be a {.cls SpatRaster}.")
  }
  fun <- match.fun(fun)
  validate_dates(dates, terra::nlyr(x))

  # Probe one non-NA cell to learn the output period labels and length.
  ref_idx <- NA_integer_
  for (i in seq_len(terra::ncell(x))) {
    vals <- as.numeric(x[i])
    if (any(!is.na(vals))) {
      ref_idx <- i
      break
    }
  }
  if (is.na(ref_idx)) {
    cli::cli_abort("All cells in {.arg x} are NA.")
  }
  ref_res <- fun(as.numeric(x[ref_idx]), dates, ...)
  if (!is.data.frame(ref_res) ||
      !all(c("period", "value") %in% names(ref_res))) {
    cli::cli_abort(
      "{.arg fun} must return a data frame with {.field period} and {.field value} columns."
    )
  }
  n_periods <- nrow(ref_res)
  period_labels <- as.character(ref_res$period)

  cell_fun <- function(vals, ...) {
    if (all(is.na(vals))) {
      return(rep(NA_real_, n_periods))
    }
    res <- fun(vals, dates, ...)
    if (nrow(res) != n_periods) {
      return(rep(NA_real_, n_periods))
    }
    res$value
  }

  out <- terra::app(x, cell_fun, ...)
  names(out) <- period_labels
  out
}

# Tests for the gridded helpers (ck_apply_grid, ck_from_netcdf).
# Skip cleanly on systems without terra / ncdf4 in Suggests.

test_that("ck_apply_grid errors on non-SpatRaster input", {
  skip_if_not_installed("terra")
  expect_error(
    ck_apply_grid(matrix(1:9, 3, 3), ck_txx, as.Date("2024-01-01") + 0:8),
    "SpatRaster"
  )
})

test_that("ck_apply_grid runs ck_txx over a tiny synthetic grid", {
  skip_if_not_installed("terra")
  set.seed(1)
  dates <- seq(as.Date("1961-01-01"), as.Date("1961-12-31"), by = "day")
  n_layers <- length(dates)
  # 4-cell grid with realistic seasonal Tmax
  cells <- 4L
  vals <- matrix(
    rep(15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365),
        each = cells) +
      rnorm(cells * n_layers),
    nrow = cells
  )
  r <- terra::rast(nrows = 2, ncols = 2, nlyrs = n_layers,
                   xmin = 0, xmax = 2, ymin = 0, ymax = 2)
  for (i in seq_len(n_layers)) {
    terra::values(r[[i]]) <- vals[, i]
  }

  out <- ck_apply_grid(r, ck_txx, dates = dates, period = "annual")
  expect_s4_class(out, "SpatRaster")
  expect_equal(terra::nlyr(out), 1L)
  expect_equal(terra::ncell(out), 4L)
  # Each cell's TXx should be near the seasonal max plus noise (~25 °C)
  vals_out <- as.numeric(terra::values(out))
  expect_true(all(vals_out > 20 & vals_out < 35))
})

test_that("ck_apply_grid layer names match output periods", {
  skip_if_not_installed("terra")
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  n_layers <- length(dates)
  set.seed(2)
  vals <- matrix(rnorm(2 * n_layers, 15, 5), nrow = 2)
  r <- terra::rast(nrows = 1, ncols = 2, nlyrs = n_layers,
                   xmin = 0, xmax = 2, ymin = 0, ymax = 1)
  for (i in seq_len(n_layers)) {
    terra::values(r[[i]]) <- vals[, i]
  }
  out <- ck_apply_grid(r, ck_txx, dates = dates, period = "annual")
  expect_equal(terra::nlyr(out), 2L)
  expect_equal(names(out), c("1961-01-01", "1962-01-01"))
})

test_that("ck_apply_grid handles all-NA cells gracefully", {
  skip_if_not_installed("terra")
  dates <- seq(as.Date("1961-01-01"), as.Date("1961-12-31"), by = "day")
  n_layers <- length(dates)
  set.seed(3)
  vals <- matrix(rnorm(2 * n_layers, 15, 5), nrow = 2)
  vals[2, ] <- NA_real_  # second cell entirely NA
  r <- terra::rast(nrows = 1, ncols = 2, nlyrs = n_layers,
                   xmin = 0, xmax = 2, ymin = 0, ymax = 1)
  for (i in seq_len(n_layers)) {
    terra::values(r[[i]]) <- vals[, i]
  }
  out <- ck_apply_grid(r, ck_txx, dates = dates, period = "annual")
  vals_out <- as.numeric(terra::values(out))
  expect_false(is.na(vals_out[1]))
  expect_true(is.na(vals_out[2]))
})

test_that("ck_from_netcdf errors on missing file", {
  skip_if_not_installed("terra")
  expect_error(ck_from_netcdf("does_not_exist.nc"), "not found")
})

test_that("ck_from_netcdf rejects non-string path", {
  skip_if_not_installed("terra")
  expect_error(ck_from_netcdf(c("a", "b")), "single character")
})

#' Convert Temperature Units
#'
#' Convert between Celsius, Fahrenheit, and Kelvin.
#'
#' @param x Numeric vector of temperatures.
#' @param from Character. Source unit: `"C"`, `"F"`, or `"K"`.
#' @param to Character. Target unit: `"C"`, `"F"`, or `"K"`.
#'
#' @return Numeric vector of converted temperatures.
#'
#' @export
#' @examples
#' ck_convert_temp(c(0, 100), from = "C", to = "F")
#' ck_convert_temp(32, from = "F", to = "C")
ck_convert_temp <- function(x, from, to) {
  validate_numeric(x, "x")
  from <- match.arg(from, c("C", "F", "K"))
  to <- match.arg(to, c("C", "F", "K"))

  if (from == to) return(x)

  # Convert to Celsius first
  celsius <- switch(from,
    "C" = x,
    "F" = (x - 32) * 5 / 9,
    "K" = x - 273.15
  )

  # Convert from Celsius to target
  switch(to,
    "C" = celsius,
    "F" = celsius * 9 / 5 + 32,
    "K" = celsius + 273.15
  )
}

#' Clear Cache
#'
#' Removes any cached reference data stored by `climatekit`.
#'
#' @return Invisibly returns `TRUE` if cache was cleared, `FALSE` if no cache
#'   existed.
#'
#' @export
#' @examples
#' \donttest{
#' op <- options(climatekit.cache_dir = tempdir())
#' clear_cache()
#' options(op)
#' }
clear_cache <- function() {
  cache_dir <- getOption("climatekit.cache_dir",
                         default = tools::R_user_dir("climatekit", "cache"))
  if (dir.exists(cache_dir)) {
    files <- list.files(cache_dir, full.names = TRUE)
    if (length(files)) unlink(files, recursive = TRUE)
    cli::cli_alert_success("Cache cleared.")
    invisible(TRUE)
  } else {
    cli::cli_alert_info("No cache to clear.")
    invisible(FALSE)
  }
}

#' Compute a Climate Index by Name
#'
#' A generic dispatcher that calls the appropriate `ck_*` function based on
#' a string index name. Useful for programmatic workflows where the index
#' is selected at runtime.
#'
#' @param data A named list or data frame containing the required input
#'   vectors. Column names should match function argument names (e.g.
#'   `tmin`, `tmax`, `precip`, `dates`).
#' @param index Character. Name of the index to compute (e.g.
#'   `"frost_days"`). Use [ck_available()] to see valid names.
#' @param ... Additional arguments passed to the underlying function (e.g.
#'   `period`, `threshold`, `base`).
#'
#' @return A data frame as returned by the underlying `ck_*` function.
#'
#' @export
#' @examples
#' d <- data.frame(
#'   dates = as.Date("2024-01-01") + 0:9,
#'   tmin = c(-2, 3, -1, 5, -3, 0, 2, -4, 1, -1)
#' )
#' ck_compute(d, "frost_days")
ck_compute <- function(data, index, ...) {
  if (!is.character(index) || length(index) != 1) {
    cli::cli_abort("{.arg index} must be a single character string.")
  }

  # Map index names to functions and required columns
  dispatch <- list(
    frost_days       = list(fn = ck_frost_days, args = c("tmin", "dates")),
    ice_days         = list(fn = ck_ice_days, args = c("tmax", "dates")),
    summer_days      = list(fn = ck_summer_days, args = c("tmax", "dates")),
    tropical_nights  = list(fn = ck_tropical_nights, args = c("tmin", "dates")),
    growing_season   = list(fn = ck_growing_season, args = c("tavg", "dates")),
    txx              = list(fn = ck_txx, args = c("tmax", "dates")),
    tnx              = list(fn = ck_tnx, args = c("tmin", "dates")),
    txn              = list(fn = ck_txn, args = c("tmax", "dates")),
    tnn              = list(fn = ck_tnn, args = c("tmin", "dates")),
    tx10p            = list(fn = ck_tx10p, args = c("tmax", "dates")),
    tn10p            = list(fn = ck_tn10p, args = c("tmin", "dates")),
    tx90p            = list(fn = ck_tx90p, args = c("tmax", "dates")),
    tn90p            = list(fn = ck_tn90p, args = c("tmin", "dates")),
    r95p             = list(fn = ck_r95p, args = c("precip", "dates")),
    r99p             = list(fn = ck_r99p, args = c("precip", "dates")),
    wsdi             = list(fn = ck_wsdi, args = c("tmax", "dates")),
    csdi             = list(fn = ck_csdi, args = c("tmin", "dates")),
    hwn              = list(fn = ck_hwn, args = c("tmax", "dates")),
    hwf              = list(fn = ck_hwf, args = c("tmax", "dates")),
    hwd              = list(fn = ck_hwd, args = c("tmax", "dates")),
    hwm              = list(fn = ck_hwm, args = c("tmax", "dates")),
    hwa              = list(fn = ck_hwa, args = c("tmax", "dates")),
    cwn              = list(fn = ck_cwn, args = c("tmin", "dates")),
    cwf              = list(fn = ck_cwf, args = c("tmin", "dates")),
    cwd              = list(fn = ck_cwd, args = c("tmin", "dates")),
    cwm              = list(fn = ck_cwm, args = c("tmin", "dates")),
    cwa              = list(fn = ck_cwa, args = c("tmin", "dates")),
    ehf              = list(fn = ck_ehf, args = c("tmax", "tmin", "dates")),
    heating_degree_days = list(fn = ck_heating_degree_days, args = c("tavg", "dates")),
    cooling_degree_days = list(fn = ck_cooling_degree_days, args = c("tavg", "dates")),
    growing_degree_days = list(fn = ck_growing_degree_days, args = c("tavg", "dates")),
    diurnal_range    = list(fn = ck_diurnal_range, args = c("tmin", "tmax", "dates")),
    warm_spell       = list(fn = ck_warm_spell, args = c("tmax", "dates")),
    dry_days         = list(fn = ck_dry_days, args = c("precip", "dates")),
    wet_days         = list(fn = ck_wet_days, args = c("precip", "dates")),
    total_precip     = list(fn = ck_total_precip, args = c("precip", "dates")),
    heavy_precip     = list(fn = ck_heavy_precip, args = c("precip", "dates")),
    very_heavy_precip = list(fn = ck_very_heavy_precip, args = c("precip", "dates")),
    max_1day_precip  = list(fn = ck_max_1day_precip, args = c("precip", "dates")),
    max_5day_precip  = list(fn = ck_max_5day_precip, args = c("precip", "dates")),
    precip_intensity = list(fn = ck_precip_intensity, args = c("precip", "dates")),
    spi              = list(fn = ck_spi, args = c("precip", "dates")),
    spei             = list(fn = ck_spei, args = c("precip", "pet", "dates")),
    pet              = list(fn = ck_pet, args = c("tmin", "tmax", "lat", "dates")),
    pet_pm           = list(fn = ck_pet_pm, args = c("tmin", "tmax", "lat", "dates")),
    huglin           = list(fn = ck_huglin, args = c("tmin", "tmax", "dates", "lat")),
    winkler          = list(fn = ck_winkler, args = c("tavg", "dates")),
    branas           = list(fn = ck_branas, args = c("precip", "tavg", "dates")),
    first_frost      = list(fn = ck_first_frost, args = c("tmin", "dates")),
    last_frost       = list(fn = ck_last_frost, args = c("tmin", "dates")),
    wind_chill       = list(fn = ck_wind_chill, args = c("tavg", "wind_speed")),
    heat_index       = list(fn = ck_heat_index, args = c("tavg", "humidity")),
    humidex          = list(fn = ck_humidex, args = c("tavg", "dewpoint")),
    fire_danger      = list(fn = ck_fire_danger,
                            args = c("tavg", "humidity", "wind_speed", "precip"))
  )

  if (!index %in% names(dispatch)) {
    cli::cli_abort(
      c("Unknown index {.val {index}}.",
        "i" = "Run {.fn ck_available} to see valid index names.")
    )
  }

  entry <- dispatch[[index]]
  required <- entry$args

  # Extract columns from data
  if (is.data.frame(data)) {
    data <- as.list(data)
  }

  missing_cols <- setdiff(required, names(data))
  if (length(missing_cols) > 0) {
    cli::cli_abort(
      "Missing required column{?s}: {.field {missing_cols}}."
    )
  }

  call_args <- data[required]
  call_args <- c(call_args, list(...))
  do.call(entry$fn, call_args)
}

# Zhang et al. (2005) in-base bootstrap for percentile-day indices.
#
# When a percentile-based index (TX10p / TN10p / TX90p / TN90p) is
# computed for a year that lies inside the reference period, the year
# being analysed contributes to its own threshold and biases the
# result toward the central tendency. The Zhang (2005) bootstrap
# corrects this by leaving the analysis year out and substituting one
# of the other (N - 1) reference years in turn, then averaging the
# resulting exceedance proportions.
#
# Reference: Zhang, X., Hegerl, G. C., Zwiers, F. W., & Kenyon, J.
# (2005). Avoiding inhomogeneity in percentile-based indices of
# temperature extremes. *Journal of Climate*, 18(11), 1641-1651.
# \doi{10.1175/JCLI3366.1}.

#' Bootstrap-corrected per-day exceedance proportions
#'
#' Returns a numeric vector aligned with `values` and `dates`.
#' Out-of-base years yield 0/1 booleans (NA where the value or
#' threshold is missing). In-base years yield averages in [0, 1]
#' from the leave-one-out resampling. Used by [ck_tx10p()],
#' [ck_tn10p()], [ck_tx90p()], and [ck_tn90p()] when
#' `bootstrap = TRUE`.
#' @noRd
.bootstrap_pct_exceedance <- function(values, dates, percentile, op,
                                      ref_start, ref_end, window = 5L) {
  doy <- as.integer(format(dates, "%j"))
  years <- as.integer(format(dates, "%Y"))

  in_ref <- years >= ref_start & years <= ref_end
  if (!any(in_ref)) {
    cli::cli_abort(
      "No data in reference period {ref_start}-{ref_end}; supply data covering the period or pass {.arg ref_start} / {.arg ref_end}."
    )
  }

  # Standard full-period thresholds for out-of-base years.
  std_thresh <- .calendar_day_percentile(values, dates, percentile,
                                         ref_start, ref_end, window = window)

  out <- rep(NA_real_, length(values))

  # --- Out-of-base years: standard binary exceedance ---
  oob_idx <- which(!in_ref)
  if (length(oob_idx) > 0L) {
    oob_thr <- std_thresh[doy[oob_idx]]
    oob_vals <- values[oob_idx]
    valid <- !is.na(oob_vals) & !is.na(oob_thr)
    if (identical(op, ">")) {
      out[oob_idx] <- ifelse(valid, as.numeric(oob_vals > oob_thr), NA_real_)
    } else {
      out[oob_idx] <- ifelse(valid, as.numeric(oob_vals < oob_thr), NA_real_)
    }
  }

  # --- In-base years: leave-one-out bootstrap ---
  #
  # The naive implementation rebuilds the calendar-day percentile pool
  # from the full `values` vector for every (y, w) pair, costing
  # O(N^2 * 366 * N_window) operations for an N-year reference. The
  # optimised implementation precomputes per-year-per-DOY contribution
  # vectors once, then for each (y, w) simply concatenates the slices
  # for "all years except y, plus year w again" and runs `quantile` on
  # the result. That cuts the inner cost to O(366) per (y, w) and the
  # total to O(N^2 * 366), roughly a 30x speedup for a 30-year base.
  ref_years <- seq(ref_start, ref_end)
  n_ref <- length(ref_years)
  half  <- (window - 1L) %/% 2L

  # Per-year-per-DOY contribution: year_doy_pool[[i]][[d]] = numeric
  # vector of values from ref-year `ref_years[i]` whose day-of-year
  # falls in the (window)-day window centred on calendar day d.
  year_doy_pool <- vector("list", n_ref)
  for (i in seq_along(ref_years)) {
    yr <- ref_years[i]
    yr_idx_full <- which(years == yr)
    yr_doy_full <- doy[yr_idx_full]
    yr_vals_full <- values[yr_idx_full]
    pool_i <- vector("list", 366L)
    for (d in seq_len(366L)) {
      target_doys <- ((d - half - 1L):(d + half - 1L)) %% 366L + 1L
      slice <- yr_vals_full[yr_doy_full %in% target_doys]
      slice <- slice[!is.na(slice)]
      pool_i[[d]] <- slice
    }
    year_doy_pool[[i]] <- pool_i
  }

  for (i_y in seq_along(ref_years)) {
    yr <- ref_years[i_y]
    yr_idx <- which(years == yr)
    if (length(yr_idx) == 0L) next

    yr_doy <- doy[yr_idx]
    yr_vals <- values[yr_idx]

    other_idx <- setdiff(seq_along(ref_years), i_y)

    # Single-year reference: bootstrap is undefined; fall back to standard.
    if (length(other_idx) == 0L) {
      yr_thr <- std_thresh[yr_doy]
      valid <- !is.na(yr_vals) & !is.na(yr_thr)
      if (identical(op, ">")) {
        out[yr_idx] <- ifelse(valid, as.numeric(yr_vals > yr_thr), NA_real_)
      } else {
        out[yr_idx] <- ifelse(valid, as.numeric(yr_vals < yr_thr), NA_real_)
      }
      next
    }

    # Pool with year `i_y` removed, per DOY (one-time per y).
    minus_y_pool <- vector("list", 366L)
    for (d in seq_len(366L)) {
      minus_y_pool[[d]] <- unlist(
        lapply(other_idx, function(j) year_doy_pool[[j]][[d]]),
        use.names = FALSE
      )
    }

    bs_sum   <- rep(0, length(yr_idx))
    bs_count <- rep(0L, length(yr_idx))

    for (i_w in other_idx) {
      # Synthetic ref pool: everything except year y, with year w's
      # values appended (representing the substitution for year y).
      mod_thresh <- vapply(seq_len(366L), function(d) {
        pool <- c(minus_y_pool[[d]], year_doy_pool[[i_w]][[d]])
        if (length(pool) == 0L) return(NA_real_)
        stats::quantile(pool, percentile, na.rm = TRUE,
                        names = FALSE, type = 8L)
      }, numeric(1))

      yr_thr <- mod_thresh[yr_doy]
      valid <- !is.na(yr_vals) & !is.na(yr_thr)
      if (identical(op, ">")) {
        ex <- ifelse(valid, as.numeric(yr_vals > yr_thr), NA_real_)
      } else {
        ex <- ifelse(valid, as.numeric(yr_vals < yr_thr), NA_real_)
      }
      bs_sum[valid]   <- bs_sum[valid]   + ex[valid]
      bs_count[valid] <- bs_count[valid] + 1L
    }

    out[yr_idx] <- ifelse(bs_count > 0L, bs_sum / bs_count, NA_real_)
  }

  out
}

#' Per-day exceedance proportions for percentile indices
#'
#' Single entry point for the percentile-day functions. When
#' `bootstrap = FALSE`, returns standard 0/1 booleans (NA where the
#' value or threshold is missing). When `bootstrap = TRUE`, applies
#' the Zhang (2005) leave-one-out bootstrap for in-base years.
#' @noRd
.pct_exceedance_per_day <- function(values, dates, percentile, op,
                                    ref_start, ref_end,
                                    bootstrap = FALSE, window = 5L) {
  if (isTRUE(bootstrap)) {
    return(.bootstrap_pct_exceedance(values, dates, percentile, op,
                                     ref_start, ref_end, window))
  }

  thresholds <- .calendar_day_percentile(values, dates, percentile,
                                         ref_start, ref_end, window = window)
  doy <- as.integer(format(dates, "%j"))
  thr <- thresholds[doy]
  valid <- !is.na(values) & !is.na(thr)

  if (identical(op, ">")) {
    ifelse(valid, as.numeric(values > thr), NA_real_)
  } else if (identical(op, "<")) {
    ifelse(valid, as.numeric(values < thr), NA_real_)
  } else {
    cli::cli_abort("Internal: unsupported {.arg op}.")
  }
}

#' Aggregate per-day exceedance proportions to per-period percentages
#'
#' Helper used by the percentile-day functions to convert a daily
#' soft exceedance vector (in [0, 1] or NA) into a per-period
#' percentage. Periods with no valid days return NA.
#' @noRd
.aggregate_pct_exceedance <- function(soft_ex, dates, period) {
  periods <- get_periods(dates, period)
  unique_periods <- unique(periods)
  values <- vapply(unique_periods, function(p) {
    e <- soft_ex[periods == p]
    e <- e[!is.na(e)]
    if (length(e) == 0L) return(NA_real_)
    100 * mean(e)
  }, numeric(1))
  list(periods = unique_periods, values = values)
}

#' List All Available Climate Indices
#'
#' Returns a data frame listing every index that `climatekit` can compute,
#' along with its category, unit, and a short description.
#'
#' @return A data frame with columns `index`, `category`, `unit`, and
#'   `description`.
#'
#' @export
#' @examples
#' ck_available()
ck_available <- function() {
  idx <- .index_registry()
  data.frame(
    index = vapply(idx, `[[`, character(1), "index"),
    category = vapply(idx, `[[`, character(1), "category"),
    unit = vapply(idx, `[[`, character(1), "unit"),
    description = vapply(idx, `[[`, character(1), "description"),
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

#' Get Metadata for a Climate Index
#'
#' Returns metadata (unit, category, description, reference) for a named
#' climate index.
#'
#' @param index Character string. The index name (e.g. `"frost_days"`).
#'   Use [ck_available()] to see valid names.
#'
#' @return A list with elements `index`, `category`, `unit`, `description`,
#'   and `reference`.
#'
#' @export
#' @examples
#' ck_metadata("frost_days")
ck_metadata <- function(index) {
  if (!is.character(index) || length(index) != 1) {
    cli::cli_abort("{.arg index} must be a single character string.")
  }
  registry <- .index_registry()
  names_vec <- vapply(registry, `[[`, character(1), "index")
  pos <- match(index, names_vec)
  if (is.na(pos)) {
    cli::cli_abort(
      c("Unknown index {.val {index}}.",
        "i" = "Run {.fn ck_available} to see valid index names.")
    )
  }
  registry[[pos]]
}

#' Canonical ETCCDI 27 Indices
#'
#' Returns the 27 canonical Expert Team on Climate Change Detection and
#' Indices (ETCCDI) indices as documented by Alexander et al. (2006) and
#' Zhang et al. (2011), with each row showing the canonical short code,
#' full name, input variable, unit, definition, and the corresponding
#' `climatekit` function (or `NA` where the index is not yet
#' implemented).
#'
#' Use this table to audit coverage, locate the `ck_*` function for a
#' given ETCCDI code, or filter to indices that climatekit currently
#' implements: `subset(ck_etccdi_27(), !is.na(ck_function))`.
#'
#' @return A data frame with one row per canonical ETCCDI index and
#'   columns `code`, `name`, `variable`, `unit`, `definition`,
#'   `ck_function`, and `status`.
#'
#' @references
#' Alexander, L. V. et al. (2006). Global observed changes in daily
#' climate extremes of temperature and precipitation. *Journal of
#' Geophysical Research: Atmospheres*, 111(D5).
#' \doi{10.1029/2005JD006290}.
#'
#' Zhang, X. et al. (2011). Indices for monitoring changes in extremes
#' based on daily temperature and precipitation data. *Wiley
#' Interdisciplinary Reviews: Climate Change*, 2(6), 851-870.
#' \doi{10.1002/wcc.147}.
#'
#' @export
#' @examples
#' tab <- ck_etccdi_27()
#' head(tab)
#' # Indices currently implemented in climatekit:
#' subset(tab, !is.na(ck_function))[, c("code", "ck_function")]
#' # Coverage:
#' table(tab$status)
ck_etccdi_27 <- function() {
  rows <- list(
    # Temperature: extremes (4)
    c("TXx", "Max Tmax", "Tmax", "\u00b0C",
      "Maximum value of daily maximum temperature in the period.",
      "ck_txx", "implemented"),
    c("TNx", "Max Tmin", "Tmin", "\u00b0C",
      "Maximum value of daily minimum temperature in the period (warmest night).",
      "ck_tnx", "implemented"),
    c("TXn", "Min Tmax", "Tmax", "\u00b0C",
      "Minimum value of daily maximum temperature in the period (coldest day).",
      "ck_txn", "implemented"),
    c("TNn", "Min Tmin", "Tmin", "\u00b0C",
      "Minimum value of daily minimum temperature in the period (coldest night).",
      "ck_tnn", "implemented"),

    # Temperature: thresholds (4)
    c("FD", "Frost days", "Tmin", "days",
      "Annual count of days when Tmin < 0\u00b0C.",
      "ck_frost_days", "implemented"),
    c("ID", "Ice days", "Tmax", "days",
      "Annual count of days when Tmax < 0\u00b0C.",
      "ck_ice_days", "implemented"),
    c("SU", "Summer days", "Tmax", "days",
      "Annual count of days when Tmax > 25\u00b0C.",
      "ck_summer_days", "implemented"),
    c("TR", "Tropical nights", "Tmin", "days",
      "Annual count of days when Tmin > 20\u00b0C.",
      "ck_tropical_nights", "implemented"),

    # Temperature: percentile and spell (5)
    c("TX10p", "Cool days", "Tmax", "%",
      "Percentage of days when Tmax < 10th percentile of 1961-1990 calendar-day base.",
      "ck_tx10p", "implemented"),
    c("TN10p", "Cool nights", "Tmin", "%",
      "Percentage of days when Tmin < 10th percentile of 1961-1990 calendar-day base.",
      "ck_tn10p", "implemented"),
    c("TX90p", "Warm days", "Tmax", "%",
      "Percentage of days when Tmax > 90th percentile of 1961-1990 calendar-day base.",
      "ck_tx90p", "implemented"),
    c("TN90p", "Warm nights", "Tmin", "%",
      "Percentage of days when Tmin > 90th percentile of 1961-1990 calendar-day base.",
      "ck_tn90p", "implemented"),
    c("DTR", "Diurnal range", "Tmax, Tmin", "\u00b0C",
      "Mean of (Tmax - Tmin) across the period.",
      "ck_diurnal_range", "implemented"),

    # Temperature: spells (3)
    c("WSDI", "Warm spell duration", "Tmax", "days",
      "Annual count of days in spans of >=6 consecutive days with Tmax > 90th percentile of 1961-1990 calendar-day base.",
      "ck_wsdi", "implemented"),
    c("CSDI", "Cold spell duration", "Tmin", "days",
      "Annual count of days in spans of >=6 consecutive days with Tmin < 10th percentile of 1961-1990 calendar-day base.",
      "ck_csdi", "implemented"),
    c("GSL", "Growing season length", "Tmean", "days",
      "Length of growing season as defined by the first 6-day spell with Tmean > 5\u00b0C and the first 6-day spell with Tmean < 5\u00b0C after mid-year.",
      "ck_growing_season", "implemented"),

    # Precipitation: maxima and intensity (3)
    c("RX1day", "Max 1-day precipitation", "Precip", "mm",
      "Maximum precipitation amount in a single day within the period.",
      "ck_max_1day_precip", "implemented"),
    c("RX5day", "Max 5-day precipitation", "Precip", "mm",
      "Maximum precipitation total over any 5 consecutive days within the period.",
      "ck_max_5day_precip", "implemented"),
    c("SDII", "Simple daily intensity", "Precip", "mm/day",
      "Mean precipitation amount on wet days (Precip >= 1 mm).",
      "ck_precip_intensity", "implemented"),

    # Precipitation: thresholds (3)
    c("R10mm", "Heavy precipitation days", "Precip", "days",
      "Annual count of days when precipitation >= 10 mm.",
      "ck_heavy_precip", "implemented"),
    c("R20mm", "Very heavy precipitation days", "Precip", "days",
      "Annual count of days when precipitation >= 20 mm.",
      "ck_very_heavy_precip", "implemented"),
    c("Rnnmm", "Days with precipitation above threshold", "Precip", "days",
      "Annual count of days when precipitation >= user-supplied threshold.",
      "ck_heavy_precip", "implemented"),

    # Precipitation: spells (2)
    c("CDD", "Consecutive dry days", "Precip", "days",
      "Maximum number of consecutive days with precipitation < 1 mm.",
      "ck_dry_days", "implemented"),
    c("CWD", "Consecutive wet days", "Precip", "days",
      "Maximum number of consecutive days with precipitation >= 1 mm.",
      "ck_wet_days", "implemented"),

    # Precipitation: percentile and total (3)
    c("R95p", "Very wet days total", "Precip", "mm",
      "Annual total precipitation when daily precip > 95th percentile of 1961-1990 wet-day baseline.",
      "ck_r95p", "implemented"),
    c("R99p", "Extremely wet days total", "Precip", "mm",
      "Annual total precipitation when daily precip > 99th percentile of 1961-1990 wet-day baseline.",
      "ck_r99p", "implemented"),
    c("PRCPTOT", "Annual total wet-day precip", "Precip", "mm",
      "Annual total precipitation in wet days (Precip >= 1 mm).",
      "ck_total_precip", "implemented")
  )

  m <- do.call(rbind, rows)
  data.frame(
    code = m[, 1],
    name = m[, 2],
    variable = m[, 3],
    unit = m[, 4],
    definition = m[, 5],
    ck_function = m[, 6],
    status = m[, 7],
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

#' Browse the climatekit Index Catalogue
#'
#' Filter the comprehensive climatekit index catalogue (see
#' [ck_catalogue()]) by sector, applicable standard (ETCCDI / ET-SCI /
#' agroclimatic / comfort / drought / energy), or a free-text search
#' across the function name, full name, and ETCCDI code.
#'
#' @param sector Character (length 1) or `NULL`. Filter to indices tagged
#'   with this sector. Common values: `"agriculture"`, `"health"`,
#'   `"water"`, `"energy"`. `NULL` (default) returns all rows.
#' @param standard Character (length 1) or `NULL`. Filter to indices
#'   under this standard. Common values: `"ETCCDI"`, `"ET-SCI"`,
#'   `"ETCCDI-approx"`, `"agroclimatic"`, `"comfort"`, `"drought"`,
#'   `"energy"`. `NULL` (default) returns all rows.
#' @param search Character (length 1) or `NULL`. Free-text search; rows
#'   are kept where the term appears (case-insensitive) in the function
#'   name, the full name, or the ETCCDI code.
#'
#' @return The catalogue, filtered to matching rows. Same column structure
#'   as [ck_catalogue()].
#'
#' @export
#' @examples
#' ck_browse(standard = "ETCCDI")
#' ck_browse(sector = "agriculture")
#' ck_browse(search = "heat")
ck_browse <- function(sector = NULL, standard = NULL, search = NULL) {
  cat <- ck_catalogue()

  if (!is.null(sector)) {
    if (!is.character(sector) || length(sector) != 1L) {
      cli::cli_abort("{.arg sector} must be a single character string or NULL.")
    }
    keep <- !is.na(cat$sector) & cat$sector == sector
    cat <- cat[keep, , drop = FALSE]
  }

  if (!is.null(standard)) {
    if (!is.character(standard) || length(standard) != 1L) {
      cli::cli_abort("{.arg standard} must be a single character string or NULL.")
    }
    cat <- cat[cat$standard == standard, , drop = FALSE]
  }

  if (!is.null(search)) {
    if (!is.character(search) || length(search) != 1L) {
      cli::cli_abort("{.arg search} must be a single character string or NULL.")
    }
    pat <- tolower(search)
    in_fn   <- grepl(pat, tolower(cat$ck_function), fixed = TRUE)
    in_name <- grepl(pat, tolower(cat$name),        fixed = TRUE)
    in_code <- grepl(pat, tolower(cat$code),        fixed = TRUE)
    in_code[is.na(in_code)] <- FALSE
    cat <- cat[in_fn | in_name | in_code, , drop = FALSE]
  }

  rownames(cat) <- NULL
  cat
}

#' climatekit Index Catalogue
#'
#' Returns the complete catalogue of climate indices implemented by
#' `climatekit`, with one row per `ck_*` function and columns covering
#' the canonical short code (where applicable), the full name, the
#' index family, the relevant sector, the unit, the source standard,
#' and the principal citation key.
#'
#' Use [ck_browse()] to filter by sector or standard.
#'
#' @return A data frame with columns `ck_function`, `code`, `name`,
#'   `category`, `sector`, `unit`, `standard`, and `citation_key`.
#'
#' @export
#' @examples
#' tab <- ck_catalogue()
#' head(tab)
#' # Tally indices by standard:
#' table(tab$standard)
ck_catalogue <- function() {
  rows <- list(
    # Temperature: ETCCDI canonical
    c("ck_frost_days",     "FD",     "Frost days",                              "temperature",   "agriculture", "days",       "ETCCDI",        "alexander2006global"),
    c("ck_ice_days",       "ID",     "Ice days",                                "temperature",   NA_character_, "days",       "ETCCDI",        "alexander2006global"),
    c("ck_summer_days",    "SU",     "Summer days",                             "temperature",   "health",      "days",       "ETCCDI",        "alexander2006global"),
    c("ck_tropical_nights","TR",     "Tropical nights",                         "temperature",   "health",      "days",       "ETCCDI",        "alexander2006global"),
    c("ck_txx",            "TXx",    "Max Tmax (warmest day)",                  "temperature",   NA_character_, "\u00b0C",    "ETCCDI",        "alexander2006global"),
    c("ck_tnx",            "TNx",    "Max Tmin (warmest night)",                "temperature",   NA_character_, "\u00b0C",    "ETCCDI",        "alexander2006global"),
    c("ck_txn",            "TXn",    "Min Tmax (coldest day)",                  "temperature",   NA_character_, "\u00b0C",    "ETCCDI",        "alexander2006global"),
    c("ck_tnn",            "TNn",    "Min Tmin (coldest night)",                "temperature",   NA_character_, "\u00b0C",    "ETCCDI",        "alexander2006global"),
    c("ck_tx10p",          "TX10p",  "Cool days",                               "temperature",   "health",      "%",          "ETCCDI",        "zhang2011indices"),
    c("ck_tn10p",          "TN10p",  "Cool nights",                             "temperature",   "health",      "%",          "ETCCDI",        "zhang2011indices"),
    c("ck_tx90p",          "TX90p",  "Warm days",                               "temperature",   "health",      "%",          "ETCCDI",        "zhang2011indices"),
    c("ck_tn90p",          "TN90p",  "Warm nights",                             "temperature",   "health",      "%",          "ETCCDI",        "zhang2011indices"),
    c("ck_wsdi",           "WSDI",   "Warm spell duration index",               "temperature",   "health",      "days",       "ETCCDI",        "zhang2011indices"),
    c("ck_csdi",           "CSDI",   "Cold spell duration index",               "temperature",   "health",      "days",       "ETCCDI",        "zhang2011indices"),
    c("ck_diurnal_range",  "DTR",    "Diurnal temperature range",               "temperature",   NA_character_, "\u00b0C",    "ETCCDI",        "alexander2006global"),
    c("ck_growing_season", "GSL",    "Growing season length",                   "temperature",   "agriculture", "days",       "ETCCDI",        "alexander2006global"),

    # Temperature: ETCCDI approximation
    c("ck_warm_spell",     "WSDI*",  "Warm spell days (series-quantile approx.)","temperature",  NA_character_, "days",       "ETCCDI-approx", "zhang2011indices"),

    # Temperature: ET-SCI heatwave family (TX > 90th percentile, calendar-day base)
    c("ck_hwn", "HWN", "Heatwave number",     "temperature", "health", "events",   "ET-SCI", "perkins2013measurement"),
    c("ck_hwf", "HWF", "Heatwave frequency",  "temperature", "health", "days",     "ET-SCI", "perkins2013measurement"),
    c("ck_hwd", "HWD", "Heatwave duration",   "temperature", "health", "days",     "ET-SCI", "perkins2013measurement"),
    c("ck_hwm", "HWM", "Heatwave magnitude",  "temperature", "health", "\u00b0C",  "ET-SCI", "perkins2013measurement"),
    c("ck_hwa", "HWA", "Heatwave amplitude",  "temperature", "health", "\u00b0C",  "ET-SCI", "perkins2013measurement"),

    # Temperature: ET-SCI cold-wave family (TN < 10th percentile, calendar-day base)
    c("ck_cwn", "CWN", "Cold-wave number",    "temperature", "health", "events",   "ET-SCI", "perkins2013measurement"),
    c("ck_cwf", "CWF", "Cold-wave frequency", "temperature", "health", "days",     "ET-SCI", "perkins2013measurement"),
    c("ck_cwd", "CWD", "Cold-wave duration",  "temperature", "health", "days",     "ET-SCI", "perkins2013measurement"),
    c("ck_cwm", "CWM", "Cold-wave magnitude", "temperature", "health", "\u00b0C",  "ET-SCI", "perkins2013measurement"),
    c("ck_cwa", "CWA", "Cold-wave amplitude", "temperature", "health", "\u00b0C",  "ET-SCI", "perkins2013measurement"),

    # Temperature: Excess Heat Factor (Nairn & Fawcett 2013, Bureau of Meteorology operational metric)
    c("ck_ehf", "EHF", "Excess Heat Factor",  "temperature", "health", "\u00b0C^2", "ET-SCI", "nairn2013defining"),

    # Temperature: energy and agriculture
    c("ck_heating_degree_days", NA_character_, "Heating degree days", "temperature", "energy",      "degree-days", "energy",       NA_character_),
    c("ck_cooling_degree_days", NA_character_, "Cooling degree days", "temperature", "energy",      "degree-days", "energy",       NA_character_),
    c("ck_growing_degree_days", NA_character_, "Growing degree days", "temperature", "agriculture", "degree-days", "agroclimatic", NA_character_),

    # Precipitation: ETCCDI canonical
    c("ck_total_precip",      "PRCPTOT",     "Total wet-day precipitation",        "precipitation", "water", "mm",     "ETCCDI", "alexander2006global"),
    c("ck_dry_days",          "CDD",         "Consecutive dry days",               "precipitation", "water", "days",   "ETCCDI", "alexander2006global"),
    c("ck_wet_days",          "CWD",         "Consecutive wet days",               "precipitation", "water", "days",   "ETCCDI", "alexander2006global"),
    c("ck_heavy_precip",      "R10mm/Rnnmm", "Heavy-precipitation days",           "precipitation", "water", "days",   "ETCCDI", "alexander2006global"),
    c("ck_very_heavy_precip", "R20mm",       "Very heavy precipitation days",      "precipitation", "water", "days",   "ETCCDI", "alexander2006global"),
    c("ck_max_1day_precip",   "RX1day",      "Max 1-day precipitation",            "precipitation", "water", "mm",     "ETCCDI", "alexander2006global"),
    c("ck_max_5day_precip",   "RX5day",      "Max 5-day precipitation",            "precipitation", "water", "mm",     "ETCCDI", "alexander2006global"),
    c("ck_precip_intensity",  "SDII",        "Simple daily intensity",             "precipitation", "water", "mm/day", "ETCCDI", "alexander2006global"),
    c("ck_r95p",              "R95p",        "Very wet days total",                "precipitation", "water", "mm",     "ETCCDI", "zhang2011indices"),
    c("ck_r99p",              "R99p",        "Extremely wet days total",           "precipitation", "water", "mm",     "ETCCDI", "zhang2011indices"),

    # Drought
    c("ck_spi",  NA_character_, "Standardised Precipitation Index",                 "drought", "water", "dimensionless", "drought", "mckee1993relationship"),
    c("ck_spei", NA_character_, "Standardised Precipitation-Evapotranspiration Index","drought","water","dimensionless", "drought", "vicente2010multiscalar"),
    c("ck_pet",     NA_character_, "Potential evapotranspiration (Hargreaves)",       "drought", "water", "mm", "drought", "hargreaves1985reference"),
    c("ck_pet_pm",  NA_character_, "Reference evapotranspiration (FAO-56 Penman-Monteith)", "drought", "water", "mm", "drought", "allen1998crop"),

    # Agroclimatic
    c("ck_huglin",      NA_character_, "Huglin heliothermal index",    "agroclimatic", "agriculture", "degree-days",      "agroclimatic", "huglin1978nouveau"),
    c("ck_winkler",     NA_character_, "Winkler index",                "agroclimatic", "agriculture", "degree-days",      "agroclimatic", "winkler1974general"),
    c("ck_branas",      NA_character_, "Branas hydrothermal index",    "agroclimatic", "agriculture", "mm\u00b7\u00b0C",  "agroclimatic", NA_character_),
    c("ck_first_frost", NA_character_, "First autumn frost date",      "agroclimatic", "agriculture", "day of year",      "agroclimatic", NA_character_),
    c("ck_last_frost",  NA_character_, "Last spring frost date",       "agroclimatic", "agriculture", "day of year",      "agroclimatic", NA_character_),

    # Comfort
    c("ck_heat_index",  NA_character_, "Heat index (NWS apparent temperature)", "comfort", "health", "\u00b0C",   "comfort", "rothfusz1990heat"),
    c("ck_humidex",     NA_character_, "Canadian humidex",                      "comfort", "health", "unitless",  "comfort", "masterton1979humidex"),
    c("ck_wind_chill",  NA_character_, "Wind chill",                            "comfort", "health", "\u00b0C",   "comfort", "osczevski2005new"),
    c("ck_fire_danger", NA_character_, "Fire danger proxy",                     "comfort", NA_character_, "unitless", "comfort", NA_character_)
  )

  m <- do.call(rbind, rows)
  data.frame(
    ck_function  = m[, 1],
    code         = m[, 2],
    name         = m[, 3],
    category     = m[, 4],
    sector       = m[, 5],
    unit         = m[, 6],
    standard     = m[, 7],
    citation_key = m[, 8],
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

#' Index registry (internal)
#' @noRd
.index_registry <- function() {
  list(
    # Temperature
    list(index = "frost_days", category = "temperature", unit = "days",
         description = "Count of days where Tmin < 0\u00b0C",
         reference = "ETCCDI"),
    list(index = "ice_days", category = "temperature", unit = "days",
         description = "Count of days where Tmax < 0\u00b0C",
         reference = "ETCCDI"),
    list(index = "summer_days", category = "temperature", unit = "days",
         description = "Count of days where Tmax > 25\u00b0C",
         reference = "ETCCDI"),
    list(index = "tropical_nights", category = "temperature", unit = "days",
         description = "Count of days where Tmin > 20\u00b0C",
         reference = "ETCCDI"),
    list(index = "growing_season", category = "temperature", unit = "days",
         description = "Growing season length (ETCCDI: 6-day spells of Tmean > 5\u00b0C)",
         reference = "ETCCDI"),
    list(index = "txx", category = "temperature", unit = "\u00b0C",
         description = "Annual or monthly maximum of daily Tmax (ETCCDI TXx)",
         reference = "ETCCDI"),
    list(index = "tnx", category = "temperature", unit = "\u00b0C",
         description = "Annual or monthly maximum of daily Tmin (ETCCDI TNx, warmest night)",
         reference = "ETCCDI"),
    list(index = "txn", category = "temperature", unit = "\u00b0C",
         description = "Annual or monthly minimum of daily Tmax (ETCCDI TXn, coldest day)",
         reference = "ETCCDI"),
    list(index = "tnn", category = "temperature", unit = "\u00b0C",
         description = "Annual or monthly minimum of daily Tmin (ETCCDI TNn, coldest night)",
         reference = "ETCCDI"),
    list(index = "tx10p", category = "temperature", unit = "%",
         description = "Percentage of cool days (ETCCDI TX10p, Tmax below calendar-day 10th percentile)",
         reference = "ETCCDI"),
    list(index = "tn10p", category = "temperature", unit = "%",
         description = "Percentage of cool nights (ETCCDI TN10p, Tmin below calendar-day 10th percentile)",
         reference = "ETCCDI"),
    list(index = "tx90p", category = "temperature", unit = "%",
         description = "Percentage of warm days (ETCCDI TX90p, Tmax above calendar-day 90th percentile)",
         reference = "ETCCDI"),
    list(index = "tn90p", category = "temperature", unit = "%",
         description = "Percentage of warm nights (ETCCDI TN90p, Tmin above calendar-day 90th percentile)",
         reference = "ETCCDI"),
    list(index = "heating_degree_days", category = "temperature", unit = "degree-days",
         description = "Sum of (base - Tavg) for days below base temperature",
         reference = "ASHRAE"),
    list(index = "cooling_degree_days", category = "temperature", unit = "degree-days",
         description = "Sum of (Tavg - base) for days above base temperature",
         reference = "ASHRAE"),
    list(index = "growing_degree_days", category = "temperature", unit = "degree-days",
         description = "Sum of (Tavg - base) for days above base temperature",
         reference = "McMaster & Wilhelm 1997"),
    list(index = "diurnal_range", category = "temperature", unit = "\u00b0C",
         description = "Mean daily temperature range (Tmax - Tmin)",
         reference = "ETCCDI"),
    list(index = "warm_spell", category = "temperature", unit = "days",
         description = "Warm spell days (simplified, see ck_wsdi for ETCCDI WSDI)",
         reference = "Simplified (see documentation)"),
    list(index = "wsdi", category = "temperature", unit = "days",
         description = "Warm spell duration index (ETCCDI WSDI, 1961-1990 calendar-day Tmax 90p base)",
         reference = "ETCCDI"),
    list(index = "csdi", category = "temperature", unit = "days",
         description = "Cold spell duration index (ETCCDI CSDI, 1961-1990 calendar-day Tmin 10p base)",
         reference = "ETCCDI"),

    # Precipitation
    list(index = "dry_days", category = "precipitation", unit = "days",
         description = "Maximum consecutive dry days (precip < threshold)",
         reference = "ETCCDI"),
    list(index = "wet_days", category = "precipitation", unit = "days",
         description = "Maximum consecutive wet days (precip >= threshold)",
         reference = "ETCCDI"),
    list(index = "total_precip", category = "precipitation", unit = "mm",
         description = "Total precipitation by period",
         reference = "ETCCDI"),
    list(index = "heavy_precip", category = "precipitation", unit = "days",
         description = "Count of days with precipitation >= threshold",
         reference = "ETCCDI"),
    list(index = "very_heavy_precip", category = "precipitation", unit = "days",
         description = "Count of days with precipitation >= threshold",
         reference = "ETCCDI"),
    list(index = "max_1day_precip", category = "precipitation", unit = "mm",
         description = "Maximum 1-day precipitation",
         reference = "ETCCDI"),
    list(index = "max_5day_precip", category = "precipitation", unit = "mm",
         description = "Maximum 5-day precipitation total",
         reference = "ETCCDI"),
    list(index = "precip_intensity", category = "precipitation", unit = "mm/day",
         description = "Mean precipitation on wet days (SDII)",
         reference = "ETCCDI"),
    list(index = "r95p", category = "precipitation", unit = "mm",
         description = "Annual total precipitation on days above 95th percentile of 1961-1990 wet-day baseline (ETCCDI R95p)",
         reference = "ETCCDI"),
    list(index = "r99p", category = "precipitation", unit = "mm",
         description = "Annual total precipitation on days above 99th percentile of 1961-1990 wet-day baseline (ETCCDI R99p)",
         reference = "ETCCDI"),

    # Drought
    list(index = "spi", category = "drought", unit = "dimensionless",
         description = "Standardized Precipitation Index",
         reference = "McKee et al. 1993"),
    list(index = "spei", category = "drought", unit = "dimensionless",
         description = "Standardized Precipitation-Evapotranspiration Index",
         reference = "Vicente-Serrano et al. 2010"),
    list(index = "pet", category = "drought", unit = "mm",
         description = "Potential evapotranspiration (Hargreaves method)",
         reference = "Hargreaves & Samani 1985"),

    # Agroclimatic
    list(index = "huglin", category = "agroclimatic", unit = "degree-days",
         description = "Huglin heliothermal index for viticulture",
         reference = "Huglin 1978"),
    list(index = "winkler", category = "agroclimatic", unit = "degree-days",
         description = "Winkler index (growing degree days for wine regions)",
         reference = "Amerine & Winkler 1944"),
    list(index = "branas", category = "agroclimatic", unit = "mm\u00b7\u00b0C",
         description = "Branas hydrothermal index",
         reference = "Branas et al. 1946"),
    list(index = "first_frost", category = "agroclimatic", unit = "day of year",
         description = "Date of first autumn frost (Tmin < 0\u00b0C)",
         reference = ""),
    list(index = "last_frost", category = "agroclimatic", unit = "day of year",
         description = "Date of last spring frost (Tmin < 0\u00b0C)",
         reference = ""),

    # Comfort
    list(index = "wind_chill", category = "comfort", unit = "\u00b0C",
         description = "Wind chill temperature",
         reference = "Environment Canada / NWS"),
    list(index = "heat_index", category = "comfort", unit = "\u00b0C",
         description = "Heat index (apparent temperature)",
         reference = "Rothfusz 1990"),
    list(index = "humidex", category = "comfort", unit = "unitless",
         description = "Canadian humidex",
         reference = "Masterson & Richardson 1979"),
    list(index = "fire_danger", category = "comfort", unit = "unitless",
         description = "Simplified fire danger proxy (not FWI)",
         reference = "")
  )
}

test_that("ck_compute dispatches frost_days", {
  d <- data.frame(
    dates = as.Date("2024-01-01") + 0:9,
    tmin = c(-2, 3, -1, 5, -3, 0, 2, -4, 1, -1)
  )
  result <- ck_compute(d, "frost_days")
  expect_equal(result$value, 5)
  expect_equal(result$index, "frost_days")
})

test_that("ck_compute dispatches with extra args", {
  d <- data.frame(
    dates = as.Date("2024-01-01") + 0:9,
    tavg = c(10, 15, 20, 5, 18, 22, 8, 25, 12, 30)
  )
  result <- ck_compute(d, "heating_degree_days", base = 20)
  expect_s3_class(result, "data.frame")
})

test_that("ck_compute rejects unknown index", {
  d <- data.frame(dates = as.Date("2024-01-01"), tmin = 1)
  expect_error(ck_compute(d, "nonexistent"), "Unknown index")
})

test_that("ck_compute rejects missing columns", {
  d <- data.frame(dates = as.Date("2024-01-01"), tmax = 1)
  expect_error(ck_compute(d, "frost_days"), "Missing required")
})

test_that("ck_compute works with list input", {
  d <- list(
    dates = as.Date("2024-01-01") + 0:4,
    tmin = c(-2, 3, -1, 5, -3)
  )
  result <- ck_compute(d, "frost_days")
  expect_equal(result$value, 3)
})

test_that("ck_compute dispatches comfort indices", {
  d <- list(
    tavg = c(30, 35),
    humidity = c(60, 70)
  )
  result <- ck_compute(d, "heat_index")
  expect_equal(nrow(result), 2)
})

test_that("ck_compute dispatches precipitation indices", {
  d <- data.frame(
    dates = as.Date("2024-01-01") + 0:4,
    precip = c(0, 5, 12, 0, 15)
  )
  result <- ck_compute(d, "total_precip")
  expect_equal(result$value, 32)
})

test_that("ck_available returns correct structure", {
  result <- ck_available()
  expect_s3_class(result, "data.frame")
  expect_true(all(c("index", "category", "unit", "description") %in% names(result)))
  expect_true(nrow(result) >= 28)
})

test_that("ck_metadata returns correct info", {
  result <- ck_metadata("frost_days")
  expect_true(is.list(result))
  expect_equal(result$index, "frost_days")
  expect_equal(result$category, "temperature")
})

test_that("ck_metadata rejects unknown index", {
  expect_error(ck_metadata("fake_index"), "Unknown index")
})

test_that("ck_metadata rejects non-string", {
  expect_error(ck_metadata(123), "character string")
})

test_that("ck_convert_temp C to F", {
  expect_equal(ck_convert_temp(0, "C", "F"), 32)
  expect_equal(ck_convert_temp(100, "C", "F"), 212)
})

test_that("ck_convert_temp F to C", {
  expect_equal(ck_convert_temp(32, "F", "C"), 0)
  expect_equal(ck_convert_temp(212, "F", "C"), 100)
})

test_that("ck_convert_temp C to K", {
  expect_equal(ck_convert_temp(0, "C", "K"), 273.15)
})

test_that("ck_convert_temp same unit returns unchanged", {
  expect_equal(ck_convert_temp(42, "C", "C"), 42)
})

test_that("ck_convert_temp rejects non-numeric", {
  expect_error(ck_convert_temp("a", "C", "F"), "numeric")
})

# ck_etccdi_27 audit table ----------------------------------------------------

test_that("ck_etccdi_27 returns canonical 27 rows with required columns", {
  tab <- ck_etccdi_27()
  expect_s3_class(tab, "data.frame")
  expect_equal(nrow(tab), 27)
  expect_true(all(c("code", "name", "variable", "unit",
                    "definition", "ck_function", "status") %in% names(tab)))
})

test_that("ck_etccdi_27 codes are unique and canonical", {
  tab <- ck_etccdi_27()
  expect_equal(length(unique(tab$code)), 27)
  canonical <- c("TXx", "TNx", "TXn", "TNn", "FD", "ID", "SU", "TR",
                 "TX10p", "TN10p", "TX90p", "TN90p", "DTR", "WSDI", "CSDI",
                 "GSL", "RX1day", "RX5day", "SDII", "R10mm", "R20mm",
                 "Rnnmm", "CDD", "CWD", "R95p", "R99p", "PRCPTOT")
  expect_setequal(tab$code, canonical)
})

test_that("ck_etccdi_27 implemented entries point to real functions", {
  tab <- ck_etccdi_27()
  impl <- subset(tab, status == "implemented")
  for (fn in unique(impl$ck_function)) {
    expect_true(exists(fn, mode = "function"),
                info = paste("Function not exported:", fn))
  }
})

test_that("ck_etccdi_27 status values are within allowed set", {
  tab <- ck_etccdi_27()
  expect_true(all(tab$status %in% c("implemented", "approximation", "missing")))
})

test_that("ck_etccdi_27 missing entries have NA ck_function", {
  tab <- ck_etccdi_27()
  miss <- subset(tab, status == "missing")
  expect_true(all(is.na(miss$ck_function)))
})

# Dispatch for new extreme-value indices --------------------------------------

test_that("ck_compute dispatches txx/tnx/txn/tnn", {
  d <- data.frame(
    dates = as.Date("2024-01-01") + 0:9,
    tmax = c(5, 10, 18, 12, 4, 8, 22, 3, 7, 6),
    tmin = c(-2, 3, -1, 5, -8, 0, 2, -12, 1, -1)
  )
  expect_equal(ck_compute(d, "txx")$value, 22)
  expect_equal(ck_compute(d, "tnx")$value, 5)
  expect_equal(ck_compute(d, "txn")$value, 3)
  expect_equal(ck_compute(d, "tnn")$value, -12)
})

test_that("ck_compute dispatches percentile indices", {
  set.seed(99)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  d <- data.frame(
    dates = dates,
    tmax = 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
           rnorm(length(dates)),
    tmin = 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
           rnorm(length(dates)),
    precip = pmax(rgamma(length(dates), shape = 0.4, scale = 8) - 1, 0)
  )
  expect_s3_class(
    ck_compute(d, "tx10p", ref_start = 1961L, ref_end = 1962L),
    "data.frame"
  )
  expect_s3_class(
    ck_compute(d, "tn90p", ref_start = 1961L, ref_end = 1962L),
    "data.frame"
  )
  expect_s3_class(
    ck_compute(d, "r95p", ref_start = 1961L, ref_end = 1962L),
    "data.frame"
  )
})

test_that("ck_etccdi_27 reports 27/27 implemented after Phase 2C", {
  tab <- ck_etccdi_27()
  expect_equal(sum(tab$status == "implemented"), 27L)
  expect_equal(sum(tab$status == "missing"), 0L)
})

# Index catalogue and browse -------------------------------------------------

test_that("ck_catalogue returns enriched data frame with required columns", {
  cat <- ck_catalogue()
  expect_s3_class(cat, "data.frame")
  expect_true(all(c("ck_function", "code", "name", "category", "sector",
                    "unit", "standard", "citation_key") %in% names(cat)))
  expect_gte(nrow(cat), 40L)
})

test_that("ck_catalogue ck_function entries are real exports", {
  cat <- ck_catalogue()
  for (fn in cat$ck_function) {
    expect_true(exists(fn, mode = "function"),
                info = paste("Not exported:", fn))
  }
})

test_that("ck_catalogue ETCCDI rows match ck_etccdi_27 implemented entries", {
  cat <- ck_catalogue()
  etccdi_rows <- subset(cat, standard == "ETCCDI")
  audit <- subset(ck_etccdi_27(), status == "implemented")
  # Every implemented ETCCDI ck_function should appear in the catalogue.
  expect_true(all(audit$ck_function %in% etccdi_rows$ck_function))
})

test_that("ck_browse filters by standard", {
  cat <- ck_catalogue()
  etccdi <- ck_browse(standard = "ETCCDI")
  expect_true(all(etccdi$standard == "ETCCDI"))
  expect_equal(nrow(etccdi), sum(cat$standard == "ETCCDI"))
})

test_that("ck_browse filters by sector and ignores NA-sector rows", {
  agri <- ck_browse(sector = "agriculture")
  expect_true(all(!is.na(agri$sector) & agri$sector == "agriculture"))
})

test_that("ck_browse search is case-insensitive across function/name/code", {
  hits <- ck_browse(search = "frost")
  expect_true(any(hits$ck_function == "ck_frost_days"))
  hits_code <- ck_browse(search = "TXx")
  expect_true(any(hits_code$ck_function == "ck_txx"))
})

test_that("ck_browse rejects non-string args", {
  expect_error(ck_browse(sector = 1), "character")
  expect_error(ck_browse(standard = TRUE), "character")
  expect_error(ck_browse(search = c("a", "b")), "character")
})

test_that("ck_browse with no args returns the full catalogue", {
  expect_equal(nrow(ck_browse()), nrow(ck_catalogue()))
})

#' Wind Chill Temperature
#'
#' Compute wind chill using the North American Wind Chill Index formula
#' (Environment Canada / US NWS). Valid for temperatures at or below 10 degrees C
#' and wind speeds above 4.8 km/h.
#'
#' @param tavg Numeric vector of temperatures (degrees C).
#' @param wind_speed Numeric vector of wind speeds (km/h).
#'
#' @return A data frame with columns `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' ck_wind_chill(tavg = c(-5, -10, 0), wind_speed = c(20, 30, 15))
ck_wind_chill <- function(tavg, wind_speed) {
  validate_numeric(tavg, "tavg")
  validate_numeric(wind_speed, "wind_speed")
  if (length(tavg) != length(wind_speed)) {
    cli::cli_abort("{.arg tavg} and {.arg wind_speed} must have the same length.")
  }

  # North American formula
  wc <- 13.12 + 0.6215 * tavg -
    11.37 * wind_speed^0.16 +
    0.3965 * tavg * wind_speed^0.16

  # Only valid for T <= 10C and wind > 4.8 km/h; return tavg otherwise
  invalid <- tavg > 10 | wind_speed <= 4.8
  wc[invalid] <- tavg[invalid]

  data.frame(
    value = wc,
    index = "wind_chill",
    unit = "\u00b0C",
    stringsAsFactors = FALSE
  )
}

#' Heat Index
#'
#' Compute the heat index (apparent temperature) using the Rothfusz
#' regression equation used by the US National Weather Service.
#'
#' @param tavg Numeric vector of temperatures (degrees C).
#' @param humidity Numeric vector of relative humidity (percent, 0-100).
#'
#' @return A data frame with columns `value`, `index`, and `unit`.
#'
#' @references Rothfusz, L. P. (1990). The heat index equation.
#'   NWS Technical Attachment SR 90-23.
#'
#' @export
#' @examples
#' ck_heat_index(tavg = c(30, 35, 40), humidity = c(60, 70, 50))
ck_heat_index <- function(tavg, humidity) {
  validate_numeric(tavg, "tavg")
  validate_numeric(humidity, "humidity")
  if (length(tavg) != length(humidity)) {
    cli::cli_abort("{.arg tavg} and {.arg humidity} must have the same length.")
  }

  # Convert to Fahrenheit for the NWS equation
  tf <- tavg * 9 / 5 + 32

  # Steadman's simple formula first
  hi_simple <- 0.5 * (tf + 61.0 + (tf - 68.0) * 1.2 + humidity * 0.094)

  # Use Rothfusz regression when simple formula > 80F
  hi <- ifelse(hi_simple > 80,
    -42.379 + 2.04901523 * tf + 10.14333127 * humidity -
      0.22475541 * tf * humidity - 6.83783e-03 * tf^2 -
      5.481717e-02 * humidity^2 + 1.22874e-03 * tf^2 * humidity +
      8.5282e-04 * tf * humidity^2 - 1.99e-06 * tf^2 * humidity^2,
    hi_simple
  )

  # NWS low-humidity adjustment
  low_rh <- !is.na(humidity) & !is.na(tf) & humidity < 13 & tf > 80 & tf < 112
  if (any(low_rh)) {
    adj1 <- ((13 - humidity[low_rh]) / 4) *
      sqrt(pmax((17 - abs(tf[low_rh] - 95)) / 17, 0))
    hi[low_rh] <- hi[low_rh] - adj1
  }

  # NWS high-humidity adjustment
  high_rh <- !is.na(humidity) & !is.na(tf) & humidity > 85 & tf >= 80 & tf <= 87
  if (any(high_rh)) {
    adj2 <- ((humidity[high_rh] - 85) / 10) * ((87 - tf[high_rh]) / 5)
    hi[high_rh] <- hi[high_rh] + adj2
  }

  # Convert back to Celsius
  hi_c <- (hi - 32) * 5 / 9

  data.frame(
    value = hi_c,
    index = "heat_index",
    unit = "\u00b0C",
    stringsAsFactors = FALSE
  )
}

#' Humidex
#'
#' Compute the Canadian humidex from temperature and dewpoint.
#'
#' @param tavg Numeric vector of temperatures (degrees C).
#' @param dewpoint Numeric vector of dewpoint temperatures (degrees C).
#'
#' @return A data frame with columns `value`, `index`, and `unit`.
#'
#' @references Masterson, J., & Richardson, F. A. (1979). Humidex: A method
#'   of quantifying human discomfort due to excessive heat and humidity.
#'   Environment Canada.
#'
#' @export
#' @examples
#' ck_humidex(tavg = c(30, 35), dewpoint = c(20, 25))
ck_humidex <- function(tavg, dewpoint) {
  validate_numeric(tavg, "tavg")
  validate_numeric(dewpoint, "dewpoint")
  if (length(tavg) != length(dewpoint)) {
    cli::cli_abort("{.arg tavg} and {.arg dewpoint} must have the same length.")
  }

  # Vapour pressure from dewpoint
  e <- 6.11 * exp(5417.7530 * (1 / 273.16 - 1 / (273.15 + dewpoint)))

  hx <- tavg + 5 / 9 * (e - 10)

  data.frame(
    value = hx,
    index = "humidex",
    unit = "unitless",
    stringsAsFactors = FALSE
  )
}

#' Fire Danger Index (Simplified)
#'
#' A simplified fire danger proxy based on temperature, humidity, wind speed,
#' and recent precipitation. This is NOT the Canadian Forest Fire Weather
#' Index (Van Wagner 1987); for the full FWI system, use the cffdrs package.
#'
#' @param tavg Numeric vector of temperatures (degrees C).
#' @param humidity Numeric vector of relative humidity (percent, 0-100).
#' @param wind_speed Numeric vector of wind speeds (km/h).
#' @param precip Numeric vector of daily precipitation (mm).
#'
#' @return A data frame with columns `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' ck_fire_danger(
#'   tavg = c(30, 25, 35),
#'   humidity = c(20, 40, 15),
#'   wind_speed = c(25, 10, 30),
#'   precip = c(0, 5, 0)
#' )
ck_fire_danger <- function(tavg, humidity, wind_speed, precip) {
  validate_numeric(tavg, "tavg")
  validate_numeric(humidity, "humidity")
  validate_numeric(wind_speed, "wind_speed")
  validate_numeric(precip, "precip")

  lens <- c(length(tavg), length(humidity), length(wind_speed), length(precip))
  if (length(unique(lens)) != 1) {
    cli::cli_abort("All input vectors must have the same length.")
  }

  # Simplified FWI: higher temp + wind = higher risk; higher humidity + precip = lower
  # This is a linearised approximation, not the full iterative CFFWIS
  moisture_factor <- pmax(1 - humidity / 100, 0)
  rain_dampening <- pmax(1 - precip / 10, 0)
  wind_factor <- wind_speed / 30

  fwi <- (tavg / 30) * moisture_factor * rain_dampening * (1 + wind_factor) * 10
  fwi[fwi < 0] <- 0

  data.frame(
    value = fwi,
    index = "fire_danger",
    unit = "unitless",
    stringsAsFactors = FALSE
  )
}

test_that("ck_wind_chill computes correctly", {
  result <- ck_wind_chill(tavg = c(-5, -10), wind_speed = c(20, 30))
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_true(all(result$value < c(-5, -10)))  # wind chill is colder
  expect_equal(result$index[1], "wind_chill")
})

test_that("ck_wind_chill returns tavg for warm temps", {
  result <- ck_wind_chill(tavg = c(15), wind_speed = c(20))
  expect_equal(result$value, 15)
})

test_that("ck_wind_chill returns tavg for calm winds", {
  result <- ck_wind_chill(tavg = c(-5), wind_speed = c(3))
  expect_equal(result$value, -5)
})

test_that("ck_wind_chill rejects mismatched lengths", {
  expect_error(ck_wind_chill(c(1, 2), c(10)), "same length")
})

test_that("ck_heat_index computes correctly", {
  result <- ck_heat_index(tavg = c(30, 35), humidity = c(60, 70))
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_true(all(result$value > c(30, 35)))  # heat index is hotter
  expect_equal(result$index[1], "heat_index")
})

test_that("ck_heat_index for cool temps uses simple formula", {
  result <- ck_heat_index(tavg = c(20), humidity = c(50))
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 1)
})

test_that("ck_heat_index rejects mismatched lengths", {
  expect_error(ck_heat_index(c(1, 2), c(50)), "same length")
})

test_that("ck_humidex computes correctly", {
  result <- ck_humidex(tavg = c(30, 35), dewpoint = c(20, 25))
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_true(all(result$value > c(30, 35)))  # humidex is higher
  expect_equal(result$index[1], "humidex")
})

test_that("ck_humidex rejects mismatched lengths", {
  expect_error(ck_humidex(c(1, 2), c(10)), "same length")
})

test_that("ck_fire_danger computes positive values", {
  result <- ck_fire_danger(
    tavg = c(30, 25),
    humidity = c(20, 40),
    wind_speed = c(25, 10),
    precip = c(0, 0)
  )
  expect_s3_class(result, "data.frame")
  expect_true(all(result$value > 0))
  expect_equal(result$index[1], "fire_danger")
})

test_that("ck_fire_danger dampened by rain", {
  dry <- ck_fire_danger(tavg = 30, humidity = 20, wind_speed = 25, precip = 0)
  wet <- ck_fire_danger(tavg = 30, humidity = 20, wind_speed = 25, precip = 10)
  expect_true(dry$value > wet$value)
})

test_that("ck_fire_danger rejects mismatched lengths", {
  expect_error(
    ck_fire_danger(c(1, 2), c(50), c(10, 20), c(0, 0)),
    "same length"
  )
})

# --- Reference value tests ---

test_that("wind chill reference: T=-10C, V=20 km/h", {
  # WC = 13.12 + 0.6215*(-10) - 11.37*(20^0.16) + 0.3965*(-10)*(20^0.16)
  expected <- 13.12 + 0.6215 * (-10) - 11.37 * (20^0.16) + 0.3965 * (-10) * (20^0.16)
  result <- ck_wind_chill(tavg = -10, wind_speed = 20)
  expect_equal(result$value, expected, tolerance = 0.01)
})

test_that("heat index reference: T=90F (32.2C), RH=65%", {
  # NWS reference: ~104.7F => ~40.4C (approximately)
  result <- ck_heat_index(tavg = 32.222, humidity = 65)
  # Should be substantially above input temperature
  expect_true(result$value > 35)
  expect_true(result$value < 50)
})

test_that("heat index low-humidity adjustment applies", {
  # T=95F (35C), RH=10% — should trigger low-humidity adjustment
  result <- ck_heat_index(tavg = 35, humidity = 10)
  expect_s3_class(result, "data.frame")
  expect_true(result$value > 30)
})

test_that("heat index high-humidity adjustment applies", {
  # T=82F (~27.8C), RH=90% — should trigger high-humidity adjustment
  result <- ck_heat_index(tavg = 27.8, humidity = 90)
  expect_s3_class(result, "data.frame")
})

test_that("humidex reference: T=30C, Td=20C", {
  # e = 6.11 * exp(5417.7530 * (1/273.16 - 1/293.15))
  # humidex = 30 + 5/9 * (e - 10)
  e <- 6.11 * exp(5417.7530 * (1 / 273.16 - 1 / 293.15))
  expected <- 30 + 5 / 9 * (e - 10)
  result <- ck_humidex(tavg = 30, dewpoint = 20)
  expect_equal(result$value, expected, tolerance = 0.1)
})

test_that("ck_compute dispatches fire_danger", {
  d <- list(
    tavg = c(30, 25),
    humidity = c(20, 40),
    wind_speed = c(25, 10),
    precip = c(0, 0)
  )
  result <- ck_compute(d, "fire_danger")
  expect_equal(result$index[1], "fire_danger")
})

# ET-SCI heatwave and cold-wave family indices.
#
# Definitions follow the Expert Team on Sector-specific Climate Indices.
# A heat wave is a period of at least three consecutive days with daily
# maximum temperature exceeding the 90th percentile of the calendar-day
# distribution from a reference period (default 1961-1990). The cold-wave
# dual uses daily minimum temperature below the 10th percentile of the
# calendar-day base. Each heatwave / cold-wave family produces five
# annual statistics:
#
# - HWN / CWN: number of events
# - HWF / CWF: frequency (total days in events)
# - HWD / CWD: duration (longest event in days)
# - HWM / CWM: magnitude (mean excess of TX over the threshold, or
#   threshold minus TN, on event days)
# - HWA / CWA: amplitude (peak excess on event days)
#
# Note: `ck_cwd` here is ET-SCI Cold Wave Duration; the ETCCDI CWD
# (Consecutive Wet Days) is a different index and lives in
# precipitation.R as `ck_wet_days`.

#' Per-year heatwave / cold-wave family statistics
#'
#' Returns one row per analysis year with columns:
#'   year, n (count), f (total days in events), d (longest event),
#'   m_excess, a_excess, m_value, a_extreme.
#'
#' `m_excess` and `a_excess` are mean and peak of (value - threshold)
#' for `op = ">"` heatwave or (threshold - value) for `op = "<"` cold
#' wave (always non-negative). `m_value` is the mean of raw values on
#' event days. `a_extreme` is the max raw value on event days for `op
#' = ">"` and the min raw value for `op = "<"`. NA for years with no
#' qualifying events.
#' @noRd
.spell_family_stats <- function(values, dates, ref_start, ref_end,
                                percentile, op, min_spell) {
  thresholds <- .calendar_day_percentile(values, dates, percentile,
                                         ref_start, ref_end, window = 5L)
  doy <- as.integer(format(dates, "%j"))
  threshold_per_day <- thresholds[doy]

  if (identical(op, ">")) {
    matches <- !is.na(values) & values > threshold_per_day
    excess <- values - threshold_per_day
  } else if (identical(op, "<")) {
    matches <- !is.na(values) & values < threshold_per_day
    excess <- threshold_per_day - values
  } else {
    cli::cli_abort("Internal: unsupported {.arg op}.")
  }

  in_spell <- .find_spells(matches, as.integer(min_spell))

  years <- as.integer(format(dates, "%Y"))
  unique_years <- unique(years)

  rows <- lapply(unique_years, function(yr) {
    sel <- years == yr
    in_y <- in_spell[sel]
    ex_y <- excess[sel]
    val_y <- values[sel]

    if (!any(in_y, na.rm = TRUE)) {
      return(c(yr, 0L, 0L, 0L,
               NA_real_, NA_real_, NA_real_, NA_real_))
    }

    r <- rle(in_y)
    seg_lens <- r$lengths[!is.na(r$values) & r$values]
    n_events <- length(seg_lens)
    f_total  <- sum(seg_lens)
    d_max    <- max(seg_lens)

    keep <- in_y & !is.na(in_y) & !is.na(ex_y) & !is.na(val_y)
    ex_in  <- ex_y[keep]
    val_in <- val_y[keep]

    if (length(ex_in) > 0L) {
      m_excess  <- mean(ex_in)
      a_excess  <- max(ex_in)
      m_value   <- mean(val_in)
      a_extreme <- if (identical(op, ">")) max(val_in) else min(val_in)
    } else {
      m_excess  <- NA_real_
      a_excess  <- NA_real_
      m_value   <- NA_real_
      a_extreme <- NA_real_
    }

    c(yr, n_events, f_total, d_max,
      m_excess, a_excess, m_value, a_extreme)
  })

  m <- do.call(rbind, rows)
  data.frame(
    year      = as.integer(m[, 1]),
    n         = as.integer(m[, 2]),
    f         = as.integer(m[, 3]),
    d         = as.integer(m[, 4]),
    m_excess  = m[, 5],
    a_excess  = m[, 6],
    m_value   = m[, 7],
    a_extreme = m[, 8],
    stringsAsFactors = FALSE
  )
}

#' Validate the spell-family `mode` argument
#' @noRd
.validate_spell_mode <- function(mode) {
  match.arg(mode, c("excess", "absolute"))
}

# Heatwave family (TX > 90th percentile, calendar-day base) -----------------

#' Heatwave Number (HWN)
#'
#' ET-SCI heatwave family index. Annual count of distinct heatwave events,
#' where a heatwave is a span of at least three consecutive days with
#' daily Tmax above the 90th percentile of the calendar-day distribution
#' from a reference period (default 1961 to 1990).
#'
#' Single-threshold definition (TX-only). For the dual-threshold
#' Perkins-Alexander variant (TX and TN both above 90th percentile) see
#' `climpact`.
#'
#' @inheritParams ck_tx10p
#' @param min_spell Integer. Minimum spell length in days (default 3,
#'   the ET-SCI standard).
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_hwn(tmax, dates))
ck_hwn <- function(tmax, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L) {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  stats <- .spell_family_stats(tmax, dates, ref_start, ref_end,
                               percentile = 0.90, op = ">",
                               min_spell = min_spell)
  build_result(as.character(stats$year), as.numeric(stats$n),
               "hwn", "events", "annual")
}

#' Heatwave Frequency (HWF)
#'
#' ET-SCI heatwave family index. Annual total number of days inside any
#' heatwave event (see [ck_hwn()] for the heatwave definition).
#'
#' @inheritParams ck_hwn
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_hwf(tmax, dates))
ck_hwf <- function(tmax, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L) {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  stats <- .spell_family_stats(tmax, dates, ref_start, ref_end,
                               percentile = 0.90, op = ">",
                               min_spell = min_spell)
  build_result(as.character(stats$year), as.numeric(stats$f),
               "hwf", "days", "annual")
}

#' Heatwave Duration (HWD)
#'
#' ET-SCI heatwave family index. Length in days of the longest heatwave
#' event in each year (see [ck_hwn()] for the heatwave definition).
#'
#' @inheritParams ck_hwn
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_hwd(tmax, dates))
ck_hwd <- function(tmax, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L) {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  stats <- .spell_family_stats(tmax, dates, ref_start, ref_end,
                               percentile = 0.90, op = ">",
                               min_spell = min_spell)
  build_result(as.character(stats$year), as.numeric(stats$d),
               "hwd", "days", "annual")
}

#' Heatwave Magnitude (HWM)
#'
#' ET-SCI heatwave family index. Reports the mean magnitude of daily
#' Tmax across all heatwave days in the year. `mode = "excess"`
#' (default) gives the mean of (Tmax - threshold), matching the
#' ET-SCI / climpact convention. `mode = "absolute"` gives the mean
#' raw Tmax across heatwave days, matching Perkins-Alexander (2013).
#' Returns `NA` for years with no heatwaves.
#'
#' @inheritParams ck_hwn
#' @param mode One of `"excess"` (default) or `"absolute"`. See details.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_hwm(tmax, dates))
ck_hwm <- function(tmax, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L, mode = c("excess", "absolute")) {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  mode <- .validate_spell_mode(mode)
  stats <- .spell_family_stats(tmax, dates, ref_start, ref_end,
                               percentile = 0.90, op = ">",
                               min_spell = min_spell)
  vals <- if (mode == "excess") stats$m_excess else stats$m_value
  build_result(as.character(stats$year), vals,
               "hwm", "\u00b0C", "annual")
}

#' Heatwave Amplitude (HWA)
#'
#' ET-SCI heatwave family index. Reports the peak magnitude of daily
#' Tmax across all heatwave days in the year. `mode = "excess"`
#' (default) gives the maximum of (Tmax - threshold). `mode =
#' "absolute"` gives the maximum raw Tmax across heatwave days
#' (matching Perkins-Alexander 2013). Returns `NA` for years with
#' no heatwaves.
#'
#' @inheritParams ck_hwn
#' @param mode One of `"excess"` (default) or `"absolute"`. See details.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_hwa(tmax, dates))
ck_hwa <- function(tmax, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L, mode = c("excess", "absolute")) {
  validate_numeric(tmax, "tmax")
  validate_dates(dates, length(tmax))
  mode <- .validate_spell_mode(mode)
  stats <- .spell_family_stats(tmax, dates, ref_start, ref_end,
                               percentile = 0.90, op = ">",
                               min_spell = min_spell)
  vals <- if (mode == "excess") stats$a_excess else stats$a_extreme
  build_result(as.character(stats$year), vals,
               "hwa", "\u00b0C", "annual")
}

# Cold-wave family (TN < 10th percentile, calendar-day base) ----------------

#' Cold-Wave Number (CWN)
#'
#' ET-SCI cold-wave family index. Annual count of distinct cold-wave
#' events, where a cold wave is a span of at least three consecutive
#' days with daily Tmin below the 10th percentile of the calendar-day
#' distribution from a reference period (default 1961 to 1990).
#'
#' @inheritParams ck_tn10p
#' @param min_spell Integer. Minimum spell length in days (default 3,
#'   the ET-SCI standard).
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_cwn(tmin, dates))
ck_cwn <- function(tmin, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  stats <- .spell_family_stats(tmin, dates, ref_start, ref_end,
                               percentile = 0.10, op = "<",
                               min_spell = min_spell)
  build_result(as.character(stats$year), as.numeric(stats$n),
               "cwn", "events", "annual")
}

#' Cold-Wave Frequency (CWF)
#'
#' ET-SCI cold-wave family index. Annual total number of days inside any
#' cold-wave event (see [ck_cwn()] for the cold-wave definition).
#'
#' @inheritParams ck_cwn
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_cwf(tmin, dates))
ck_cwf <- function(tmin, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  stats <- .spell_family_stats(tmin, dates, ref_start, ref_end,
                               percentile = 0.10, op = "<",
                               min_spell = min_spell)
  build_result(as.character(stats$year), as.numeric(stats$f),
               "cwf", "days", "annual")
}

#' Cold-Wave Duration (CWD, ET-SCI)
#'
#' ET-SCI cold-wave family index. Length in days of the longest
#' cold-wave event in each year (see [ck_cwn()] for the cold-wave
#' definition).
#'
#' Note: the same letters CWD also denote the ETCCDI **Consecutive Wet
#' Days** precipitation index, which is unrelated and is implemented in
#' [ck_wet_days()]. These are two different indices that share an
#' acronym in the climate-extremes literature.
#'
#' @inheritParams ck_cwn
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_cwd(tmin, dates))
ck_cwd <- function(tmin, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  stats <- .spell_family_stats(tmin, dates, ref_start, ref_end,
                               percentile = 0.10, op = "<",
                               min_spell = min_spell)
  build_result(as.character(stats$year), as.numeric(stats$d),
               "cwd", "days", "annual")
}

#' Cold-Wave Magnitude (CWM)
#'
#' ET-SCI cold-wave family index. `mode = "excess"` (default) returns
#' the mean of (threshold - Tmin) across cold-wave days, expressed as
#' a positive magnitude. `mode = "absolute"` returns the mean raw
#' Tmin across cold-wave days. Returns `NA` for years with no cold
#' waves.
#'
#' @inheritParams ck_cwn
#' @param mode One of `"excess"` (default) or `"absolute"`. See details.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_cwm(tmin, dates))
ck_cwm <- function(tmin, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L, mode = c("excess", "absolute")) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  mode <- .validate_spell_mode(mode)
  stats <- .spell_family_stats(tmin, dates, ref_start, ref_end,
                               percentile = 0.10, op = "<",
                               min_spell = min_spell)
  vals <- if (mode == "excess") stats$m_excess else stats$m_value
  build_result(as.character(stats$year), vals,
               "cwm", "\u00b0C", "annual")
}

#' Cold-Wave Amplitude (CWA)
#'
#' ET-SCI cold-wave family index. `mode = "excess"` (default) returns
#' the peak (threshold - Tmin) across cold-wave days, expressed as a
#' positive magnitude. `mode = "absolute"` returns the minimum raw
#' Tmin across cold-wave days (the coldest event-day value). Returns
#' `NA` for years with no cold waves.
#'
#' @inheritParams ck_cwn
#' @param mode One of `"excess"` (default) or `"absolute"`. See details.
#'
#' @return A data frame with columns `period`, `value`, `index`, and `unit`.
#'
#' @export
#' @examples
#' set.seed(1)
#' dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
#' tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
#'         rnorm(length(dates))
#' tail(ck_cwa(tmin, dates))
ck_cwa <- function(tmin, dates, ref_start = 1961L, ref_end = 1990L,
                   min_spell = 3L, mode = c("excess", "absolute")) {
  validate_numeric(tmin, "tmin")
  validate_dates(dates, length(tmin))
  mode <- .validate_spell_mode(mode)
  stats <- .spell_family_stats(tmin, dates, ref_start, ref_end,
                               percentile = 0.10, op = "<",
                               min_spell = min_spell)
  vals <- if (mode == "excess") stats$a_excess else stats$a_extreme
  build_result(as.character(stats$year), vals,
               "cwa", "\u00b0C", "annual")
}

# ET-SCI heatwave / cold-wave family tests.
# Reference period 1961-1961 (one year) keeps thresholds out of the
# analysis years so injected events show up cleanly.

make_tmax_two_years <- function(seed = 31) {
  set.seed(seed)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  list(dates = dates, tmax = tmax)
}

make_tmin_two_years <- function(seed = 32) {
  set.seed(seed)
  dates <- seq(as.Date("1961-01-01"), as.Date("1962-12-31"), by = "day")
  tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "%j")) / 365) +
          rnorm(length(dates))
  list(dates = dates, tmin = tmin)
}

# Heatwave family ------------------------------------------------------------

test_that("heatwave family functions return sane structure and units", {
  d <- make_tmax_two_years()
  for (fn in c("ck_hwn", "ck_hwf", "ck_hwd", "ck_hwm", "ck_hwa")) {
    result <- do.call(fn, list(d$tmax, d$dates,
                               ref_start = 1961L, ref_end = 1961L))
    expect_s3_class(result, "data.frame")
    expect_true(all(c("period", "value", "index", "unit") %in% names(result)),
                label = fn)
  }
})

test_that("ck_hwn detects injected 7-day heatwave in 1962", {
  d <- make_tmax_two_years(seed = 41)
  hot <- which(d$dates >= as.Date("1962-07-15") & d$dates <= as.Date("1962-07-21"))
  d$tmax[hot] <- d$tmax[hot] + 25  # well above any reasonable threshold
  result <- ck_hwn(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L)
  yr_1962 <- result$value[format(result$period, "%Y") == "1962"]
  expect_gte(yr_1962, 1L)
})

test_that("HWF >= HWD >= 0 (frequency dominates duration)", {
  d <- make_tmax_two_years(seed = 42)
  hot <- which(d$dates >= as.Date("1962-07-15") & d$dates <= as.Date("1962-07-21"))
  d$tmax[hot] <- d$tmax[hot] + 25
  hwf <- ck_hwf(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L)
  hwd <- ck_hwd(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L)
  expect_true(all(hwf$value >= hwd$value))
  expect_true(all(hwd$value >= 0))
})

test_that("HWA >= HWM (peak excess >= mean excess)", {
  d <- make_tmax_two_years(seed = 43)
  hot <- which(d$dates >= as.Date("1962-07-15") & d$dates <= as.Date("1962-07-21"))
  d$tmax[hot] <- d$tmax[hot] + 25
  hwm <- ck_hwm(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L)
  hwa <- ck_hwa(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L)
  yr <- format(hwm$period, "%Y") == "1962"
  expect_true(hwa$value[yr] >= hwm$value[yr])
})

test_that("HWM and HWA return NA for years with no heatwaves", {
  d <- make_tmax_two_years(seed = 44)
  # No injection: noisy data shouldn't reliably produce 3-day spells in 1962
  hwm <- ck_hwm(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L)
  hwa <- ck_hwa(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L)
  # Where HWN is 0, HWM/HWA should be NA
  hwn <- ck_hwn(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L)
  zero_years <- hwn$value == 0
  if (any(zero_years)) {
    expect_true(all(is.na(hwm$value[zero_years])))
    expect_true(all(is.na(hwa$value[zero_years])))
  }
})

# Cold-wave family -----------------------------------------------------------

test_that("cold-wave family functions return sane structure", {
  d <- make_tmin_two_years()
  for (fn in c("ck_cwn", "ck_cwf", "ck_cwd", "ck_cwm", "ck_cwa")) {
    result <- do.call(fn, list(d$tmin, d$dates,
                               ref_start = 1961L, ref_end = 1961L))
    expect_s3_class(result, "data.frame")
  }
})

test_that("ck_cwn detects injected cold spell", {
  d <- make_tmin_two_years(seed = 51)
  cold <- which(d$dates >= as.Date("1962-02-01") & d$dates <= as.Date("1962-02-07"))
  d$tmin[cold] <- d$tmin[cold] - 25
  result <- ck_cwn(d$tmin, d$dates, ref_start = 1961L, ref_end = 1961L)
  yr_1962 <- result$value[format(result$period, "%Y") == "1962"]
  expect_gte(yr_1962, 1L)
})

test_that("cold-wave magnitudes are positive (threshold - tmin)", {
  d <- make_tmin_two_years(seed = 52)
  cold <- which(d$dates >= as.Date("1962-02-01") & d$dates <= as.Date("1962-02-07"))
  d$tmin[cold] <- d$tmin[cold] - 25
  cwm <- ck_cwm(d$tmin, d$dates, ref_start = 1961L, ref_end = 1961L)
  cwa <- ck_cwa(d$tmin, d$dates, ref_start = 1961L, ref_end = 1961L)
  yr <- format(cwm$period, "%Y") == "1962"
  expect_true(cwm$value[yr] > 0)
  expect_true(cwa$value[yr] >= cwm$value[yr])
})

# Argument validation --------------------------------------------------------

test_that("heatwave / cold-wave functions reject mismatched lengths", {
  expect_error(ck_hwn(c(1, 2), as.Date("1961-01-01")), "same length")
  expect_error(ck_cwa(c(1, 2), as.Date("1961-01-01")), "same length")
})

test_that("heatwave / cold-wave functions error on missing reference data", {
  dates <- seq(as.Date("2000-01-01"), as.Date("2000-12-31"), by = "day")
  tmax <- rnorm(length(dates), 15, 5)
  tmin <- rnorm(length(dates), 5, 5)
  expect_error(ck_hwn(tmax, dates, ref_start = 1961L, ref_end = 1990L),
               "reference period")
  expect_error(ck_cwn(tmin, dates, ref_start = 1961L, ref_end = 1990L),
               "reference period")
})

# Excess vs absolute mode (Phase B) ------------------------------------------

test_that("ck_hwm / ck_hwa absolute mode returns raw Tmax", {
  d <- make_tmax_two_years(seed = 61)
  hot <- which(d$dates >= as.Date("1962-07-15") & d$dates <= as.Date("1962-07-21"))
  d$tmax[hot] <- d$tmax[hot] + 25  # injected heatwave
  hwm_ex  <- ck_hwm(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L,
                    mode = "excess")
  hwm_abs <- ck_hwm(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L,
                    mode = "absolute")
  hwa_ex  <- ck_hwa(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L,
                    mode = "excess")
  hwa_abs <- ck_hwa(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L,
                    mode = "absolute")
  yr <- format(hwm_ex$period, "%Y") == "1962"
  # Absolute Tmax must be larger than the excess (absolute = excess + threshold).
  expect_gt(hwm_abs$value[yr], hwm_ex$value[yr])
  expect_gt(hwa_abs$value[yr], hwa_ex$value[yr])
})

test_that("ck_cwm / ck_cwa absolute mode returns raw Tmin", {
  d <- make_tmin_two_years(seed = 62)
  cold <- which(d$dates >= as.Date("1962-02-01") & d$dates <= as.Date("1962-02-07"))
  d$tmin[cold] <- d$tmin[cold] - 25
  cwm_ex  <- ck_cwm(d$tmin, d$dates, ref_start = 1961L, ref_end = 1961L,
                    mode = "excess")
  cwm_abs <- ck_cwm(d$tmin, d$dates, ref_start = 1961L, ref_end = 1961L,
                    mode = "absolute")
  cwa_ex  <- ck_cwa(d$tmin, d$dates, ref_start = 1961L, ref_end = 1961L,
                    mode = "excess")
  cwa_abs <- ck_cwa(d$tmin, d$dates, ref_start = 1961L, ref_end = 1961L,
                    mode = "absolute")
  yr <- format(cwm_ex$period, "%Y") == "1962"
  # excess is positive; absolute Tmin during deep cold should be negative or much lower
  expect_lt(cwm_abs$value[yr], cwm_ex$value[yr])
  expect_lt(cwa_abs$value[yr], cwa_ex$value[yr])
})

test_that("invalid mode is rejected", {
  d <- make_tmax_two_years(seed = 63)
  expect_error(
    ck_hwm(d$tmax, d$dates, ref_start = 1961L, ref_end = 1961L,
           mode = "bogus"),
    "should be one of"
  )
})

---
title: "Migrating from climdex.pcic"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Migrating from climdex.pcic}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

`climdex.pcic` (Pacific Climate Impacts Consortium) was for many years the standard R implementation of the ETCCDI canonical 27 indices. It was archived from CRAN in 2023. `climatekit` covers the same ETCCDI 27 plus the ET-SCI heatwave family, on a tidy data-frame interface with no compiled code. This vignette walks through what changes when you migrate.

## Function-name crosswalk

Every ETCCDI 27 index in `climdex.pcic` maps to a `climatekit` function. The function names mirror the canonical ETCCDI codes where possible.

| ETCCDI code | climdex.pcic                      | climatekit              |
| :---------- | :-------------------------------- | :---------------------- |
| FD          | `climdex.fd()`                    | `ck_frost_days()`       |
| ID          | `climdex.id()`                    | `ck_ice_days()`         |
| SU          | `climdex.su()`                    | `ck_summer_days()`      |
| TR          | `climdex.tr()`                    | `ck_tropical_nights()`  |
| TXx         | `climdex.txx()`                   | `ck_txx()`              |
| TNx         | `climdex.tnx()`                   | `ck_tnx()`              |
| TXn         | `climdex.txn()`                   | `ck_txn()`              |
| TNn         | `climdex.tnn()`                   | `ck_tnn()`              |
| TX10p       | `climdex.tx10p()`                 | `ck_tx10p()`            |
| TN10p       | `climdex.tn10p()`                 | `ck_tn10p()`            |
| TX90p       | `climdex.tx90p()`                 | `ck_tx90p()`            |
| TN90p       | `climdex.tn90p()`                 | `ck_tn90p()`            |
| WSDI        | `climdex.wsdi()`                  | `ck_wsdi()`             |
| CSDI        | `climdex.csdi()`                  | `ck_csdi()`             |
| DTR         | `climdex.dtr()`                   | `ck_diurnal_range()`    |
| GSL         | `climdex.gsl()`                   | `ck_growing_season()`   |
| RX1day      | `climdex.rx1day()`                | `ck_max_1day_precip()`  |
| RX5day      | `climdex.rx5day()`                | `ck_max_5day_precip()`  |
| SDII        | `climdex.sdii()`                  | `ck_precip_intensity()` |
| R10mm       | `climdex.r10mm()`                 | `ck_heavy_precip()` (default `threshold = 10`) |
| R20mm       | `climdex.r20mm()`                 | `ck_very_heavy_precip()` (default `threshold = 20`) |
| Rnnmm       | `climdex.rnnmm(threshold = nn)`   | `ck_heavy_precip(threshold = nn)` |
| CDD         | `climdex.cdd()`                   | `ck_dry_days()`         |
| CWD         | `climdex.cwd()`                   | `ck_wet_days()`         |
| R95p        | `climdex.r95p()` / `climdex.r95ptot()` | `ck_r95p()`        |
| R99p        | `climdex.r99p()` / `climdex.r99ptot()` | `ck_r99p()`        |
| PRCPTOT     | `climdex.prcptot()`               | `ck_total_precip()`     |

Beyond ETCCDI 27, `climatekit` also exposes the ET-SCI heatwave family (`ck_hwn()`, `ck_hwf()`, `ck_hwd()`, `ck_hwm()`, `ck_hwa()`) and the cold-wave duals (`ck_cwn()` etc.), agroclimatic indices (Huglin, Winkler, Branas), drought indices (SPI, SPEI), and comfort indices (heat index, humidex, wind chill).

`ck_etccdi_27()` returns an audit table that maps every ETCCDI code to its `climatekit` function in one place.

## Interface shifts

**Inputs.** `climdex.pcic` builds a `climdexInput` object from your daily temperature and precipitation series and then runs index functions against that object. `climatekit` takes the daily numeric vectors and a `Date` vector directly. There is no preliminary input object.

```{r, eval = FALSE}
# climdex.pcic
ci <- climdexInput.raw(
  tmax, tmin, precip, tmax.dates, tmin.dates, prec.dates,
  base.range = c(1961, 1990)
)
fd <- climdex.fd(ci)

# climatekit
fd <- ck_frost_days(tmin, dates)
```

**Outputs.** `climdex.pcic` returns a numeric vector indexed by year (or month). `climatekit` returns a tidy data frame with `period`, `value`, `index`, and `unit` columns:

```{r, eval = FALSE}
ck_frost_days(tmin, dates)
#> # period       value index        unit
#> # 2020-01-01      72 frost_days   days
#> # 2021-01-01      68 frost_days   days
#> # ...
```

The tidy shape composes directly with `rbind()`, `ggplot2`, and downstream tabulation. No reshaping needed.

**Reference period.** The percentile-based indices (`ck_tx10p()`, `ck_tn10p()`, `ck_tx90p()`, `ck_tn90p()`, `ck_r95p()`, `ck_r99p()`, `ck_wsdi()`, `ck_csdi()`, `ck_hwn()` and family) use `ref_start = 1961L`, `ref_end = 1990L` by default. Pass other values to override:

```{r, eval = FALSE}
ck_tx90p(tmax, dates, ref_start = 1981L, ref_end = 2010L)
```

**In-base bootstrap.** `climdex.pcic` implements the Zhang et al. (2005) in-base bootstrap to remove the self-inclusion bias for years inside the reference period. `climatekit` does **not** implement the bootstrap in v0.2.0. Years inside the reference period therefore show a small bias (this is documented in each function's roxygen). For climate-change attribution work, restrict interpretation to years outside the reference window, or apply the bootstrap externally. Future versions may add it.

**Calendar-day window.** Both packages use a 5-day window centred on each day of year for percentile estimation, pooled across the reference period.

**Spell-counting conventions.** Both packages use 6 days minimum for WSDI / CSDI. `climatekit`'s ET-SCI heatwave family uses 3 days minimum (`min_spell = 3L`), matching ET-SCI / `climpact`.

## Sanity-check workflow

When migrating a real station record, sanity-check the new outputs against your previous `climdex.pcic` runs at the annual scale:

```{r, eval = FALSE}
# Old climdex.pcic outputs
old_fd <- climdex.fd(ci)
old_txx <- climdex.txx(ci)

# New climatekit outputs
new_fd  <- ck_frost_days(tmin, dates)
new_txx <- ck_txx(tmax, dates)

# Compare; expect agreement up to NA-handling and base-period self-inclusion
all.equal(old_fd, new_fd$value)
all.equal(old_txx, new_txx$value)
```

Most ETCCDI indices are NA-handling-tolerant by construction. Differences typically come from:

1. **Default missing-day handling.** `climdex.pcic` has more elaborate rules (e.g. min number of valid days per month) that `climatekit` does not enforce in v0.2.0.
2. **In-base bootstrap.** As noted above, percentile indices for years 1961-1990 will differ slightly.
3. **Calendar leap-day handling.** Both packages handle DOY 60 (29 February) without special-casing in v0.2.0.

For all other ETCCDI 27 indices, `climatekit` and `climdex.pcic` should agree to within numerical noise on the same input.

## See also

- [`ck_etccdi_27()`](../reference/ck_etccdi_27.html): canonical ETCCDI 27 audit table.
- [`ck_catalogue()`](../reference/ck_catalogue.html) and [`ck_browse()`](../reference/ck_browse.html): full implementation catalogue and filter.
- Alexander, L. V. et al. (2006). Global observed changes in daily climate extremes of temperature and precipitation. *Journal of Geophysical Research: Atmospheres*, 111(D5). \doi{10.1029/2005JD006290}.
- Zhang, X. et al. (2011). Indices for monitoring changes in extremes based on daily temperature and precipitation data. *WIREs Climate Change*, 2(6), 851-870. \doi{10.1002/wcc.147}.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temperature.R
\name{ck_tn90p}
\alias{ck_tn90p}
\title{Percentage of Warm Nights (TN90p)}
\usage{
ck_tn90p(
  tmin,
  dates,
  ref_start = 1961L,
  ref_end = 1990L,
  period = "annual",
  bootstrap = FALSE
)
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmax}. Must contain
data covering the reference period.}

\item{ref_start, ref_end}{Integer. Reference period boundary years
(inclusive). Defaults to 1961 and 1990.}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}

\item{bootstrap}{Logical. If \code{TRUE}, apply the Zhang (2005) in-base
bootstrap correction. Default \code{FALSE} for backward compatibility
and speed.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
ETCCDI canonical index TN90p. Percentage of days where daily Tmin
exceeds the 90th percentile of the calendar-day distribution from a
reference period (default 1961 to 1990). Computation follows the
same convention as \code{\link[=ck_tx10p]{ck_tx10p()}} and supports the same \code{bootstrap}
argument.
}
\examples{
set.seed(1)
dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "\%j")) / 365) +
        rnorm(length(dates))
tail(ck_tn90p(tmin, dates))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute.R
\name{ck_compute}
\alias{ck_compute}
\title{Compute a Climate Index by Name}
\usage{
ck_compute(data, index, ...)
}
\arguments{
\item{data}{A named list or data frame containing the required input
vectors. Column names should match function argument names (e.g.
\code{tmin}, \code{tmax}, \code{precip}, \code{dates}).}

\item{index}{Character. Name of the index to compute (e.g.
\code{"frost_days"}). Use \code{\link[=ck_available]{ck_available()}} to see valid names.}

\item{...}{Additional arguments passed to the underlying function (e.g.
\code{period}, \code{threshold}, \code{base}).}
}
\value{
A data frame as returned by the underlying \verb{ck_*} function.
}
\description{
A generic dispatcher that calls the appropriate \verb{ck_*} function based on
a string index name. Useful for programmatic workflows where the index
is selected at runtime.
}
\examples{
d <- data.frame(
  dates = as.Date("2024-01-01") + 0:9,
  tmin = c(-2, 3, -1, 5, -3, 0, 2, -4, 1, -1)
)
ck_compute(d, "frost_days")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/agroclimatic.R
\name{ck_last_frost}
\alias{ck_last_frost}
\title{Last Frost Date}
\usage{
ck_last_frost(tmin, dates, lat = 50)
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmin}.}

\item{lat}{Numeric. Latitude in decimal degrees, used to select the
hemisphere convention. Default 50 (Northern Hemisphere).}
}
\value{
A data frame with columns \code{period}, \code{value} (day of year),
\code{date} (the frost date), \code{index}, and \code{unit}.
}
\description{
Day of year of the last spring frost (Tmin < 0 degrees C) in each
year. Hemisphere is selected by \code{lat}: in the Northern Hemisphere
the search runs up to July 1 (DOY 183); in the Southern Hemisphere
up to October 1 (DOY 274), matching the spring boundary for each.
}
\examples{
dates <- seq(as.Date("2024-01-01"), as.Date("2024-06-30"), by = "day")
set.seed(42)
tmin <- -10 + seq_along(dates) * 0.12 + rnorm(length(dates), sd = 3)
ck_last_frost(tmin, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/heatwave.R
\name{ck_cwa}
\alias{ck_cwa}
\title{Cold-Wave Amplitude (CWA)}
\usage{
ck_cwa(
  tmin,
  dates,
  ref_start = 1961L,
  ref_end = 1990L,
  min_spell = 3L,
  mode = c("excess", "absolute")
)
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmax}. Must contain
data covering the reference period.}

\item{ref_start, ref_end}{Integer. Reference period boundary years
(inclusive). Defaults to 1961 and 1990.}

\item{min_spell}{Integer. Minimum spell length in days (default 3,
the ET-SCI standard).}

\item{mode}{One of \code{"excess"} (default) or \code{"absolute"}. See details.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
ET-SCI cold-wave family index. \code{mode = "excess"} (default) returns
the peak (threshold - Tmin) across cold-wave days, expressed as a
positive magnitude. \code{mode = "absolute"} returns the minimum raw
Tmin across cold-wave days (the coldest event-day value). Returns
\code{NA} for years with no cold waves.
}
\examples{
set.seed(1)
dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "\%j")) / 365) +
        rnorm(length(dates))
tail(ck_cwa(tmin, dates))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comfort.R
\name{ck_heat_index}
\alias{ck_heat_index}
\title{Heat Index}
\usage{
ck_heat_index(tavg, humidity)
}
\arguments{
\item{tavg}{Numeric vector of temperatures (degrees C).}

\item{humidity}{Numeric vector of relative humidity (percent, 0-100).}
}
\value{
A data frame with columns \code{value}, \code{index}, and \code{unit}.
}
\description{
Compute the heat index (apparent temperature) using the Rothfusz
regression equation used by the US National Weather Service.
}
\examples{
ck_heat_index(tavg = c(30, 35, 40), humidity = c(60, 70, 50))
}
\references{
Rothfusz, L. P. (1990). The heat index equation.
NWS Technical Attachment SR 90-23.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drought.R
\name{ck_pet_pm}
\alias{ck_pet_pm}
\title{Reference Evapotranspiration (FAO-56 Penman-Monteith)}
\usage{
ck_pet_pm(
  tmin,
  tmax,
  lat,
  dates,
  elev = 0,
  wind = 2,
  rh_min = NULL,
  rh_max = NULL,
  rs = NULL,
  albedo = 0.23,
  krs = 0.16
)
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{tmax}{Numeric vector of daily maximum temperatures (degrees C),
same length as \code{tmin}.}

\item{lat}{Numeric. Latitude in decimal degrees.}

\item{dates}{Date vector of the same length as \code{tmin}.}

\item{elev}{Numeric. Elevation above sea level in metres
(default 0).}

\item{wind}{Numeric vector or single value. 2-m wind speed (m/s).
Default \code{2} (FAO-56 fallback for unmeasured wind).}

\item{rh_min, rh_max}{Optional numeric vectors of daily minimum and
maximum relative humidity in percent. Both must be supplied
together; otherwise vapour pressure falls back to \code{e0(tmin)}.}

\item{rs}{Optional numeric vector of daily incoming solar
radiation (MJ m^-2 day^-1). If \code{NULL}, Hargreaves-Samani
estimate is used.}

\item{albedo}{Numeric. Surface albedo (default 0.23).}

\item{krs}{Numeric. Hargreaves-Samani coefficient for the Rs
fallback (default 0.16 for inland sites; 0.19 for coastal).}
}
\value{
A data frame with columns \code{date}, \code{value}, \code{index}
(\code{"pet_pm"}), and \code{unit}.
}
\description{
Compute reference evapotranspiration ETo using the FAO-56
Penman-Monteith equation (Allen et al. 1998), the international
standard for ETo estimation. Required inputs are daily Tmin and
Tmax; optional inputs (humidity, wind speed, incoming solar
radiation, elevation) increase accuracy. Where humidity, wind, or
solar radiation are missing, FAO-56 fallback estimators are used.
}
\details{
Inputs and units:
\itemize{
\item \code{tmin}, \code{tmax}: daily minimum and maximum temperature (degrees C).
\item \code{lat}: latitude in decimal degrees, used for the
extraterrestrial-radiation calculation.
\item \code{elev}: elevation above sea level in metres (default 0).
\item \code{wind}: 2-metre wind speed (m/s). Default 2 m/s (the
FAO-56 fallback value when wind data are unavailable).
\item \code{rh_min}, \code{rh_max}: minimum and maximum daily relative
humidity (\\%). If both are NULL (default), actual vapour
pressure is estimated as \code{e0(tmin)} (FAO-56 Eq. 48).
\item \code{rs}: daily incoming solar radiation (MJ m^-2 day^-1).
If NULL, estimated by Hargreaves-Samani: \code{Rs = krs * Ra * sqrt(Tmax - Tmin)} with \code{krs = 0.16} for inland sites.
\item \code{albedo}: surface albedo (default 0.23, the FAO grass
reference).
}
}
\examples{
dates <- as.Date("2024-07-01") + 0:9
tmin <- c(15, 16, 14, 17, 15, 13, 16, 14, 15, 16)
tmax <- c(30, 32, 28, 33, 31, 27, 34, 29, 30, 32)
ck_pet_pm(tmin, tmax, lat = 45, dates = dates)
}
\references{
Allen, R. G., Pereira, L. S., Raes, D., & Smith, M.
(1998). Crop evapotranspiration: guidelines for computing crop
water requirements. \emph{FAO Irrigation and Drainage Paper 56}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{ck_available}
\alias{ck_available}
\title{List All Available Climate Indices}
\usage{
ck_available()
}
\value{
A data frame with columns \code{index}, \code{category}, \code{unit}, and
\code{description}.
}
\description{
Returns a data frame listing every index that \code{climatekit} can compute,
along with its category, unit, and a short description.
}
\examples{
ck_available()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{clear_cache}
\alias{clear_cache}
\title{Clear Cache}
\usage{
clear_cache()
}
\value{
Invisibly returns \code{TRUE} if cache was cleared, \code{FALSE} if no cache
existed.
}
\description{
Removes any cached reference data stored by \code{climatekit}.
}
\examples{
\donttest{
op <- options(climatekit.cache_dir = tempdir())
clear_cache()
options(op)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/climatekit-package.R
\docType{package}
\name{climatekit-package}
\alias{climatekit}
\alias{climatekit-package}
\title{climatekit: Unified Climate Indices for Temperature, Precipitation, and Drought}
\description{
Compute the standard suite of climate indices from daily weather observations. Provides the canonical 'ETCCDI' 27 (Expert Team on Climate Change Detection and Indices), the 'ET-SCI' heatwave and cold-wave families plus the Excess Heat Factor of Nairn and Fawcett (2013), and agroclimatic, drought, and human-comfort families. Drought indices ('SPI', 'SPEI') accept a choice of distribution (gamma or Pearson III for SPI; log-logistic or generalised extreme value for SPEI). Reference evapotranspiration is available via Hargreaves and the FAO-56 Penman-Monteith method (Allen et al. 1998). Percentile-based indices support the Zhang (2005) in-base bootstrap. Daily inputs are numeric vectors plus a 'Date' vector; outputs are tidy data frames. Optional gridded support via 'terra' applies any index over a 'SpatRaster' and reads 'netCDF' input. No external API calls; pairs with data packages such as 'readnoaa'. References: Alexander et al. (2006) \doi{10.1029/2005JD006290}; Zhang et al. (2011) \doi{10.1002/wcc.147}; Zhang et al. (2005) \doi{10.1175/JCLI3366.1}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://charlescoverdale.github.io/climatekit/}
  \item \url{https://github.com/charlescoverdale/climatekit}
  \item Report bugs at \url{https://github.com/charlescoverdale/climatekit/issues}
}

}
\author{
\strong{Maintainer}: Charles Coverdale \email{charlesfcoverdale@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridded.R
\name{ck_apply_grid}
\alias{ck_apply_grid}
\title{Apply a climatekit Index Function Over a SpatRaster}
\usage{
ck_apply_grid(x, fun, dates, ...)
}
\arguments{
\item{x}{A SpatRaster. Layers correspond one-to-one with \code{dates}.}

\item{fun}{A \verb{ck_*} function (or any function with the same
signature: numeric vector + Date vector + optional named
arguments, returning a data frame with \code{period} and \code{value}).}

\item{dates}{Date vector of length \code{terra::nlyr(x)}.}

\item{...}{Additional named arguments forwarded to \code{fun} (for
example \code{period = "annual"}, \code{ref_start}, \code{ref_end}).}
}
\value{
A SpatRaster with one layer per output period (layer names
are the period labels).
}
\description{
Compute a daily climatekit index function for every cell in a
SpatRaster \code{x} whose layers represent successive days, and return
the per-period results as a SpatRaster. The supplied function \code{fun}
must accept a numeric vector and a Date vector and return a data
frame with \code{period} and \code{value} columns (the standard climatekit
shape).
}
\details{
All cells must share the same \code{dates}. Cells that are entirely \code{NA}
are returned as \code{NA}. Run-time scales linearly with the number of
cells; for very large grids consider sub-setting first.
}
\examples{
\donttest{
  if (requireNamespace("terra", quietly = TRUE)) {
    dates <- seq(as.Date("2024-01-01"), as.Date("2024-12-31"), by = "day")
    n <- length(dates)
    # Tiny 2x2 SpatRaster of synthetic daily Tmax
    r <- terra::rast(nrows = 2, ncols = 2, nlyrs = n,
                     xmin = 0, xmax = 2, ymin = 0, ymax = 2)
    set.seed(1)
    for (i in seq_len(n)) {
      terra::values(r[[i]]) <- rnorm(4, 15, 5)
    }
    txx_r <- ck_apply_grid(r, ck_txx, dates = dates, period = "annual")
    terra::nlyr(txx_r)
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temperature.R
\name{ck_heating_degree_days}
\alias{ck_heating_degree_days}
\title{Heating Degree Days}
\usage{
ck_heating_degree_days(tavg, dates, base = 18, period = "annual")
}
\arguments{
\item{tavg}{Numeric vector of daily average temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tavg}.}

\item{base}{Numeric. Base temperature in degrees C (default 18).}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Sum of \code{(base - Tavg)} for all days where daily average temperature is
below the base temperature (default 18 degrees C).
}
\examples{
dates <- as.Date("2024-01-01") + 0:9
tavg <- c(5, 10, 15, 20, 8, 12, 18, 3, 25, 7)
ck_heating_degree_days(tavg, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comfort.R
\name{ck_humidex}
\alias{ck_humidex}
\title{Humidex}
\usage{
ck_humidex(tavg, dewpoint)
}
\arguments{
\item{tavg}{Numeric vector of temperatures (degrees C).}

\item{dewpoint}{Numeric vector of dewpoint temperatures (degrees C).}
}
\value{
A data frame with columns \code{value}, \code{index}, and \code{unit}.
}
\description{
Compute the Canadian humidex from temperature and dewpoint.
}
\examples{
ck_humidex(tavg = c(30, 35), dewpoint = c(20, 25))
}
\references{
Masterson, J., & Richardson, F. A. (1979). Humidex: A method
of quantifying human discomfort due to excessive heat and humidity.
Environment Canada.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drought.R
\name{ck_pet}
\alias{ck_pet}
\title{Potential Evapotranspiration (Hargreaves Method)}
\usage{
ck_pet(tmin, tmax, lat, dates)
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{tmax}{Numeric vector of daily maximum temperatures (degrees C).}

\item{lat}{Numeric. Latitude in decimal degrees.}

\item{dates}{Date vector of the same length as \code{tmin}.}
}
\value{
A data frame with columns \code{date}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Estimate daily PET using the Hargreaves-Samani equation, which requires
only daily temperature extremes and latitude.
}
\examples{
dates <- as.Date("2024-07-01") + 0:9
tmin <- c(15, 16, 14, 17, 15, 13, 16, 14, 15, 16)
tmax <- c(30, 32, 28, 33, 31, 27, 34, 29, 30, 32)
ck_pet(tmin, tmax, lat = 45, dates = dates)
}
\references{
Hargreaves, G. H., & Samani, Z. A. (1985). Reference crop
evapotranspiration from temperature. \emph{Applied Engineering in Agriculture},
1(2), 96-99.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temperature.R
\name{ck_summer_days}
\alias{ck_summer_days}
\title{Summer Days}
\usage{
ck_summer_days(tmax, dates, period = "annual")
}
\arguments{
\item{tmax}{Numeric vector of daily maximum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmax}.}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Count the number of days where maximum temperature exceeds 25 degrees C.
}
\examples{
dates <- as.Date("2024-07-01") + 0:9
tmax <- c(22, 26, 28, 24, 30, 25, 27, 23, 31, 29)
ck_summer_days(tmax, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temperature.R
\name{ck_txn}
\alias{ck_txn}
\title{Annual or Monthly Minimum of Daily Maximum Temperature (TXn)}
\usage{
ck_txn(tmax, dates, period = "annual")
}
\arguments{
\item{tmax}{Numeric vector of daily maximum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmax}.}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
ETCCDI canonical index TXn. The minimum value of daily maximum
temperature (Tmax) within each reporting period (coldest day).
}
\examples{
dates <- as.Date("2024-01-01") + 0:9
tmax <- c(5, 10, -3, 12, 4, 8, 22, -8, 7, 6)
ck_txn(tmax, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comfort.R
\name{ck_wind_chill}
\alias{ck_wind_chill}
\title{Wind Chill Temperature}
\usage{
ck_wind_chill(tavg, wind_speed)
}
\arguments{
\item{tavg}{Numeric vector of temperatures (degrees C).}

\item{wind_speed}{Numeric vector of wind speeds (km/h).}
}
\value{
A data frame with columns \code{value}, \code{index}, and \code{unit}.
}
\description{
Compute wind chill using the North American Wind Chill Index formula
(Environment Canada / US NWS). Valid for temperatures at or below 10 degrees C
and wind speeds above 4.8 km/h.
}
\examples{
ck_wind_chill(tavg = c(-5, -10, 0), wind_speed = c(20, 30, 15))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precipitation.R
\name{ck_max_1day_precip}
\alias{ck_max_1day_precip}
\title{Maximum 1-Day Precipitation}
\usage{
ck_max_1day_precip(precip, dates, period = "annual")
}
\arguments{
\item{precip}{Numeric vector of daily precipitation (mm).}

\item{dates}{Date vector of the same length as \code{precip}.}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Maximum precipitation recorded in a single day per period.
}
\examples{
dates <- as.Date("2024-01-01") + 0:9
precip <- c(0, 5, 22, 0, 15, 25, 0, 11, 4, 30)
ck_max_1day_precip(precip, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/heatwave.R
\name{ck_cwn}
\alias{ck_cwn}
\title{Cold-Wave Number (CWN)}
\usage{
ck_cwn(tmin, dates, ref_start = 1961L, ref_end = 1990L, min_spell = 3L)
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmax}. Must contain
data covering the reference period.}

\item{ref_start, ref_end}{Integer. Reference period boundary years
(inclusive). Defaults to 1961 and 1990.}

\item{min_spell}{Integer. Minimum spell length in days (default 3,
the ET-SCI standard).}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
ET-SCI cold-wave family index. Annual count of distinct cold-wave
events, where a cold wave is a span of at least three consecutive
days with daily Tmin below the 10th percentile of the calendar-day
distribution from a reference period (default 1961 to 1990).
}
\examples{
set.seed(1)
dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
tmin <- 5 + 8 * sin(2 * pi * as.integer(format(dates, "\%j")) / 365) +
        rnorm(length(dates))
tail(ck_cwn(tmin, dates))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/agroclimatic.R
\name{ck_winkler}
\alias{ck_winkler}
\title{Winkler Index}
\usage{
ck_winkler(tavg, dates)
}
\arguments{
\item{tavg}{Numeric vector of daily average temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tavg}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
The Winkler index (also called growing degree days for viticulture)
accumulates daily mean temperature above 10 degrees C during the growing season
(April-October in NH, October-April in SH).
}
\examples{
dates <- seq(as.Date("2024-04-01"), as.Date("2024-10-31"), by = "day")
set.seed(42)
tavg <- rnorm(length(dates), mean = 18, sd = 4)
ck_winkler(tavg, dates)
}
\references{
Amerine, M. A., & Winkler, A. J. (1944). Composition and
quality of musts and wines of California grapes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ehf.R
\name{ck_ehf}
\alias{ck_ehf}
\title{Excess Heat Factor (EHF, Nairn & Fawcett 2013)}
\usage{
ck_ehf(
  tmax,
  tmin,
  dates,
  ref_start = 1961L,
  ref_end = 1990L,
  stat = c("max", "n_positive", "sum_positive")
)
}
\arguments{
\item{tmax}{Numeric vector of daily maximum temperatures (degrees C).}

\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmax}. Must contain
data covering the reference period.}

\item{ref_start, ref_end}{Integer. Reference period boundary years
(inclusive). Defaults to 1961 and 1990.}

\item{stat}{One of \code{"max"}, \code{"n_positive"}, or \code{"sum_positive"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Annual summary of the daily Excess Heat Factor heatwave intensity
metric. EHF combines a 3-day mean daily temperature anomaly above
the 95th percentile of the reference period with an acclimatisation
term (3-day mean minus previous 30-day mean). Positive EHF days
indicate heatwave conditions; larger values indicate more severe
or less-acclimatised events. This is the operational heatwave
metric used by the Australian Bureau of Meteorology.
}
\details{
Three annual summaries are exposed via the \code{stat} argument:
\itemize{
\item \code{"max"} (default): peak EHF in the year. Strongest
single-day intensity.
\item \code{"n_positive"}: count of days with \code{EHF > 0}. A
frequency-of-heatwave-conditions measure.
\item \code{"sum_positive"}: sum of EHF on days with \code{EHF > 0}.
A severity-weighted total.
}
}
\examples{
set.seed(1)
dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
s <- 2 * pi * as.integer(format(dates, "\%j")) / 365
tmax <- 20 + 10 * sin(s) + rnorm(length(dates))
tmin <- 10 +  8 * sin(s) + rnorm(length(dates))
tail(ck_ehf(tmax, tmin, dates))
}
\references{
Nairn, J. R., & Fawcett, R. J. B. (2013). Defining heatwaves:
heatwave defined as a heat-impact event servicing all community and
business sectors in Australia. \emph{CAWCR Technical Report No. 060}.

Perkins, S. E., & Alexander, L. V. (2013). On the measurement of
heatwaves. \emph{Journal of Climate}, 26(13), 4500-4517.
\doi{10.1175/JCLI-D-12-00383.1}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drought.R
\name{ck_spei}
\alias{ck_spei}
\title{Standardized Precipitation-Evapotranspiration Index (SPEI)}
\usage{
ck_spei(precip, pet, dates, scale = 3, distribution = c("log-logistic", "gev"))
}
\arguments{
\item{precip}{Numeric vector of daily precipitation (mm).}

\item{pet}{Numeric vector of daily potential evapotranspiration (mm).}

\item{dates}{Date vector of the same length as \code{precip} and \code{pet}.}

\item{scale}{Integer. Accumulation period in months (default 3).}

\item{distribution}{Character. Either \code{"log-logistic"} (default,
Vicente-Serrano et al. 2010) or \code{"gev"} (Generalised Extreme
Value, fitted by L-moments; preferred for water-balance series
with heavy upper or lower tails).}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Compute the SPEI by fitting a log-logistic distribution to the monthly
climatic water balance (precipitation minus potential evapotranspiration)
accumulated over a rolling window.
}
\examples{
dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
set.seed(42)
precip <- rgamma(length(dates), shape = 0.5, rate = 0.1)
pet <- rep(3, length(dates))
ck_spei(precip, pet, dates, scale = 3)
}
\references{
Vicente-Serrano, S. M., Begueria, S., & Lopez-Moreno, J. I.
(2010). A multiscalar drought index sensitive to global warming: the
Standardized Precipitation Evapotranspiration Index. \emph{Journal of Climate},
23(7), 1696-1718.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precipitation.R
\name{ck_r95p}
\alias{ck_r95p}
\title{Very Wet Days Total (R95p)}
\usage{
ck_r95p(precip, dates, ref_start = 1961L, ref_end = 1990L, period = "annual")
}
\arguments{
\item{precip}{Numeric vector of daily precipitation (mm).}

\item{dates}{Date vector of the same length as \code{precip}. Must contain
data covering the reference period.}

\item{ref_start, ref_end}{Integer. Reference period boundary years
(inclusive). Defaults to 1961 and 1990.}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
ETCCDI canonical index R95p. Annual total precipitation on days where
daily precipitation exceeds the 95th percentile of wet-day precipitation
in a reference period (default 1961 to 1990). A wet day is one with
precipitation at or above 1 mm. The threshold is a single value derived
from all wet days in the reference period (not calendar-day specific).
}
\examples{
set.seed(1)
dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
precip <- pmax(rgamma(length(dates), shape = 0.4, scale = 8) - 1, 0)
tail(ck_r95p(precip, dates))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comfort.R
\name{ck_fire_danger}
\alias{ck_fire_danger}
\title{Fire Danger Index (Simplified)}
\usage{
ck_fire_danger(tavg, humidity, wind_speed, precip)
}
\arguments{
\item{tavg}{Numeric vector of temperatures (degrees C).}

\item{humidity}{Numeric vector of relative humidity (percent, 0-100).}

\item{wind_speed}{Numeric vector of wind speeds (km/h).}

\item{precip}{Numeric vector of daily precipitation (mm).}
}
\value{
A data frame with columns \code{value}, \code{index}, and \code{unit}.
}
\description{
A simplified fire danger proxy based on temperature, humidity, wind speed,
and recent precipitation. This is NOT the Canadian Forest Fire Weather
Index (Van Wagner 1987); for the full FWI system, use the cffdrs package.
}
\examples{
ck_fire_danger(
  tavg = c(30, 25, 35),
  humidity = c(20, 40, 15),
  wind_speed = c(25, 10, 30),
  precip = c(0, 5, 0)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{ck_etccdi_27}
\alias{ck_etccdi_27}
\title{Canonical ETCCDI 27 Indices}
\usage{
ck_etccdi_27()
}
\value{
A data frame with one row per canonical ETCCDI index and
columns \code{code}, \code{name}, \code{variable}, \code{unit}, \code{definition},
\code{ck_function}, and \code{status}.
}
\description{
Returns the 27 canonical Expert Team on Climate Change Detection and
Indices (ETCCDI) indices as documented by Alexander et al. (2006) and
Zhang et al. (2011), with each row showing the canonical short code,
full name, input variable, unit, definition, and the corresponding
\code{climatekit} function (or \code{NA} where the index is not yet
implemented).
}
\details{
Use this table to audit coverage, locate the \verb{ck_*} function for a
given ETCCDI code, or filter to indices that climatekit currently
implements: \code{subset(ck_etccdi_27(), !is.na(ck_function))}.
}
\examples{
tab <- ck_etccdi_27()
head(tab)
# Indices currently implemented in climatekit:
subset(tab, !is.na(ck_function))[, c("code", "ck_function")]
# Coverage:
table(tab$status)
}
\references{
Alexander, L. V. et al. (2006). Global observed changes in daily
climate extremes of temperature and precipitation. \emph{Journal of
Geophysical Research: Atmospheres}, 111(D5).
\doi{10.1029/2005JD006290}.

Zhang, X. et al. (2011). Indices for monitoring changes in extremes
based on daily temperature and precipitation data. \emph{Wiley
Interdisciplinary Reviews: Climate Change}, 2(6), 851-870.
\doi{10.1002/wcc.147}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/heatwave.R
\name{ck_hwd}
\alias{ck_hwd}
\title{Heatwave Duration (HWD)}
\usage{
ck_hwd(tmax, dates, ref_start = 1961L, ref_end = 1990L, min_spell = 3L)
}
\arguments{
\item{tmax}{Numeric vector of daily maximum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmax}. Must contain
data covering the reference period.}

\item{ref_start, ref_end}{Integer. Reference period boundary years
(inclusive). Defaults to 1961 and 1990.}

\item{min_spell}{Integer. Minimum spell length in days (default 3,
the ET-SCI standard).}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
ET-SCI heatwave family index. Length in days of the longest heatwave
event in each year (see \code{\link[=ck_hwn]{ck_hwn()}} for the heatwave definition).
}
\examples{
set.seed(1)
dates <- seq(as.Date("1961-01-01"), as.Date("1991-12-31"), by = "day")
tmax <- 15 + 10 * sin(2 * pi * as.integer(format(dates, "\%j")) / 365) +
        rnorm(length(dates))
tail(ck_hwd(tmax, dates))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridded.R
\name{ck_from_netcdf}
\alias{ck_from_netcdf}
\title{Read a netCDF File as a SpatRaster}
\usage{
ck_from_netcdf(path, var = NULL)
}
\arguments{
\item{path}{Character. Path to a netCDF file.}

\item{var}{Character or \code{NULL}. Variable to extract. If \code{NULL}, the
default behaviour of \code{\link[terra:rast]{terra::rast()}} applies.}
}
\value{
A SpatRaster (one layer per time step in the netCDF file).
}
\description{
Convenience wrapper that delegates to \code{\link[terra:rast]{terra::rast()}}. Reads the
file at \code{path} and returns a SpatRaster, optionally restricted to a
single variable. terra and ncdf4 must be installed (both are listed
in \verb{Suggests:}).
}
\examples{
\dontrun{
  r <- ck_from_netcdf("tas_day.nc", var = "tas")
  terra::nlyr(r)  # number of daily layers
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temperature.R
\name{ck_warm_spell}
\alias{ck_warm_spell}
\title{Warm Spell Days (Series-Quantile Approximation)}
\usage{
ck_warm_spell(tmax, dates, threshold = 0.9, period = "annual")
}
\arguments{
\item{tmax}{Numeric vector of daily maximum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmax}.}

\item{threshold}{Numeric. Quantile threshold (default 0.9, i.e. 90th
percentile).}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Count the number of days in warm spells, where a warm spell is defined
as at least six consecutive days with Tmax above the \code{threshold}
quantile of the full input series. This is a quick approximation
driven by a single series-wide quantile, and does not require a
reference period.
}
\details{
For the canonical ETCCDI WSDI definition (1961-1990 calendar-day base,
six-day spell rule), use \code{\link[=ck_wsdi]{ck_wsdi()}}.
}
\examples{
dates <- as.Date("2024-01-01") + 0:364
set.seed(42)
tmax <- rnorm(365, mean = 20, sd = 5)
ck_warm_spell(tmax, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precipitation.R
\name{ck_total_precip}
\alias{ck_total_precip}
\title{Total Wet-Day Precipitation (PRCPTOT)}
\usage{
ck_total_precip(precip, dates, period = "annual", wet_day_threshold = 1)
}
\arguments{
\item{precip}{Numeric vector of daily precipitation (mm).}

\item{dates}{Date vector of the same length as \code{precip}.}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}

\item{wet_day_threshold}{Numeric (mm). Days with precipitation
strictly below this threshold are excluded from the sum. Default 1.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Annual or monthly total precipitation in wet days, where a wet day is
a day with precipitation at or above \code{wet_day_threshold} mm (default
1 mm, the ETCCDI standard). This is the canonical ETCCDI 'PRCPTOT'
definition (Alexander et al. 2006; Zhang et al. 2011).
}
\details{
Sub-threshold trace amounts are excluded. Pass
\code{wet_day_threshold = 0} to recover the previous behaviour of
summing all daily values.
}
\examples{
dates <- as.Date("2024-01-01") + 0:9
precip <- c(0, 5, 3, 0, 8, 2, 0, 1, 4, 0)
ck_total_precip(precip, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{ck_catalogue}
\alias{ck_catalogue}
\title{climatekit Index Catalogue}
\usage{
ck_catalogue()
}
\value{
A data frame with columns \code{ck_function}, \code{code}, \code{name},
\code{category}, \code{sector}, \code{unit}, \code{standard}, and \code{citation_key}.
}
\description{
Returns the complete catalogue of climate indices implemented by
\code{climatekit}, with one row per \verb{ck_*} function and columns covering
the canonical short code (where applicable), the full name, the
index family, the relevant sector, the unit, the source standard,
and the principal citation key.
}
\details{
Use \code{\link[=ck_browse]{ck_browse()}} to filter by sector or standard.
}
\examples{
tab <- ck_catalogue()
head(tab)
# Tally indices by standard:
table(tab$standard)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drought.R
\name{ck_spi}
\alias{ck_spi}
\title{Standardized Precipitation Index (SPI)}
\usage{
ck_spi(precip, dates, scale = 3, distribution = c("gamma", "pearsonIII"))
}
\arguments{
\item{precip}{Numeric vector of daily precipitation (mm).}

\item{dates}{Date vector of the same length as \code{precip}.}

\item{scale}{Integer. Accumulation period in months (default 3).}

\item{distribution}{Character. Either \code{"gamma"} (default) or
\code{"pearsonIII"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Compute the SPI by fitting a parametric distribution to rolling monthly
precipitation accumulations and transforming to standard normal
deviates. Two distributions are supported: the two-parameter gamma
(default; WMO-1090 standard) and the three-parameter Pearson III. The
Pearson III tail is heavier and is preferred in arid regions where the
wet-day distribution is highly skewed (Stagge et al. 2015).
}
\examples{
dates <- seq(as.Date("2020-01-01"), as.Date("2023-12-31"), by = "day")
set.seed(42)
precip <- rgamma(length(dates), shape = 0.5, rate = 0.1)
ck_spi(precip, dates, scale = 3)
}
\references{
McKee, T. B., Doesken, N. J., & Kleist, J. (1993). The relationship of
drought frequency and duration to time scales.

Stagge, J. H., Tallaksen, L. M., Gudmundsson, L., Van Loon, A. F., &
Stahl, K. (2015). Candidate distributions for climatological drought
indices (SPI and SPEI). \emph{International Journal of Climatology}, 35(13),
4027-4040. \doi{10.1002/joc.4267}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{ck_metadata}
\alias{ck_metadata}
\title{Get Metadata for a Climate Index}
\usage{
ck_metadata(index)
}
\arguments{
\item{index}{Character string. The index name (e.g. \code{"frost_days"}).
Use \code{\link[=ck_available]{ck_available()}} to see valid names.}
}
\value{
A list with elements \code{index}, \code{category}, \code{unit}, \code{description},
and \code{reference}.
}
\description{
Returns metadata (unit, category, description, reference) for a named
climate index.
}
\examples{
ck_metadata("frost_days")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/agroclimatic.R
\name{ck_branas}
\alias{ck_branas}
\title{Branas Hydrothermal Index}
\usage{
ck_branas(precip, tavg, dates, lat = 50)
}
\arguments{
\item{precip}{Numeric vector of daily precipitation (mm).}

\item{tavg}{Numeric vector of daily mean temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{precip}.}

\item{lat}{Numeric. Latitude in decimal degrees, used to select the
hemisphere convention. Default 50 (Northern Hemisphere).}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
The Branas index combines temperature and precipitation during the
growing season to estimate disease pressure (especially downy mildew)
in vineyards. It is the sum of (monthly mean temperature) times
(monthly precipitation total) over the five months of the growing
season: April-August in the Northern Hemisphere; October-February
in the Southern Hemisphere. The Southern Hemisphere season spans
two calendar years and is reported under the year in which it
starts.
}
\examples{
dates <- seq(as.Date("2024-04-01"), as.Date("2024-08-31"), by = "day")
set.seed(42)
tavg <- rnorm(length(dates), mean = 12, sd = 3)
precip <- rgamma(length(dates), shape = 0.5, rate = 0.2)
ck_branas(precip, tavg, dates)
}
\references{
Branas, J., Bernon, G., & Levadoux, L. (1946).
Elements de viticulture generale.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temperature.R
\name{ck_tropical_nights}
\alias{ck_tropical_nights}
\title{Tropical Nights}
\usage{
ck_tropical_nights(tmin, dates, period = "annual")
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmin}.}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Count the number of days where minimum temperature exceeds 20 degrees C.
}
\examples{
dates <- as.Date("2024-07-01") + 0:9
tmin <- c(18, 21, 22, 19, 25, 20, 23, 17, 24, 21)
ck_tropical_nights(tmin, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temperature.R
\name{ck_growing_season}
\alias{ck_growing_season}
\title{Growing Season Length}
\usage{
ck_growing_season(tavg, dates, lat = 50)
}
\arguments{
\item{tavg}{Numeric vector of daily mean temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tavg}.}

\item{lat}{Numeric. Latitude in decimal degrees (used to determine
hemisphere for end-of-season rule). Default 50 (Northern Hemisphere).}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
Compute the growing season length following the ETCCDI definition: the
number of days between the first occurrence of at least 6 consecutive days
with daily mean temperature above 5 degrees C and the first span of 6
consecutive days with Tmean below 5 degrees C after July 1 (Northern
Hemisphere) or January 1 (Southern Hemisphere). Calculated per year.
}
\examples{
dates <- as.Date("2024-01-01") + 0:364
set.seed(42)
tavg <- sin(seq(0, 2 * pi, length.out = 365)) * 15 + 5
ck_growing_season(tavg, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{ck_convert_temp}
\alias{ck_convert_temp}
\title{Convert Temperature Units}
\usage{
ck_convert_temp(x, from, to)
}
\arguments{
\item{x}{Numeric vector of temperatures.}

\item{from}{Character. Source unit: \code{"C"}, \code{"F"}, or \code{"K"}.}

\item{to}{Character. Target unit: \code{"C"}, \code{"F"}, or \code{"K"}.}
}
\value{
Numeric vector of converted temperatures.
}
\description{
Convert between Celsius, Fahrenheit, and Kelvin.
}
\examples{
ck_convert_temp(c(0, 100), from = "C", to = "F")
ck_convert_temp(32, from = "F", to = "C")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temperature.R
\name{ck_tnn}
\alias{ck_tnn}
\title{Annual or Monthly Minimum of Daily Minimum Temperature (TNn)}
\usage{
ck_tnn(tmin, dates, period = "annual")
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmin}.}

\item{period}{Character. Aggregation period: \code{"annual"} (default) or
\code{"monthly"}.}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
ETCCDI canonical index TNn. The minimum value of daily minimum
temperature (Tmin) within each reporting period (coldest night).
}
\examples{
dates <- as.Date("2024-01-01") + 0:9
tmin <- c(-2, 3, -1, 5, -8, 0, 2, -12, 1, -1)
ck_tnn(tmin, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/agroclimatic.R
\name{ck_first_frost}
\alias{ck_first_frost}
\title{First Frost Date}
\usage{
ck_first_frost(tmin, dates, lat = 50)
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmin}.}

\item{lat}{Numeric. Latitude in decimal degrees, used to select the
hemisphere convention. Default 50 (Northern Hemisphere).}
}
\value{
A data frame with columns \code{period}, \code{value} (day of year),
\code{date} (the frost date), \code{index}, and \code{unit}.
}
\description{
Day of year of the first autumn frost (Tmin < 0 degrees C) in each
year. Hemisphere is selected by \code{lat}: in the Northern Hemisphere
(\code{lat >= 0}) the search starts at July 1 (DOY 183); in the Southern
Hemisphere it starts at March 1 (DOY 60), matching the autumn entry
for each.
}
\examples{
dates <- seq(as.Date("2024-07-01"), as.Date("2024-12-31"), by = "day")
set.seed(42)
tmin <- 15 - seq_along(dates) * 0.15 + rnorm(length(dates), sd = 3)
ck_first_frost(tmin, dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/agroclimatic.R
\name{ck_huglin}
\alias{ck_huglin}
\title{Huglin Heliothermal Index}
\usage{
ck_huglin(tmin, tmax, dates, lat)
}
\arguments{
\item{tmin}{Numeric vector of daily minimum temperatures (degrees C).}

\item{tmax}{Numeric vector of daily maximum temperatures (degrees C).}

\item{dates}{Date vector of the same length as \code{tmin}.}

\item{lat}{Numeric. Latitude in decimal degrees (used to determine
hemisphere and day-length coefficient).}
}
\value{
A data frame with columns \code{period}, \code{value}, \code{index}, and \code{unit}.
}
\description{
The Huglin index is used in viticulture to characterise the thermal
potential of a region for grape growing. It is computed over the growing
season (April 1 to September 30 in the Northern Hemisphere; October 1 to
March 31 in the Southern Hemisphere).
}
\examples{
dates <- seq(as.Date("2024-04-01"), as.Date("2024-09-30"), by = "day")
set.seed(42)
tmin <- rnorm(length(dates), mean = 12, sd = 3)
tmax <- tmin + runif(length(dates), 8, 15)
ck_huglin(tmin, tmax, dates, lat = 45)
}
\references{
Huglin, P. (1978). Nouveau mode d'evaluation des possibilites
heliothermiques d'un milieu viticole. \emph{Comptes Rendus de l'Academie
d'Agriculture de France}, 64, 1117-1126.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{ck_browse}
\alias{ck_browse}
\title{Browse the climatekit Index Catalogue}
\usage{
ck_browse(sector = NULL, standard = NULL, search = NULL)
}
\arguments{
\item{sector}{Character (length 1) or \code{NULL}. Filter to indices tagged
with this sector. Common values: \code{"agriculture"}, \code{"health"},
\code{"water"}, \code{"energy"}. \code{NULL} (default) returns all rows.}

\item{standard}{Character (length 1) or \code{NULL}. Filter to indices
under this standard. Common values: \code{"ETCCDI"}, \code{"ET-SCI"},
\code{"ETCCDI-approx"}, \code{"agroclimatic"}, \code{"comfort"}, \code{"drought"},
\code{"energy"}. \code{NULL} (default) returns all rows.}

\item{search}{Character (length 1) or \code{NULL}. Free-text search; rows
are kept where the term appears (case-insensitive) in the function
name, the full name, or the ETCCDI code.}
}
\value{
The catalogue, filtered to matching rows. Same column structure
as \code{\link[=ck_catalogue]{ck_catalogue()}}.
}
\description{
Filter the comprehensive climatekit index catalogue (see
\code{\link[=ck_catalogue]{ck_catalogue()}}) by sector, applicable standard (ETCCDI / ET-SCI /
agroclimatic / comfort / drought / energy), or a free-text search
across the function name, full name, and ETCCDI code.
}
\examples{
ck_browse(standard = "ETCCDI")
ck_browse(sector = "agriculture")
ck_browse(search = "heat")
}
